searchNodes=[{"doc":"Oban is a robust job processing library which uses PostgreSQL for storage and coordination.Each Oban instance is a supervision tree and not an application. That means it won't be started automatically and must be included in your application's supervision tree. All of your configuration is passed into the supervisor, allowing you to configure Oban like the rest of your application:# config/config.exs config :my_app, Oban, repo: MyApp.Repo, plugins: [Oban.Plugins.Pruner], queues: [default: 10, events: 50, media: 20] # lib/my_app/application.ex defmodule MyApp.Application do @moduledoc false use Application alias MyApp.Repo alias MyAppWeb.Endpoint def start(_type, _args) do children = [ Repo, Endpoint, {Oban, oban_config()} ] Supervisor.start_link(children, strategy: :one_for_one, name: MyApp.Supervisor) end # Conditionally disable crontab, queues, or plugins here. defp oban_config do Application.get_env(:my_app, Oban) end endIf you are running tests (which you should be) you'll want to disable plugins, enqueueing scheduled jobs and job dispatching altogether when testing:# config/test.exs config :my_app, Oban, crontab: false, queues: false, plugins: falseSee the installation instructions in the README or on the Hexdocs guide for details on how to migrate your database.Configuring QueuesQueues are specified as a keyword list where the key is the name of the queue and the value is the maximum number of concurrent jobs. The following configuration would start four queues with concurrency ranging from 5 to 50:queues: [default: 10, mailers: 20, events: 50, media: 5]You may also use an expanded form to configure queues with individual overrides:queues: [ default: 10, mailers: [limit: 20, poll_interval: :timer.seconds(30)], events: [limit: 50, paused: true] ]Here we've declared the mailers with the same limit as above, but it will only poll to deschedule jobs every 30 seconds. The events queue will start in a paused state, which means it won't process anything until Oban.resume_queue/2 is called to start it.There isn't a limit to the number of queues or how many jobs may execute concurrently in each queue. Some additional guidelines:Caveats &amp; GuidelinesEach queue will run as many jobs as possible concurrently, up to the configured limit. Make sure your system has enough resources (i.e. database connections) to handle the concurrent load.Queue limits are local (per-node), not global (per-cluster). For example, running a queue with a local limit of one on three separate nodes is effectively a global limit of three. If you require a global limit you must restrict the number of nodes running a particular queue.Only jobs in the configured queues will execute. Jobs in any other queue will stay in the database untouched.Be careful how many concurrent jobs make expensive system calls (i.e. FFMpeg, ImageMagick). The BEAM ensures that the system stays responsive under load, but those guarantees don't apply when using ports or shelling out commands.Defining WorkersWorker modules do the work of processing a job. At a minimum they must define a perform/1 function, which is called with an %Oban.Job{} struct.Note that the args field of the job struct will always have string keys, regardless of the key type when the job was enqueued. The args are stored as jsonb in PostgreSQL and the serialization process automatically stringifies all keys.Define a worker to process jobs in the events queue:defmodule MyApp.Business do use Oban.Worker, queue: :events @impl Oban.Worker def perform(%Oban.Job{args: %{&quot;id&quot; =&gt; id} = args}) do model = MyApp.Repo.get(MyApp.Business.Man, id) case args do %{&quot;in_the&quot; =&gt; &quot;business&quot;} -&gt; IO.inspect(model) %{&quot;vote_for&quot; =&gt; vote} -&gt; IO.inspect([vote, model]) _ -&gt; IO.inspect(model) end :ok end endThe use macro also accepts options to customize max attempts, priority, tags, and uniqueness:defmodule MyApp.LazyBusiness do use Oban.Worker, queue: :events, priority: 3, max_attempts: 3, tags: [&quot;business&quot;], unique: [period: 30] @impl Oban.Worker def perform(_job) do # do business slowly :ok end endSuccessful jobs should return :ok or an {:ok, value} tuple. The value returned from perform/1 is used to control whether the job is treated as a success, a failure, discarded completely or deferred until later.See the Oban.Worker docs for more details on failure conditions and Oban.Telemetry for details on job reporting.Enqueueing JobsJobs are simply Ecto structs and are enqueued by inserting them into the database. For convenience and consistency all workers provide a new/2 function that converts an args map into a job changeset suitable for insertion:%{id: 1, in_the: &quot;business&quot;, of_doing: &quot;business&quot;} |&gt; MyApp.Business.new() |&gt; Oban.insert()The worker's defaults may be overridden by passing options:%{id: 1, vote_for: &quot;none of the above&quot;} |&gt; MyApp.Business.new(queue: :special, max_attempts: 5) |&gt; Oban.insert()Jobs may be scheduled at a specific datetime in the future:%{id: 1} |&gt; MyApp.Business.new(scheduled_at: ~U[2020-12-25 19:00:56.0Z]) |&gt; Oban.insert()Jobs may also be scheduled down to the second any time in the future:%{id: 1} |&gt; MyApp.Business.new(schedule_in: 5) |&gt; Oban.insert()Unique jobs can be configured in the worker, or when the job is built:%{email: &quot;brewster@example.com&quot;} |&gt; MyApp.Mailer.new(unique: [period: 300, fields: [:queue, :worker]) |&gt; Oban.insert()Job priority can be specified using an integer from 0 to 3, with 0 being the default and highest priority:%{id: 1} |&gt; MyApp.Backfiller.new(priority: 2) |&gt; Oban.insert()Any number of tags can be added to a job dynamically, at the time it is inserted:id = 1 %{id: id} |&gt; MyApp.OnboardMailer.new(tags: [&quot;mailer&quot;, &quot;record-\#{id}&quot;]) |&gt; Oban.insert()Multiple jobs can be inserted in a single transaction:Ecto.Multi.new() |&gt; Oban.insert(:b_job, MyApp.Business.new(%{id: 1})) |&gt; Oban.insert(:m_job, MyApp.Mailer.new(%{email: &quot;brewser@example.com&quot;})) |&gt; Repo.transaction()Occasionally you may need to insert a job for a worker that exists in another application. In that case you can use Oban.Job.new/2 to build the changeset manually:%{id: 1, user_id: 2} |&gt; Oban.Job.new(queue: :default, worker: OtherApp.Worker) |&gt; Oban.insert()Oban.insert/2,4 is the preferred way of inserting jobs as it provides some of Oban's advanced features (i.e., unique jobs). However, you can use your application's Repo.insert/2 function if necessary.See Oban.Job.new/2 for a full list of job options.Pruning Historic JobsJob stats and queue introspection are built on keeping job rows in the database after they have completed. This allows administrators to review completed jobs and build informative aggregates, at the expense of storage and an unbounded table size. To prevent the oban_jobs table from growing indefinitely, Oban provides active pruning of completed and discarded jobs.By default, pruning retains jobs for 60 seconds. You can configure a longer retention period by providing a max_age in seconds to the Pruner plugin.# Set the max_age for 5 minutes config :my_app, Oban, plugins: [{Oban.Plugins.Pruner, max_age: 300}] ...Caveats &amp; GuidelinesPruning is best-effort and performed out-of-band. This means that all limits are soft; jobs beyond a specified age may not be pruned immediately after jobs complete.Pruning is only applied to jobs that are completed or discarded (has reached the maximum number of retries or has been manually killed). It'll never delete a new job, a scheduled job or a job that failed and will be retried.Unique JobsThe unique jobs feature lets you specify constraints to prevent enqueueing duplicate jobs. Uniqueness is based on a combination of args, queue, worker, state and insertion time. It is configured at the worker or job level using the following options::period — The number of seconds until a job is no longer considered duplicate. You should always specify a period. :infinity can be used to indicate the job be considered a duplicate as long as jobs are retained.:fields — The fields to compare when evaluating uniqueness. The available fields are :args, :queue and :worker, by default all three are used.:keys — A specific subset of the :args to consider when comparing against historic jobs. This allows a job with multiple key/value pairs in the args to be compared using only a subset of them.:states — The job states that are checked for duplicates. The available states are :available, :scheduled, :executing, :retryable, :completed and :discarded. By default all states except for :discarded are checked, which prevents duplicates even if the previous job has been completed.For example, configure a worker to be unique across all fields and states for 60 seconds:use Oban.Worker, unique: [period: 60]Configure the worker to be unique only by :worker and :queue:use Oban.Worker, unique: [fields: [:queue, :worker], period: 60]Or, configure a worker to be unique until it has executed:use Oban.Worker, unique: [period: 300, states: [:available, :scheduled, :executing]]Only consider the :url key rather than the entire args:use Oban.Worker, unique: [fields: [:args, :worker], keys: [:url]]You can use Oban.Job.states/0 to specify uniqueness across all states, including :discarded:use Oban.Worker, unique: [period: 300, states: Oban.Job.states()]Strong GuaranteesUnique jobs are guaranteed through transactional locks and database queries: they do not rely on unique constraints in the database. This makes uniqueness entirely configurable by application code, without the need for database migrations.Performance NoteIf your application makes heavy use of unique jobs you may want to add an index on the args column of the oban_jobs table. The other columns considered for uniqueness are already covered by indexes.Periodic JobsOban allows jobs to be registered with a cron-like schedule and enqueued automatically. Periodic jobs are registered as a list of {cron, worker} or {cron, worker, options} tuples:config :my_app, Oban, repo: MyApp.Repo, crontab: [ {&quot;* * * * *&quot;, MyApp.MinuteWorker}, {&quot;0 * * * *&quot;, MyApp.HourlyWorker, args: %{custom: &quot;arg&quot;}}, {&quot;0 0 * * *&quot;, MyApp.DailyWorker, max_attempts: 1}, {&quot;0 12 * * MON&quot;, MyApp.MondayWorker, queue: :scheduled, tags: [&quot;mondays&quot;]}, {&quot;@daily&quot;, MyApp.AnotherDailyWorker} ]These jobs would be executed as follows:MyApp.MinuteWorker — Executed once every minuteMyApp.HourlyWorker — Executed at the first minute of every hour with custom argsMyApp.DailyWorker — Executed at midnight every day with no retriesMyApp.MondayWorker — Executed at noon every Monday in the &quot;scheduled&quot; queueMyApp.AnotherDailyWorker — Executed at midnight every day with no retriesThe crontab format respects all standard rules and has one minute resolution. Jobs are considered unique for most of each minute, which prevents duplicate jobs with multiple nodes and across node restarts.Like other jobs, recurring jobs will use the :queue specified by the worker module (or :default if one is not specified).Cron ExpressionsStandard Cron expressions are composed of rules specifying the minutes, hours, days, months and weekdays. Rules for each field are comprised of literal values, wildcards, step values or ranges:* — Wildcard, matches any value (0, 1, 2, ...)0 — Literal, matches only itself (only 0)*/15 — Step, matches any value that is a multiple (0, 15, 30, 45)0-5 — Range, matches any value within the range (0, 1, 2, 3, 4, 5)0-23/2 - Step values can be used in conjunction with ranges (0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22)Each part may have multiple rules, where rules are separated by a comma. The allowed values for each field are as follows:minute — 0-59hour — 0-23days — 1-31month — 1-12 (or aliases, JAN, FEB, MAR, etc.)weekdays — 0-6 (or aliases, SUN, MON, TUE, etc.)The following Cron extensions are supported:@hourly — 0 * * * *@daily (as well as @midnight) — 0 0 * * *@weekly — 0 0 * * 0@monthly — 0 0 1 * *@yearly (as well as @annually) — 0 0 1 1 *@reboot — Run once at boot across the entire clusterSome specific examples that demonstrate the full range of expressions:0 * * * * — The first minute of every hour*/15 9-17 * * * — Every fifteen minutes during standard business hours0 0 * DEC * — Once a day at midnight during december0 7-9,4-6 13 * FRI — Once an hour during both rush hours on Friday the 13thFor more in depth information see the man documentation for cron and crontab in your system. Alternatively you can experiment with various expressions online at Crontab Guru.Caveats &amp; GuidelinesAll schedules are evaluated as UTC unless a different timezone is configured. See Oban.start_link/1 for information about configuring a timezone.Workers can be used for regular and scheduled jobs so long as they accept different arguments.Duplicate jobs are prevented through transactional locks and unique constraints. Workers that are used for regular and scheduled jobs must not specify unique options less than 60s.Long running jobs may execute simultaneously if the scheduling interval is shorter than it takes to execute the job. You can prevent overlap by passing custom unique opts in the crontab config:custom_args = %{scheduled: true} unique_opts = [ period: 60 * 60 * 24, states: [:available, :scheduled, :executing] ] config :my_app, Oban, repo: MyApp.Repo, crontab: [ {&quot;* * * * *&quot;, MyApp.SlowWorker, args: custom_args, unique: unique_opts}, ]Prioritizing JobsNormally, all available jobs within a queue are executed in the order they were scheduled. You can override the normal behavior and prioritize or de-prioritize a job by assigning a numerical priority.Priorities from 0-3 are allowed, where 0 is the highest priority and 3 is the lowest.The default priority is 0, unless specified all jobs have an equally high priority.All jobs with a higher priority will execute before any jobs with a lower priority. Within a particular priority jobs are executed in their scheduled order.TestingOban provides some helpers to facilitate testing. The helpers handle the boilerplate of making assertions on which jobs are enqueued. To use the perform_job/2,3, assert_enqueued/1 and refute_enqueued/1 helpers in your tests you must include them in your testing module and specify your app's Ecto repo:use Oban.Testing, repo: MyApp.RepoNow you can assert, refute or list jobs that have been enqueued within your integration tests:assert_enqueued worker: MyWorker, args: %{id: 1} # or refute_enqueued queue: :special, args: %{id: 2} # or assert [%{args: %{&quot;id&quot; =&gt; 1}}] = all_enqueued worker: MyWorkerYou can also easily unit test workers with the perform_job/2,3 function, which automates validating job args, options, and worker results from a single function call:assert :ok = perform_job(MyWorker, %{id: 1}) # or assert :ok = perform_job(MyWorker, %{id: 1}, attempt: 3, max_attempts: 3) # or assert {:error, _} = perform_job(MyWorker, %{bad: :arg})See the Oban.Testing module for more details.Caveats &amp; GuidelinesAs noted in Usage, there are some guidelines for running tests:Disable all job dispatching by setting queues: false or queues: nil in your test.exs config. Keyword configuration is deep merged, so setting queues: [] won't have any effect.Disable plugins via plugins: false. Default plugins, such as the fixed pruner, aren't necessary in testing mode because jobs created within the sandbox are rolled back at the end of the test. Additionally, the periodic pruning queries will raise DBConnection.OwnershipError when the application boots.Disable cron jobs via crontab: false. Periodic jobs aren't useful while testing and scheduling can lead to random ownership issues.Be sure to use the Ecto Sandbox for testing. Oban makes use of database pubsub events to dispatch jobs, but pubsub events never fire within a transaction. Since sandbox tests run within a transaction no events will fire and jobs won't be dispatched.config :my_app, MyApp.Repo, pool: Ecto.Adapters.SQL.SandboxIntegration TestingDuring integration testing it may be necessary to run jobs because they do work essential for the test to complete, i.e. sending an email, processing media, etc. You can execute all available jobs in a particular queue by calling Oban.drain_queue/1,2 directly from your tests.For example, to process all pending jobs in the &quot;mailer&quot; queue while testing some business logic:defmodule MyApp.BusinessTest do use MyApp.DataCase, async: true alias MyApp.{Business, Worker} test &quot;we stay in the business of doing business&quot; do :ok = Business.schedule_a_meeting(%{email: &quot;monty@brewster.com&quot;}) assert %{success: 1, failure: 0} == Oban.drain_queue(queue: :mailer) # Now, make an assertion about the email delivery end endSee Oban.drain_queue/1,2 for additional details.Error HandlingWhen a job returns an error value, raises an error or exits during execution the details are recorded within the errors array on the job. When the number of execution attempts is below the configured max_attempts limit, the job will automatically be retried in the future.The retry delay has an exponential backoff, meaning the job's second attempt will be after 16s, third after 31s, fourth after 1m 36s, etc.See the Oban.Worker documentation on &quot;Customizing Backoff&quot; for alternative backoff strategies.Error DetailsExecution errors are stored as a formatted exception along with metadata about when the failure occurred and which attempt caused it. Each error is stored with the following keys:at The utc timestamp when the error occurred atattempt The attempt number when the error occurrederror A formatted error message and stacktraceSee the Instrumentation docs for an example of integrating with external error reporting systems.Limiting RetriesBy default, jobs are retried up to 20 times. The number of retries is controlled by the max_attempts value, which can be set at the Worker or Job level. For example, to instruct a worker to discard jobs after three failures:use Oban.Worker, queue: :limited, max_attempts: 3Limiting Execution TimeBy default, individual jobs may execute indefinitely. If this is undesirable you may define a timeout in milliseconds with the timeout/1 callback on your worker module.For example, to limit a worker's execution time to 30 seconds:def MyApp.Worker do use Oban.Worker @impl Oban.Worker def perform(_job) do something_that_may_take_a_long_time() :ok end @impl Oban.Worker def timeout(_job), do: :timer.seconds(30) endThe timeout/1 function accepts an Oban.Job struct, so you can customize the timeout using any job attributes.Define the timeout value through job args:def timeout(%_{args: %{&quot;timeout&quot; =&gt; timeout}}), do: timeoutDefine the timeout based on the number of attempts:def timeout(%_{attempt: attempt}), do: attempt * :timer.seconds(5)Instrumentation and LoggingOban provides integration with Telemetry, a dispatching library for metrics. It is easy to report Oban metrics to any backend by attaching to :oban events.Here is an example of a sample unstructured log handler:defmodule MyApp.ObanLogger do require Logger def handle_event([:oban, :job, :start], measure, meta, _) do Logger.warn(&quot;[Oban] :started \#{meta.worker} at \#{measure.system_time}&quot;) end def handle_event([:oban, :job, event], measure, meta, _) do Logger.warn(&quot;[Oban] \#{event} \#{meta.worker} ran in \#{measure.duration}&quot;) end endAttach the handler to success and failure events in application.ex:events = [[:oban, :job, :start], [:oban, :job, :stop], [:oban, :job, :exception]] :telemetry.attach_many(&quot;oban-logger&quot;, events, &amp;MyApp.ObanLogger.handle_event/4, [])The Oban.Telemetry module provides a robust structured logger that handles all of Oban's telemetry events. As in the example above, attach it within your application.ex module::ok = Oban.Telemetry.attach_default_logger()For more details on the default structured logger and information on event metadata see docs for the Oban.Telemetry module.Reporting ErrorsAnother great use of execution data is error reporting. Here is an example of integrating with Sentry to report job failures:defmodule ErrorReporter do def handle_event([:oban, :job, :exception], measure, meta, _) do extra = meta |&gt; Map.take([:id, :args, :queue, :worker]) |&gt; Map.merge(measure) Sentry.capture_exception(meta.error, stacktrace: meta.stacktrace, extra: extra) end def handle_event([:oban, :circuit, :trip], _measure, meta, _) do Sentry.capture_exception(meta.error, stacktrace: meta.stacktrace, extra: meta) end end :telemetry.attach_many( &quot;oban-errors&quot;, [[:oban, :job, :exception], [:oban, :circuit, :trip]], &amp;ErrorReporter.handle_event/4, %{} )You can use exception events to send error reports to Honeybadger, Rollbar, AppSignal or any other application monitoring platform.IsolationOban supports namespacing through PostgreSQL schemas, also called &quot;prefixes&quot; in Ecto. With prefixes your jobs table can reside outside of your primary schema (usually public) and you can have multiple separate job tables.To use a prefix you first have to specify it within your migration:defmodule MyApp.Repo.Migrations.AddPrefixedObanJobsTable do use Ecto.Migration def up do Oban.Migrations.up(prefix: &quot;private&quot;) end def down do Oban.Migrations.down(prefix: &quot;private&quot;) end endThe migration will create the &quot;private&quot; schema and all tables, functions and triggers within that schema. With the database migrated you'll then specify the prefix in your configuration:config :my_app, Oban, prefix: &quot;private&quot;, repo: MyApp.Repo, queues: [default: 10]Now all jobs are inserted and executed using the private.oban_jobs table. Note that Oban.insert/2,4 will write jobs in the private.oban_jobs table, you'll need to specify a prefix manually if you insert jobs directly through a repo.Supervisor IsolationNot only is the oban_jobs table isolated within the schema, but all notification events are also isolated. That means that insert/update events will only dispatch new jobs for their prefix. You can run multiple Oban instances with different prefixes on the same system and have them entirely isolated, provided you give each supervisor a distinct id.Here we configure our application to start three Oban supervisors using the &quot;public&quot;, &quot;special&quot; and &quot;private&quot; prefixes, respectively:def start(_type, _args) do children = [ Repo, Endpoint, {Oban, name: ObanA, repo: Repo}, {Oban, name: ObanB, repo: Repo, prefix: &quot;special&quot;}, {Oban, name: ObanC, repo: Repo, prefix: &quot;private&quot;} ] Supervisor.start_link(children, strategy: :one_for_one, name: MyApp.Supervisor) endDynamic RepositoriesOban supports Ecto dynamic repositories through the :get_dynamic_repo option. To make this work, you need to run a separate Oban instance per each dynamic repo instance. Most often it's worth bundling each Oban and repo instance under the same supervisor:def start_repo_and_oban(instance_id) do children = [ {MyDynamicRepo, name: nil, url: repo_url(instance_id)}, {Oban, name: instance_id, get_dynamic_repo: fn -&gt; repo_pid(instance_id) end} ] Supervisor.start_link(children, strategy: :one_for_one) endThe function repo_pid/1 must return the pid of the repo for the given instance. You can use Registry to register the repo (for example in the repo's init/2 callback) and discover it.","ref":"Oban.html","title":"Oban","type":"module"},{"doc":"Cancel an available, scheduled or retryable job and mark it as discarded to prevent it from running. If the job is currently executing it will be killed and otherwise it is ignored.If an executing job happens to fail before it can be cancelled the state is set to discarded. However, if it manages to complete successfully then the state will still be completed.ExampleCancel a scheduled job with the id 1:Oban.cancel_job(1) :ok","ref":"Oban.html#cancel_job/2","title":"Oban.cancel_job/2","type":"function"},{"doc":"Check the current state of a queue producer.This allows you to introspect on a queue's health by retrieving key attributes of the producer's state; values such as the current limit, the running job ids, and when the producer was started.Options:queue - a string or atom specifying the queue to check, requiredExampleOban.check_queue(queue: :default) %{ limit: 10, node: &quot;me@local&quot;, none: &quot;a1b2c3d4&quot;, paused: false, queue: &quot;default&quot;, running: [100, 102], started_at: ~D[2020-10-07 15:31:00] }","ref":"Oban.html#check_queue/2","title":"Oban.check_queue/2","type":"function"},{"doc":"Returns a specification to start this module under a supervisor.See Supervisor.","ref":"Oban.html#child_spec/1","title":"Oban.child_spec/1","type":"function"},{"doc":"Retrieve the config struct for a named Oban supervision tree.","ref":"Oban.html#config/1","title":"Oban.config/1","type":"function"},{"doc":"Synchronously execute all available jobs in a queue.All execution happens within the current process and it is guaranteed not to raise an error or exit.Draining a queue from within the current process is especially useful for testing. Jobs that are enqueued by a process when Ecto is in sandbox mode are only visible to that process. Calling drain_queue/2 allows you to control when the jobs are executed and to wait synchronously for all jobs to complete.Failures &amp; RetriesDraining a queue uses the same execution mechanism as regular job dispatch. That means that any job failures or crashes are captured and result in a retry. Retries are scheduled in the future with backoff and won't be retried immediately.By default jobs are executed in safe mode, just as they are in production. Safe mode catches any errors or exits and records the formatted error in the job's errors array. That means exceptions and crashes are not bubbled up to the calling process.If you expect jobs to fail, would like to track failures, or need to check for specific errors you can pass the with_safety: false flag.Scheduled JobsBy default, drain_queue/2 will execute all currently available jobs. In order to execute scheduled jobs, you may pass the :with_scheduled flag which will cause scheduled jobs to be marked as available beforehand.Options:queue - a string or atom specifying the queue to drain, required:with_scheduled — whether to include any scheduled jobs when draining, default false:with_safety — whether to silently catch errors when draining, default trueExampleDrain a queue with three available jobs, two of which succeed and one of which fails:Oban.drain_queue(queue: :default) %{success: 2, failure: 1}Drain a queue including any scheduled jobs:Oban.drain_queue(queue: :default, with_scheduled: true) %{success: 1, failure: 0}Drain a queue and assert an error is raised:assert_raise RuntimeError, fn -&gt; Oban.drain_queue(queue: :risky, with_safety: false) end","ref":"Oban.html#drain_queue/2","title":"Oban.drain_queue/2","type":"function"},{"doc":"Insert a new job into the database for execution.This and the other insert variants are the recommended way to enqueue jobs because they support features like unique jobs.See the section on &quot;Unique Jobs&quot; for more details.ExampleInsert a single job:{:ok, job} = Oban.insert(MyApp.Worker.new(%{id: 1}))Insert a job while ensuring that it is unique within the past 30 seconds:{:ok, job} = Oban.insert(MyApp.Worker.new(%{id: 1}, unique: [period: 30]))","ref":"Oban.html#insert/2","title":"Oban.insert/2","type":"function"},{"doc":"Put a job insert operation into an Ecto.Multi.Like insert/2, this variant is recommended over Ecto.Multi.insert because it supports all of Oban's features, i.e. unique jobs.See the section on &quot;Unique Jobs&quot; for more details.ExampleEcto.Multi.new() |&gt; Oban.insert(&quot;job-1&quot;, MyApp.Worker.new(%{id: 1})) |&gt; Oban.insert(&quot;job-2&quot;, fn _ -&gt; MyApp.Worker.new(%{id: 2}) end) |&gt; MyApp.Repo.transaction()","ref":"Oban.html#insert/4","title":"Oban.insert/4","type":"function"},{"doc":"Similar to insert/2, but raises an Ecto.InvalidChangesetError if the job can't be inserted.Examplejob = Oban.insert!(MyApp.Worker.new(%{id: 1}))","ref":"Oban.html#insert!/2","title":"Oban.insert!/2","type":"function"},{"doc":"Insert multiple jobs into the database for execution.Insertion respects prefix and log settings, but it does not use per-job unique configuration. You must use insert/2,4 or insert!/2 for per-job unique support.There are a few important differences between this function and Ecto.Repo.insert_all/3:This function always returns a list rather than a tuple of {count, records}This function requires a list of changesets rather than a list of maps or keyword listsExample1..100 |&gt; Enum.map(&amp;MyApp.Worker.new(%{id: &amp;1})) |&gt; Oban.insert_all()","ref":"Oban.html#insert_all/2","title":"Oban.insert_all/2","type":"function"},{"doc":"Put an insert_all operation into an Ecto.Multi.This function supports the same features and has the same caveats as insert_all/2.Examplechangesets = Enum.map(0..100, MyApp.Worker.new(%{id: &amp;1})) Ecto.Multi.new() |&gt; Oban.insert_all(:jobs, changesets) |&gt; MyApp.Repo.transaction()","ref":"Oban.html#insert_all/4","title":"Oban.insert_all/4","type":"function"},{"doc":"Pause a running queue, preventing it from executing any new jobs. All running jobs will remain running until they are finished.When shutdown begins all queues are paused.Options:queue - a string or atom specifying the queue to pause, required:local_only - whether the queue will be paused only on the local node, default: falseExamplePause the default queue:Oban.pause_queue(queue: :default) :okPause the default queue, but only on the local node:Oban.pause_queue(queue: :default, local_only: true) :ok","ref":"Oban.html#pause_queue/2","title":"Oban.pause_queue/2","type":"function"},{"doc":"Resume executing jobs in a paused queue.Options:queue - a string or atom specifying the queue to resume, required:local_only - whether the queue will be resumed only on the local node, default: falseExampleResume a paused default queue:Oban.resume_queue(:default) :okResume the default queue, but only on the local node:Oban.resume_queue(queue: :default, local_only: true) :ok","ref":"Oban.html#resume_queue/2","title":"Oban.resume_queue/2","type":"function"},{"doc":"Sets a job as available, adding attempts if already maxed out. If the job is currently available, executing or scheduled it will be ignored. The job is scheduled for immediate execution.ExampleRetry a discarded job with the id 1:Oban.retry_job(1) :ok","ref":"Oban.html#retry_job/2","title":"Oban.retry_job/2","type":"function"},{"doc":"Scale the concurrency for a queue.Options:queue - a string or atom specifying the queue to scale, required:limit — the new concurrency limit:local_only — whether the queue will be scaled only on the local node, default: falseExampleScale a queue up, triggering immediate execution of queued jobs:Oban.scale_queue(queue: :default, limit: 50) :okScale the queue back down, allowing executing jobs to finish:Oban.scale_queue(queue: :default, limit: 5) :okScale the queue only on the local node:Oban.scale_queue(queue: :default, limit: 10, local_only: true) :ok","ref":"Oban.html#scale_queue/2","title":"Oban.scale_queue/2","type":"function"},{"doc":"Starts an Oban supervision tree linked to the current process.OptionsThese options are required; without them the supervisor won't start:name — used for supervisor registration, defaults to Oban:repo — specifies the Ecto repo used to insert and retrieve jobsPrimary OptionsThese options determine what the system does at a high level, i.e. which queues to run.:crontab — a list of cron expressions that enqueue jobs on a periodic basis. See &quot;Periodic (CRON) Jobs&quot; in the module docs.For testing purposes :crontab may be set to false or nil, which disables scheduling.:node — used to identify the node that the supervision tree is running in. If no value is provided it will use the node name in a distributed system, or the hostname in an isolated node. See &quot;Node Name&quot; below.:prefix — the query prefix, or schema, to use for inserting and executing jobs. An oban_jobs table must exist within the prefix. See the &quot;Prefix Support&quot; section in the module documentation for more details.:queues — a keyword list where the keys are queue names and the values are the concurrency setting or a keyword list of queue options. For example, setting queues to [default: 10, exports: 5] would start the queues default and exports with a combined concurrency level of 15. The concurrency setting specifies how many jobs each queue will run concurrently.Queues accept additional override options to customize their behavior, e.g. by setting the poll_interval and the dispatch_cooldown for a specific queue. For testing purposes :queues may be set to false or nil, which effectively disables all job dispatching.:timezone — which timezone to use when scheduling cron jobs. To use a timezone other than the default of &quot;Etc/UTC&quot; you must have a timezone database like tzdata installed and configured.:log — either false to disable logging or a standard log level (:error, :warn, :info, :debug). This determines whether queries are logged or not; overriding the repo's configured log level. Defaults to false, where no queries are logged.Twiddly OptionsAdditional options used to tune system behaviour. These are primarily useful for testing or troubleshooting and don't usually need modification.:circuit_backoff — the number of milliseconds until queries are attempted after a database error. All processes communicating with the database are equipped with circuit breakers and will use this for the backoff. Defaults to 30_000ms.:dispatch_cooldown — the minimum number of milliseconds a producer will wait before fetching and running more jobs. A slight cooldown period prevents a producer from flooding with messages and thrashing the database. The cooldown period directly impacts a producer's throughput: jobs per second for a single queue is calculated by (1000 / cooldown) * limit. For example, with a 5ms cooldown and a queue limit of 25 a single queue can run 2,500 jobs/sec.The default is 5ms and the minimum is 1ms, which is likely faster than the database can return new jobs to run.:poll_interval - the number of milliseconds between polling for new jobs in a queue. This is directly tied to the resolution of scheduled jobs. For example, with a poll_interval of 5_000ms, scheduled jobs are checked every 5 seconds. The default is 1_000ms.:shutdown_grace_period - the amount of time a queue will wait for executing jobs to complete before hard shutdown, specified in milliseconds. The default is 15_000, or 15 seconds.ExampleTo start an Oban supervisor within an application's supervision tree:def start(_type, _args) do children = [MyApp.Repo, {Oban, queues: [default: 50]}] Supervisor.start_link(children, strategy: :one_for_one, name: MyApp.Supervisor) endNode NameWhen the node value hasn't been configured it is generated based on the environment:In a distributed system the node name is usedIn a Heroku environment the system environment's DYNO value is usedOtherwise, the system hostname is used","ref":"Oban.html#start_link/1","title":"Oban.start_link/1","type":"function"},{"doc":"Start a new supervised queue.By default this starts a new supervised queue across all nodes running Oban on the same database and prefix. You can pass the option local_only: true if you prefer to start the queue only on the local node.Options:queue - a string or atom specifying the queue to start, required:limit - set the concurrency limit, required:local_only - whether the queue will be started only on the local node, default: falseExampleStart the :priority queue with a concurrency limit of 10 across the connected nodes.Oban.start_queue(queue: :priority, limit: 10) :okStart the :media queue with a concurrency limit of 5 only on the local node.Oban.start_queue(queue: :media, limit: 5, local_only: true) :ok","ref":"Oban.html#start_queue/2","title":"Oban.start_queue/2","type":"function"},{"doc":"Shutdown a queue's supervision tree and stop running jobs for that queue.By default this action will occur across all the running nodes. Still, if you prefer to stop the queue's supervision tree and stop running jobs for that queue only on the local node, you can pass the option: local_only: trueThe shutdown process pauses the queue first and allows current jobs to exit gracefully, provided they finish within the shutdown limit.Options:queue - a string or atom specifying the queue to stop, required:local_only - whether the queue will be stopped only on the local node, default: falseExampleOban.stop_queue(queue: :default) :ok Oban.stop_queue(queue: :media, local_only: true) :ok","ref":"Oban.html#stop_queue/2","title":"Oban.stop_queue/2","type":"function"},{"doc":"Returns the pid of the root oban process for the given name.ExampleFind the default instance:Oban.whereis(Oban)Find a dynamically named instance:Oban.whereis({:oban, 1})","ref":"Oban.html#whereis/1","title":"Oban.whereis/1","type":"function"},{"doc":"","ref":"Oban.html#t:job_changeset/0","title":"Oban.job_changeset/0","type":"type"},{"doc":"","ref":"Oban.html#t:name/0","title":"Oban.name/0","type":"type"},{"doc":"","ref":"Oban.html#t:option/0","title":"Oban.option/0","type":"type"},{"doc":"","ref":"Oban.html#t:queue_name/0","title":"Oban.queue_name/0","type":"type"},{"doc":"","ref":"Oban.html#t:queue_option/0","title":"Oban.queue_option/0","type":"type"},{"doc":"","ref":"Oban.html#t:queue_state/0","title":"Oban.queue_state/0","type":"type"},{"doc":"A Job is an Ecto schema used for asynchronous execution.Job changesets are created by your application code and inserted into the database for asynchronous execution. Jobs can be inserted along with other application data as part of a transaction, which guarantees that jobs will only be triggered from a successful transaction.","ref":"Oban.Job.html","title":"Oban.Job","type":"module"},{"doc":"Construct a new job changeset ready for insertion into the database.Options:max_attempts — the maximum number of times a job can be retried if there are errors during execution:meta — a map containing additional information about the job:priority — a numerical indicator from 0 to 3 of how important this job is relative to other jobs in the same queue. The lower the number, the higher priority the job.:queue — a named queue to push the job into. Jobs may be pushed into any queue, regardless of whether jobs are currently being processed for the queue.:schedule_in - the number of seconds until the job should be executed:replace_args - if the arguments should be replaced on a unique conflict:scheduled_at - a time in the future after which the job should be executed:tags — a list of tags to group and organize related jobs, i.e. to identify scheduled jobs:unique — a keyword list of options specifying how uniqueness will be calculated. The options define which fields will be used, for how long, with which keys, and for which states.:worker — a module to execute the job in. The module must implement the Oban.Worker behaviour.ExamplesInsert a job with the :default queue:%{id: 1, user_id: 2} |&gt; Oban.Job.new(queue: :default, worker: MyApp.Worker) |&gt; Oban.insert()Generate a pre-configured job for MyApp.Worker and push it:%{id: 1, user_id: 2} |&gt; MyApp.Worker.new() |&gt; Oban.insert()Schedule a job to run in 5 seconds:%{id: 1} |&gt; MyApp.Worker.new(schedule_in: 5) |&gt; Oban.insert()Insert a job, ensuring that it is unique within the past minute:%{id: 1} |&gt; MyApp.Worker.new(unique: [period: 60]) |&gt; Oban.insert()Insert a unique job based only on the worker field, and within multiple states:fields = [:worker] states = [:available, :scheduled, :executing, :retryable, :completed] %{id: 1} |&gt; MyApp.Worker.new(unique: [fields: fields, period: 60, states: states]) |&gt; Oban.insert()Insert a unique job considering only the worker and specified keys in the args:keys = [:account_id, :url] %{account_id: 1, url: &quot;https://example.com&quot;} |&gt; MyApp.Worker.new(unique: [fields: [:args, :worker], keys: keys]) |&gt; Oban.insert()","ref":"Oban.Job.html#new/2","title":"Oban.Job.new/2","type":"function"},{"doc":"A canonical list of all possible job states.This may be used to build up :unique options without duplicating states in application code.Examplesiex&gt; Oban.Job.states() -- [:completed, :discarded] [:scheduled, :available, :executing, :retryable, :cancelled]","ref":"Oban.Job.html#states/0","title":"Oban.Job.states/0","type":"function"},{"doc":"Convert a Job changeset into a map suitable for database insertion.ExamplesConvert a worker generated changeset into a plain map:%{id: 123} |&gt; MyApp.Worker.new() |&gt; Oban.Job.to_map()","ref":"Oban.Job.html#to_map/1","title":"Oban.Job.to_map/1","type":"function"},{"doc":"","ref":"Oban.Job.html#t:args/0","title":"Oban.Job.args/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:errors/0","title":"Oban.Job.errors/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:option/0","title":"Oban.Job.option/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:t/0","title":"Oban.Job.t/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:tags/0","title":"Oban.Job.tags/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:unique_field/0","title":"Oban.Job.unique_field/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:unique_option/0","title":"Oban.Job.unique_option/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:unique_period/0","title":"Oban.Job.unique_period/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:unique_state/0","title":"Oban.Job.unique_state/0","type":"type"},{"doc":"The Notifier coordinates listening for and publishing notifications for events in predefined channels.Every Oban supervision tree contains a notifier process, registered as Oban.Notifier, which itself maintains a single connection with an app's database. All incoming notifications are relayed through that connection to other processes.ChannelsThe notifier recognizes three predefined channels, each with a distinct responsibility:gossip — arbitrary communication between nodes or jobs are sent on the gossip channelinsert — as jobs are inserted into the database an event is published on the insert channel. Processes such as queue producers use this as a signal to dispatch new jobs.signal — instructions to take action, such as scale a queue or kill a running job, are sent through the signal channel.The insert and signal channels are primarily for internal use. Use the gossip channel to send notifications between jobs or processes in your application.CaveatsThe notifications system is built on PostgreSQL's LISTEN/NOTIFY functionality. Notifications are only delivered after a transaction completes and are de-duplicated before publishing.Most applications run Ecto in sandbox mode while testing. Sandbox mode wraps each test in a separate transaction which is rolled back after the test completes. That means the transaction is never committed, which prevents delivering any notifications.To test using notifications you must run Ecto without sandbox mode enabled.ExamplesBroadcasting after a job is completed:defmodule MyApp.Worker do use Oban.Worker @impl Oban.Worker def perform(job) do :ok = MyApp.do_work(job.args) Oban.Notifier.notify(Oban.config(), :gossip, %{complete: job.id}) :ok end endListening for job complete events from another process:def insert_and_listen(args) do {:ok, job} = args |&gt; MyApp.Worker.new() |&gt; Oban.insert() receive do {:notification, :gossip, %{&quot;complete&quot; =&gt; ^job.id}} -&gt; IO.puts(&quot;Other job complete!&quot;) after 30_000 -&gt; IO.puts(&quot;Other job didn&#39;t finish in 30 seconds!&quot;) end end","ref":"Oban.Notifier.html","title":"Oban.Notifier","type":"module"},{"doc":"Returns a specification to start this module under a supervisor.See Supervisor.","ref":"Oban.Notifier.html#child_spec/1","title":"Oban.Notifier.child_spec/1","type":"function"},{"doc":"Register the current process to receive relayed messages for the provided channels.All messages are received as JSON and decoded before they are relayed to registered processes. Each registered process receives a three element notification tuple in the following format:{:notification, channel :: channel(), decoded :: map()}ExampleRegister to listen for all :gossip channel messages:Oban.Notifier.listen([:gossip])Listen for messages on all channels:Oban.Notifier.listen([:gossip, :insert, :signal])","ref":"Oban.Notifier.html#listen/2","title":"Oban.Notifier.listen/2","type":"function"},{"doc":"Broadcast a notification to listeners on all nodes.Notifications are scoped to the configured prefix. For example, if there are instances running with the public and private prefixes, a notification published in the public prefix won't be picked up by processes listening with the private prefix.ExampleBroadcast a gossip message:Oban.Notifier.notify(Oban.config(), :gossip, %{message: &quot;hi!&quot;})","ref":"Oban.Notifier.html#notify/3","title":"Oban.Notifier.notify/3","type":"function"},{"doc":"","ref":"Oban.Notifier.html#t:channel/0","title":"Oban.Notifier.channel/0","type":"type"},{"doc":"","ref":"Oban.Notifier.html#t:option/0","title":"Oban.Notifier.option/0","type":"type"},{"doc":"Wrappers around Ecto.Repo callbacks.These functions should be used when working with an Ecto repo inside a plugin. These functions will resolve the correct repo instance, and set the schema prefix and the log level, according to the Oban configuration.","ref":"Oban.Repo.html","title":"Oban.Repo","type":"module"},{"doc":"Wraps Ecto.Repo.all/2.","ref":"Oban.Repo.html#all/3","title":"Oban.Repo.all/3","type":"function"},{"doc":"Wraps Ecto.Repo.checkout/2.","ref":"Oban.Repo.html#checkout/3","title":"Oban.Repo.checkout/3","type":"function"},{"doc":"Wraps Ecto.Repo.config/0.","ref":"Oban.Repo.html#config/1","title":"Oban.Repo.config/1","type":"function"},{"doc":"Wraps Ecto.Repo.delete_all/2.","ref":"Oban.Repo.html#delete_all/3","title":"Oban.Repo.delete_all/3","type":"function"},{"doc":"Wraps Ecto.Repo.insert/2.","ref":"Oban.Repo.html#insert/3","title":"Oban.Repo.insert/3","type":"function"},{"doc":"Wraps Ecto.Repo.insert_all/3.","ref":"Oban.Repo.html#insert_all/4","title":"Oban.Repo.insert_all/4","type":"function"},{"doc":"Wraps Ecto.Repo.one/2.","ref":"Oban.Repo.html#one/3","title":"Oban.Repo.one/3","type":"function"},{"doc":"Wraps Ecto.Adapters.SQL.Repo.query/4.","ref":"Oban.Repo.html#query/4","title":"Oban.Repo.query/4","type":"function"},{"doc":"Wraps Ecto.Adapters.SQL.Repo.to_sql/2.","ref":"Oban.Repo.html#to_sql/3","title":"Oban.Repo.to_sql/3","type":"function"},{"doc":"Wraps Ecto.Repo.transaction/2.","ref":"Oban.Repo.html#transaction/3","title":"Oban.Repo.transaction/3","type":"function"},{"doc":"Wraps Ecto.Repo.update/2.","ref":"Oban.Repo.html#update/3","title":"Oban.Repo.update/3","type":"function"},{"doc":"Wraps Ecto.Repo.update_all/3.","ref":"Oban.Repo.html#update_all/4","title":"Oban.Repo.update_all/4","type":"function"},{"doc":"","ref":"Oban.Repo.html#t:config/0","title":"Oban.Repo.config/0","type":"type"},{"doc":"Telemetry integration for event metrics, logging and error reporting.Job EventsOban emits the following telemetry events for each job:[:oban, :job, :start] — at the point a job is fetched from the database and will execute[:oban, :job, :stop] — after a job succeeds and the success is recorded in the database[:oban, :job, :exception] — after a job fails and the failure is recorded in the databaseAll job events share the same details about the job that was executed. In addition, failed jobs provide the error type, the error itself, and the stacktrace. The following chart shows which metadata you can expect for each event:eventmeasuresmetadata:start:system_time:job, :prefix:stop:duration, :queue_time:job, :prefix:exception:duration, :queue_time:job, :prefix, :kind, :error, :stacktraceFor :exception events the metadata includes details about what caused the failure. The :kind value is determined by how an error occurred. Here are the possible kinds::error — from an {:error, error} return value. Some Erlang functions may also throw an :error tuple, which will be reported as :error.:exit — from a caught process exit:throw — from a caught value, this doesn't necessarily mean that an error occurred and the error value is unpredictableProducer EventsOban emits the following telemetry span events for each queue's producer:[:oban, :producer, :start | :stop | :exception] — when a producer deschedules or dispatches new jobseventmeasuresmetadata:start:system_time:action, :queue:stop:duration:action, :queue:exception:duration:action, :queueMetadata:action — one of :deschedule or :dispatch:queue — the name of the queue as a string, e.g. &quot;default&quot; or &quot;mailers&quot;Circuit EventsAll processes that interact with the database have circuit breakers to prevent errors from crashing the entire supervision tree. Processes emit a [:oban, :circuit, :trip] event when a circuit is tripped and [:oban, :circuit, :open] when the breaker is subsequently opened again.eventmeasuresmetadata[:oban, :circuit, :trip]:error, :message, :name, :stacktrace[:oban, :circuit, :open]:nameMetadata:error — the error that tripped the circuit, see the error kinds breakdown above:name — the registered name of the process that tripped a circuit, i.e. Oban.Notifier:message — a formatted error message describing what went wrong:stacktrace — exception stacktrace, when availableDefault LoggerA default log handler that emits structured JSON is provided, see attach_default_logger/0 for usage. Otherwise, if you would prefer more control over logging or would like to instrument events you can write your own handler.Here is an example of the JSON output for the job:stop event:{ &quot;args&quot;:{&quot;action&quot;:&quot;OK&quot;,&quot;ref&quot;:1}, &quot;duration&quot;:4327295, &quot;event&quot;:&quot;job:stop&quot;, &quot;queue&quot;:&quot;alpha&quot;, &quot;queue_time&quot;:3127905, &quot;source&quot;:&quot;oban&quot;, &quot;worker&quot;:&quot;Oban.Integration.Worker&quot; }All timing measurements are recorded as native time units but logged in microseconds.ExamplesA handler that only logs a few details about failed jobs:defmodule MicroLogger do require Logger def handle_event([:oban, :job, :exception], %{duration: duration}, meta, nil) do Logger.warn(&quot;[\#{meta.queue}] \#{meta.worker} failed in \#{duration}&quot;) end end :telemetry.attach(&quot;oban-logger&quot;, [:oban, :job, :exception], &amp;MicroLogger.handle_event/4, nil)Another great use of execution data is error reporting. Here is an example of integrating with Honeybadger, but only reporting jobs that have failed 3 times or more:defmodule ErrorReporter do def handle_event([:oban, :job, :exception], _, %{attempt: attempt} = meta, _) do if attempt &gt;= 3 do context = Map.take(meta, [:id, :args, :queue, :worker]) Honeybadger.notify(meta.error, context, meta.stacktrace) end end end :telemetry.attach(&quot;oban-errors&quot;, [:oban, :job, :exception], &amp;ErrorReporter.handle_event/4, [])","ref":"Oban.Telemetry.html","title":"Oban.Telemetry","type":"module"},{"doc":"Attaches a default structured JSON Telemetry handler for logging.This function attaches a handler that outputs logs with the following fields:args — a map of the job's raw argumentsduration — the job's runtime duration, in the native time unitevent — either :success or :failure depending on whether the job succeeded or erroredqueue — the job's queuesource — always &quot;oban&quot;system_time — when the job started, in microsecondsworker — the job's worker moduleExamplesAttach a logger at the default :info level::ok = Oban.Telemetry.attach_default_logger()Attach a logger at the :debug level::ok = Oban.Telemetry.attach_default_logger(:debug)","ref":"Oban.Telemetry.html#attach_default_logger/1","title":"Oban.Telemetry.attach_default_logger/1","type":"function"},{"doc":"Measure and report :start, :stop and :exception events for a function.ExamplesEmit span timing events for a prune function::ok = Oban.Telemetry.span(:prune, &amp;MyApp.Pruner.prune/0, %{extra: :data})That will emit the following events:[:oban, :prune, :start] — before the function is invoked[:oban, :prune, :stop] — when the function completes successfully[:oban, :prune, :exception] — reported if the function throws, crashes or raises an error","ref":"Oban.Telemetry.html#span/3","title":"Oban.Telemetry.span/3","type":"function"},{"doc":"This module simplifies making assertions about enqueued jobs during testing.Assertions may be made on any property of a job, but you'll typically want to check by args, queue or worker. If you're using namespacing through PostgreSQL schemas, also called &quot;prefixes&quot; in Ecto, you should use the prefix option when doing assertions about enqueued jobs during testing. By default the prefix option is public.Using in TestsThe most convenient way to use Oban.Testing is to use the module:use Oban.Testing, repo: MyApp.RepoThat will define three helper functions, assert_enqueued/1,2, refute_enqueued/1,2 and all_enqueued/1. The functions can then be used to make assertions on the jobs that have been inserted in the database while testing.Some small examples:# Assert that a job was already enqueued assert_enqueued worker: MyWorker, args: %{id: 1} # Assert that a job was enqueued or will be enqueued in the next 100ms assert_enqueued [worker: MyWorker, args: %{id: 1}], 100 # Refute that a job was already enqueued refute_enqueued queue: &quot;special&quot;, args: %{id: 2} # Refute that a job was already enqueued or would be enqueued in the next 100ms refute_enqueued queue: &quot;special&quot;, args: %{id: 2}, 100 # Make assertions on a list of all jobs matching some options assert [%{args: %{&quot;id&quot; =&gt; 1}}] = all_enqueued(worker: MyWorker)Note that the final example, using all_enqueued/1, returns a raw list of matching jobs and does not make an assertion by itself. This makes it possible to test using pattern matching at the expense of being more verbose.ExampleGiven a simple module that enqueues a job:defmodule MyApp.Business do def work(args) do args |&gt; Oban.Job.new(worker: MyApp.Worker, queue: :special) |&gt; Oban.insert!() end endThe behaviour can be exercised in your test code:defmodule MyApp.BusinessTest do use ExUnit.Case, async: true use Oban.Testing, repo: MyApp.Repo alias MyApp.Business test &quot;jobs are enqueued with provided arguments&quot; do Business.work(%{id: 1, message: &quot;Hello!&quot;}) assert_enqueued worker: MyApp.Worker, args: %{id: 1, message: &quot;Hello!&quot;} end endMatching Scheduled Jobs and TimestampsIn order to assert a job has been scheduled at a certain time, you will need to match against the scheduled_at attribute of the enqueued job.in_an_hour = DateTime.add(DateTime.utc_now(), 3600, :second) assert_enqueued worker: MyApp.Worker, scheduled_at: in_an_hourBy default, Oban will apply a 1 second delta to all timestamp fields of jobs, so that small deviations between the actual value and the expected one are ignored. You may configure this delta by passing a tuple of value and a delta option (in seconds) to corresponding keyword:assert_enqueued worker: MyApp.Worker, scheduled_at: {in_an_hour, delta: 10}Adding to Case TemplatesTo include helpers in all of your tests you can add it to your case template:defmodule MyApp.DataCase do use ExUnit.CaseTemplate using do quote do use Oban.Testing, repo: MyApp.Repo import Ecto import Ecto.Changeset import Ecto.Query import MyApp.DataCase alias MyApp.Repo end end end","ref":"Oban.Testing.html","title":"Oban.Testing","type":"module"},{"doc":"Retrieve all currently enqueued jobs matching a set of options.Only jobs matching all of the provided arguments will be returned. Additionally, jobs are returned in descending order where the most recently enqueued job will be listed first.ExamplesAssert based on only some of a job's args:assert [%{args: %{&quot;id&quot; =&gt; 1}}] = all_enqueued(worker: MyWorker)Assert that exactly one job was inserted for a queue:assert [%Oban.Job{}] = all_enqueued(queue: :alpha)","ref":"Oban.Testing.html#all_enqueued/2","title":"Oban.Testing.all_enqueued/2","type":"function"},{"doc":"Assert that a job with particular options has been enqueued.Only values for the provided arguments will be checked. For example, an assertion made on worker: &quot;MyWorker&quot; will match any jobs for that worker, regardless of the queue or args.","ref":"Oban.Testing.html#assert_enqueued/2","title":"Oban.Testing.assert_enqueued/2","type":"function"},{"doc":"Assert that a job with particular options is or will be enqueued within a timeout period.See assert_enqueued/2 for additional details.ExamplesAssert that a job will be enqueued in the next 100ms:assert_enqueued [worker: MyWorker], 100","ref":"Oban.Testing.html#assert_enqueued/3","title":"Oban.Testing.assert_enqueued/3","type":"function"},{"doc":"Construct a job and execute it with a worker module.This reduces boiler plate when constructing jobs for unit tests and checks for common pitfalls. For example, it automatically converts args to string keys before calling perform/1, ensuring that perform clauses aren't erroneously trying to match atom keys.The helper makes the following assertions:That the worker implements the Oban.Worker behaviourThat the options provided build a valid jobThat the return is valid, e.g. :ok, {:ok, value}, {:error, value} etc.If all of the assertions pass then the function returns the result of perform/1 for you to make additional assertions on.ExamplesSuccessfully execute a job with some string arguments:assert :ok = perform_job(MyWorker, %{&quot;id&quot; =&gt; 1})Successfully execute a job and assert that it returns an error tuple:assert {:error, _} = perform_job(MyWorker, %{&quot;bad&quot; =&gt; &quot;arg&quot;})Execute a job with the args keys automatically stringified:assert :ok = perform_job(MyWorker, %{id: 1})Exercise custom attempt handling within a worker by passing options:assert :ok = perform_job(MyWorker, %{}, attempt: 42)Cause a test failure because the provided worker isn't real:assert :ok = perform_job(Vorker, %{&quot;id&quot; =&gt; 1})","ref":"Oban.Testing.html#perform_job/3","title":"Oban.Testing.perform_job/3","type":"function"},{"doc":"Refute that a job with particular options has been enqueued.See assert_enqueued/2 for additional details.","ref":"Oban.Testing.html#refute_enqueued/2","title":"Oban.Testing.refute_enqueued/2","type":"function"},{"doc":"Refute that a job with particular options is or will be enqueued within a timeout period.The minimum refute timeout is 10ms.See assert_enqueued/2 for additional details.ExamplesRefute that a job will not be enqueued in the next 100ms:refute_enqueued [worker: MyWorker], 100","ref":"Oban.Testing.html#refute_enqueued/3","title":"Oban.Testing.refute_enqueued/3","type":"function"},{"doc":"Defines a behavior and macro to guide the creation of worker modules.Worker modules do the work of processing a job. At a minimum they must define a perform/1 function, which is called with the full Oban.Job struct.Defining WorkersWorker modules are defined by using Oban.Worker. A bare use Oban.Worker invocation sets a worker with these defaults::max_attempts — 20:priority — 0:queue — :default:unique — no uniqueness setThe following example defines a worker module to process jobs in the events queue. It then dials down the priority from 0 to 1, limits retrying on failures to 10, adds a &quot;business&quot; tag, and ensures that duplicate jobs aren't enqueued within a 30 second period:defmodule MyApp.Workers.Business do use Oban.Worker, queue: :events, priority: 1, max_attempts: 10, tags: [&quot;business&quot;], unique: [period: 30] @impl Oban.Worker def perform(%Oban.Job{attempt: attempt}) when attempt &gt; 3 do IO.inspect(attempt) end def perform(job) do IO.inspect(job.args) end endThe perform/1 function receives an Oban.Job struct as argument. This allows workers to change the behavior of perform/1 based on attributes of the job, e.g. the number of execution attempts or when it was inserted.The value returned from perform/1 can control whether the job is a success or a failure::ok or {:ok, value} — the job is successful; for success tuples the value is ignored:discard or {:discard, reason} — discard the job and prevent it from being retried again. An error is recorded using the optional reason, though the job is still successful{:error, error} — the job failed, record the error and schedule a retry if possible{:snooze, seconds} — mark the job as snoozed and schedule it to run again seconds in the future. Snoozing a job does not change the number of retries remaining on the job.In addition to explicit return values, any unhandled exception, exit or throw will fail the job and schedule a retry if possible.As an example of error tuple handling, this worker will return an error tuple when the value is less than one:defmodule MyApp.Workers.ErrorExample do use Oban.Worker @impl Worker def perform(%{args: %{&quot;value&quot; =&gt; value}}) do if value &gt; 1 do :ok else {:error, &quot;invalid value given: &quot; &lt;&gt; inspect(value)} end end endThe error tuple is wrapped in an Oban.PerformError with a formatted message. The error tuple itself is available through the exception's :reason field.Enqueuing JobsAll workers implement a new/2 function that converts an args map into a job changeset suitable for inserting into the database for later execution:%{in_the: &quot;business&quot;, of_doing: &quot;business&quot;} |&gt; MyApp.Workers.Business.new() |&gt; Oban.insert()The worker's defaults may be overridden by passing options:%{vote_for: &quot;none of the above&quot;} |&gt; MyApp.Workers.Business.new(queue: &quot;special&quot;, max_attempts: 5) |&gt; Oban.insert()Uniqueness options may also be overridden by passing options:%{expensive: &quot;business&quot;} |&gt; MyApp.Workers.Business.new(unique: [period: 120, fields: [:worker]]) |&gt; Oban.insert()Note that unique options aren't merged, they are overridden entirely.See Oban.Job for all available options.Customizing BackoffWhen jobs fail they may be retried again in the future using a backoff algorithm. By default the backoff is exponential with a fixed padding of 15 seconds and a small amount of jitter. The jitter helps to prevent jobs that fail simultaneously from consistently retrying at the same time. The default backoff is clamped to a maximum of 24 days, the equivalent of the 20th attempt.If the default strategy is too aggressive or otherwise unsuited to your app's workload you can define a custom backoff function using the backoff/1 callback.The following worker defines a backoff/1 function that delays retries using a variant of the historic Resque/Sidekiq algorithm:defmodule MyApp.SidekiqBackoffWorker do use Oban.Worker @impl Worker def backoff(%Job{attempt: attempt}) do trunc(:math.pow(attempt, 4) + 15 + :rand.uniform(30) * attempt) end @impl Worker def perform(_job) do :do_business end endHere are some alternative backoff strategies to consider:constant — delay by a fixed number of seconds, e.g. 1→15, 2→15, 3→15linear — delay for the same number of seconds as the current attempt, e.g. 1→1, 2→2, 3→3squared — delay by attempt number squared, e.g. 1→1, 2→4, 3→9sidekiq — delay by a base amount plus some jitter, e.g. 1→32, 2→61, 3→135Contextual BackoffAny error, catch or throw is temporarily recorded in the job's unsaved_error map. The unsaved error map can be used by backoff/1 to calculate a custom backoff based on the exact error that failed the job. In this example the backoff/1 callback checks to see if the error was due to rate limiting and adjusts the backoff accordingly:defmodule MyApp.ApiWorker do use Oban.Worker @five_minutes 5 * 60 @impl Worker def perform(%{args: args}) do MyApp.make_external_api_call(args) end @impl Worker def backoff(%Job{attempt: attempt, unsaved_error: unsaved_error}) do %{kind: _, reason: reason, stacktrace: _} = unsaved_error case reason do %MyApp.ApiError{status: 429} -&gt; @five_minutes _ -&gt; trunc(:math.pow(attempt, 4)) end end endLimiting Execution TimeBy default, individual jobs may execute indefinitely. If this is undesirable you may define a timeout in milliseconds with the timeout/1 callback on your worker module.For example, to limit a worker's execution time to 30 seconds:def MyApp.Worker do use Oban.Worker @impl Oban.Worker def perform(_job) do something_that_may_take_a_long_time() :ok end @impl Oban.Worker def timeout(_job), do: :timer.seconds(30) endThe timeout/1 function accepts an Oban.Job struct, so you can customize the timeout using any job attributes.Define the timeout value through job args:def timeout(%_{args: %{&quot;timeout&quot; =&gt; timeout}}), do: timeoutDefine the timeout based on the number of attempts:def timeout(%_{attempt: attempt}), do: attempt * :timer.seconds(5)","ref":"Oban.Worker.html","title":"Oban.Worker","type":"behaviour"},{"doc":"Calculate the execution backoff.In this context backoff specifies the number of seconds to wait before retrying a failed job.Defaults to an exponential algorithm with a minimum delay of 15 seconds and a small amount of jitter.","ref":"Oban.Worker.html#c:backoff/1","title":"Oban.Worker.backoff/1","type":"callback"},{"doc":"Resolve a module from a worker string.Examplesiex&gt; Oban.Worker.from_string(&quot;Oban.Integration.Worker&quot;) {:ok, Oban.Integration.Worker} iex&gt; defmodule NotAWorker, do: [] ...&gt; Oban.Worker.from_string(&quot;NotAWorker&quot;) {:error, %RuntimeError{message: &quot;module is not a worker: NotAWorker&quot;}} iex&gt; Oban.Worker.from_string(&quot;RandomWorker&quot;) {:error, %RuntimeError{message: &quot;unknown worker: RandomWorker&quot;}}","ref":"Oban.Worker.html#from_string/1","title":"Oban.Worker.from_string/1","type":"function"},{"doc":"Build a job changeset for this worker with optional overrides.See Oban.Job.new/2 for the available options.","ref":"Oban.Worker.html#c:new/2","title":"Oban.Worker.new/2","type":"callback"},{"doc":"The perform/1 function is called to execute a job.Each perform/1 function should return :ok or a success tuple. When the return is an error tuple, an uncaught exception or a throw then the error is recorded and the job may be retried if there are any attempts remaining.Note that the args map provided to perform/1 will always have string keys, regardless of the key type when the job was enqueued. The args are stored as jsonb in PostgreSQL and the serialization process automatically stringifies all keys.","ref":"Oban.Worker.html#c:perform/1","title":"Oban.Worker.perform/1","type":"callback"},{"doc":"Set a job's maximum execution time in milliseconds.Jobs that exceed the time limit are considered a failure and may be retried.Defaults to :infinity.","ref":"Oban.Worker.html#c:timeout/1","title":"Oban.Worker.timeout/1","type":"callback"},{"doc":"Return a string representation of a worker module.This is particularly useful for normalizing worker names when building custom Ecto queries.Examplesiex&gt; Oban.Worker.to_string(MyApp.SomeWorker) &quot;MyApp.SomeWorker&quot; iex&gt; Oban.Worker.to_string(Elixir.MyApp.SomeWorker) &quot;MyApp.SomeWorker&quot; iex&gt; Oban.Worker.to_string(&quot;Elixir.MyApp.SomeWorker&quot;) &quot;MyApp.SomeWorker&quot;","ref":"Oban.Worker.html#to_string/1","title":"Oban.Worker.to_string/1","type":"function"},{"doc":"","ref":"Oban.Worker.html#t:result/0","title":"Oban.Worker.result/0","type":"type"},{"doc":"","ref":"Oban.Worker.html#t:t/0","title":"Oban.Worker.t/0","type":"type"},{"doc":"Wraps unhandled exits and throws that occur during job execution.","ref":"Oban.CrashError.html","title":"Oban.CrashError","type":"exception"},{"doc":"Wraps the reason returned by {:error, reason}, {:discard, reason} in a proper exception.The original return tuple is available in the :reason key.","ref":"Oban.PerformError.html","title":"Oban.PerformError","type":"exception"},{"doc":"Returned when a job is terminated early due to a custom timeout.","ref":"Oban.TimeoutError.html","title":"Oban.TimeoutError","type":"exception"},{"doc":"InstallationOban is published on Hex. Add it to your list of dependencies in mix.exs:def deps do [ {:oban, &quot;~&gt; 2.1&quot;} ] endThen run mix deps.get to install Oban and its dependencies, including Ecto, Jason and Postgrex.After the packages are installed you must create a database migration to add the oban_jobs table to your database:mix ecto.gen.migration add_oban_jobs_tableOpen the generated migration in your editor and call the up and down functions on Oban.Migrations:defmodule MyApp.Repo.Migrations.AddObanJobsTable do use Ecto.Migration def up do Oban.Migrations.up() end # We specify `version: 1` in `down`, ensuring that we&#39;ll roll all the way back down if # necessary, regardless of which version we&#39;ve migrated `up` to. def down do Oban.Migrations.down(version: 1) end endThis will run all of Oban's versioned migrations for your database. Migrations between versions are idempotent and will never change after a release. As new versions are released you may need to run additional migrations.Now, run the migration to create the table:mix ecto.migrateNext see Usage for how to integrate Oban into your application and start defining jobs!","ref":"installation.html","title":"Installation","type":"extras"},{"doc":"Troubleshooting","ref":"troubleshooting.html","title":"Troubleshooting","type":"extras"},{"doc":"Oban.Job defines an Ecto schema and the jobs table can therefore be queried as usual, e.g.:MyApp.Repo.all( from j in Oban.Job, where: j.worker == &quot;MyApp.Business&quot;, where: j.state == &quot;discarded&quot; )","ref":"troubleshooting.html#querying-the-jobs-table","title":"Troubleshooting - Querying the Jobs Table","type":"extras"},{"doc":"Elixir and Erlang VersionsIf your app crashes on launch, be sure to confirm you are running the correct version of Elixir and Erlang (view requirements). If using the hashnuke/elixir buildpack, you can update the elixir_buildpack.config file in your application's root directory to something like:# Elixir version elixir_version=1.9.0 # Erlang version erlang_version=22.0.3Available Erlang versions are available here.Database ConnectionsMake sure that you have enough available database connections when running on Heroku. Oban uses a database connection in order to listen for pubsub notifications. This is in addition to your Ecto Repo pool_size setting.Heroku's Hobby tier Postgres plans have a maximum of 20 connections, so if you're using one of those plan accordingly.","ref":"troubleshooting.html#heroku","title":"Troubleshooting - Heroku","type":"extras"},{"doc":"No Notifications with Transaction PoolingUsing PG Bouncer's &quot;Transaction Pooling&quot; setup disables all of PostgreSQL's LISTEN and NOTIFY activity. Some functionality, such as scaling queues, canceling jobs, etc. rely on those notifications.To ensure full functionality you must use a Repo that connects directly to the database, or use another mode like &quot;Session Pooling&quot; if possible.","ref":"troubleshooting.html#pg-bouncer","title":"Troubleshooting - PG Bouncer","type":"extras"},{"doc":"Writing PluginsOban supports the use of plugins to extend its base functionality. A plugin is any module that begins a process and exposes a start_link/1 function. That means a plugin may be a GenServer, an Agent, a Task, or any other OTP behaviour that manages a process. Realistically you'll want a long lived process to complement Oban's behaviour, which makes a GenServer or GenStateMachine ideal.Upon startup, Oban dynamically injects each plugin into its supervision tree and passes a few base options along with any custom configuration for the plugin.Every plugin receives these base options::conf — An Oban.Config struct, which contains all of the options provided to Oban.start_link/1 in a validated and normalized form.:name — A unique via tuple scoped to the current supervision tree. The name may be used later to support helper/interface functions.","ref":"writing_plugins.html","title":"Writing Plugins","type":"extras"},{"doc":"Let's look at a tiny example plugin to get a feel for how options are passed in and how they run. Our plugin will periodically generate a table of counts for each queue / state combination and then print it out. It isn't an amazingly useful plugin, but it demonstrates how to handle options, work with the Oban.Config struct, and periodically interact with the oban_jobs table.defmodule MyApp.Plugins.Breakdown do use GenServer import Ecto.Query, only: [group_by: 3, select: 3] def start_link(opts) do name = Keyword.get(opts, :name, __MODULE__) GenServer.start_link(__MODULE__, opts, name: name) end @impl GenServer def init(opts) do state = Map.new(opts) {:ok, schedule_poll(state)} end @impl GenServer def handle_info(:poll, %{conf: conf} = state) do breakdown = Oban.Repo.all( conf, Oban.Job |&gt; group_by([j], [j.queue, j.state]) |&gt; select([j], {j.queue, j.state, count(j.id)}) ) IO.inspect(breakdown) {:noreply, schedule_poll(state)} end defp schedule_poll(%{interval: interval}) do Process.send_after(self(), :poll, interval) state end endThe plugin's start_link/1 function expects a keyword list with :name, :conf, and :interval values. After extracting the :name for process registration, it passes the options through to init/1. The init function converts the keyword list of options into a map for easier access and then begins a polling loop.Each iteration of the loop will query the oban_jobs table and print out a list of {queue, state, count} tuples like this:[ {&quot;default&quot;, &quot;executing&quot;, 8}, {&quot;default&quot;, &quot;retryable&quot;, 1}, {&quot;default&quot;, &quot;completed&quot;, 3114}, {&quot;default&quot;, &quot;discarded&quot;, 1}, {&quot;events&quot;, &quot;scheduled&quot;, 3}, {&quot;events&quot;, &quot;executing&quot;, 21}, {&quot;events&quot;, &quot;retryable&quot;, 2}, {&quot;events&quot;, &quot;completed&quot;, 1783}, {&quot;events&quot;, &quot;discarded&quot;, 1}, ]Now all that is left is to adding the Breakdown module to Oban's plugin list:config :my_app, Oban, plugins: [ {MyApp.Plugins.Breakdown, interval: :timer.seconds(10)} ] ...In the configuration we're only providing the :interval value. Oban injects the :name and :conf automatically.","ref":"writing_plugins.html#example-plugin","title":"Writing Plugins - Example Plugin","type":"extras"},{"doc":"Plugins are named dynamically using via tuples, which is an effective way to manage process registration for multiple unique Oban instances. However, it makes writing interface functions for plugins a little more complicated. The solution is to make use of the Oban.Registry for process discovery.Imagine adding a pause interface function to the Breakdown plugin we built above:alias Oban.Registry def pause(oban_name \\\\ Oban) do oban_name |&gt; Registry.whereis({:plugin, __MODULE__}) |&gt; GenServer.call(:pause) endThe function accepts an oban_name argument with a default of Oban, which is the default name for an Oban supervision tree. It then calls whereis/2 with a {:plugin, plugin_name} tuple and uses the returned pid to call the plugin process.You can then call pause/1 from elsewhere in the application:MyApp.Plugins.Breakdown.pause() # or MyApp.Plugins.Breakdown.pause(MyApp.OtherOban)","ref":"writing_plugins.html#calling-interface-functions","title":"Writing Plugins - Calling Interface Functions","type":"extras"},{"doc":"Plugins run directly within Oban's supervision tree. A badly behaving plugin, e.g. a plugin that crashes repeatedly, may bring down the entire supervision tree. Be sure that your plugin has safety mechanisms in place to prevent repeated crashes during startup.","ref":"writing_plugins.html#caveats","title":"Writing Plugins - Caveats","type":"extras"},{"doc":"Release ConfigurationWhile having the same Oban configuration for every environment might be fine, there are certainly times you might want to make changes for a specific environment. For example, you may want to increase or decrease a queue's concurrency.","ref":"release_configuration.html","title":"Release Configuration","type":"extras"},{"doc":"If you are using Elixir Releases, this is straight forward to do using Module Config Providers:defmodule MyApp.ConfigProvider do @moduledoc &quot;&quot;&quot; Provide release configuration for Oban Queue Concurrency &quot;&quot;&quot; @behaviour Config.Provider def init(path) when is_binary(path), do: path def load(config, path) do case parse_json(path) do nil -&gt; config queues -&gt; Config.Reader.merge(config, ingestion: [{Oban, [queues: queues]}]) end end defp parse_json(path) do {:ok, _} = Application.ensure_all_started(:jason) if File.exists?(path) do path |&gt; File.read!() |&gt; Jason.decode!() |&gt; Map.fetch!(&quot;queues&quot;) |&gt; Keyword.new(fn {key, value} -&gt; {String.to_atom(key), value} end) end end endOur config provider ensures that the Jason app is loaded so that we can parse a JSON configuration file. Once the JSON is loaded we must extract the queues map and convert it to a keyword list where all of the keys are atoms. The use of String.to_atom/1 is safe because all of our queues names are already defined.Then you include this in your mix.ex file, where your release is configured:releases: [ umbrella_app: [ version: &quot;0.0.1&quot;, applications: [ child_app: :permanent ], config_providers: [{Path.To.ConfigProvider, &quot;/etc/config.json&quot;}] ] ]Then when you release your app, you ensure that you have a JSON file mounted at whatever path you specified above and that it contains all of your desired queues:{&quot;queues&quot;: {&quot;special&quot;: 1, &quot;default&quot;: 10, &quot;events&quot;: 20}}","ref":"release_configuration.html#using-config-providers","title":"Release Configuration - Using Config Providers","type":"extras"},{"doc":"Recursive JobsRecursive jobs, like recursive functions, call themselves after they have have executed. Except unlike recursive functions, where recursion happens in a tight loop, a recursive job enqueues a new version of itself and may add a slight delay to alleviate pressure on the queue.Recursive jobs are a great way to backfill large amounts of data where a database migration or a mix task may not be suitable. Here are a few reasons that a recursive job may be better suited for backfilling data:Data can't be backfilled with a database migration, it may require talking to an external serviceA task may fail partway through execution; resuming the task would mean starting over again, or tracking progress manually to resume where the failure occurredA task may be computationally intensive or put heavy pressure on the databaseA task may run for too long and would be interrupted by code releases or other node restartsA task may interface with an external service and require some rate limitingA job can be used directly for new records and to backfill existing recordsLet's explore recursive jobs with a use case that builds on several of those reasons.","ref":"recursive-jobs.html","title":"Recursive Jobs","type":"extras"},{"doc":"Consider a worker that queries an external service to determine what timezone a user resides in. The external service has a rate limit and the response time is unpredictable. We have a lot of users in our database missing timezone information, and we need to backfill.Our application has an existing TimezoneWorker that accepts a user's id, makes an external request and then updates the user's timezone. We can modify the worker to handle backfilling by adding a new clause to perform/1. The new clause explicitly checks for a backfill argument and will enqueue the next job after it executes:defmodule MyApp.Workers.TimezoneWorker do use Oban.Worker import Ecto.Query alias MyApp.{Repo, User} @backfill_delay 1 @impl true def perform(%{args: %{&quot;id&quot; =&gt; id, &quot;backfill&quot; =&gt; true}}) do with :ok &lt;- perform(%{args: %{&quot;id&quot; =&gt; id}}) do case fetch_next(id) do next_id when is_integer(next_id) -&gt; %{id: next_id, backfill: true} |&gt; new(schedule_in: @backfill_delay) |&gt; Oban.insert() nil -&gt; :ok end end end def perform(%{args: %{&quot;id&quot; =&gt; id}}) do update_timezone(id) end defp fetch_next(current_id) do User |&gt; where([u], is_nil(u.timezone)) |&gt; where([u], u.id &gt; ^current_id) |&gt; order_by(asc: :id) |&gt; limit(1) |&gt; select([u], u.id) |&gt; Repo.one() end defp update_timezone(_id), do: Enum.random([:ok, {:error, :reason}]) endThere is a lot happening in the worker module, so let's unpack it a little bit.There are two clauses for perform/1, the first only matches when a job is marked as &quot;backfill&quot; =&gt; true, the second does the actual work of updating the timezone.The backfill clause checks that the timezone update succeeds and then uses fetch_next/1 to look for the id of the next user without a timezone.When another user needing a backfill is available it enqueues a new backfill job with a one second delay.With the new perform/1 clause in place and our code deployed we can kick off the recursive backfill. Assuming the id of the first user is 1, you can start the job from an iex console:iex&gt; %{id: 1, backfill: true} |&gt; MyApp.Workers.TimezoneWorker.new() |&gt; Oban.insert()Now the jobs will chug along at a steady rate of one per second until the backfill is complete (or something fails). If there are any errors the backfill will pause until the failing job completes: especially useful for jobs relying on flaky external services. Finally, when there aren't any more user's without a timezone, the backfill is complete and recursion will stop.","ref":"recursive-jobs.html#use-case-backfilling-timezone-data","title":"Recursive Jobs - Use Case: Backfilling Timezone Data","type":"extras"},{"doc":"This was a relatively simple example, and hopefully it illustrates the power and flexibility of recursive jobs. Recursive jobs are a general pattern and aren't specific to Oban. In fact, aside from the use Oban.Worker directive there isn't anything specific to Oban in the recipe!","ref":"recursive-jobs.html#building-on-recursive-jobs","title":"Recursive Jobs - Building On Recursive Jobs","type":"extras"},{"doc":"Reliable Scheduled JobsA common variant of recursive jobs are &quot;scheduled jobs&quot;, where the goal is for a job to repeat indefinitely with a fixed amount of time between executions. The part that makes it &quot;reliable&quot; is the guarantee that we'll keep retrying the job's business logic when the job retries, but we'll only schedule the next occurrence once. In order to achieve this guarantee we'll make use of the perform function to receive a complete Oban.Job struct.Time for illustrative example!","ref":"reliable-scheduling.html","title":"Reliable Scheduled Jobs","type":"extras"},{"doc":"When a new user signs up to use our site we need to start sending them daily digest emails. We want to deliver the emails around the same time a user signed up every, repeating every 24 hours. It is important that we don't spam them with duplicate emails, so we ensure that the next email is only scheduled on our first attempt.defmodule MyApp.Workers.ScheduledWorker do use Oban.Worker, queue: :scheduled, max_attempts: 10 alias MyApp.Mailer @one_day 60 * 60 * 24 @impl true def perform(%{args: %{&quot;email&quot; =&gt; email} = args, attempt: 1}) do args |&gt; new(schedule_in: @one_day) |&gt; Oban.insert!() Mailer.deliver_email(email) end def perform(%{args: %{&quot;email&quot; =&gt; email}}) do Mailer.deliver_email(email) end endYou'll notice that the first perform/1 clause only matches a job struct on the first attempt. When it matches, the first clause schedules the next iteration immediately, before attempting to deliver the email. Any subsequent retries fall through to the second perform/1 clause, which only attempts to deliver the email again. Combined, the clauses get us close to at-most-once semantics for scheduling, and at-least-once semantics for delivery.","ref":"reliable-scheduling.html#use-case-delivering-daily-digest-emails","title":"Reliable Scheduled Jobs - Use Case: Delivering Daily Digest Emails","type":"extras"},{"doc":"Delivering around the same time using cron-style scheduling would need extra book-keeping to check when a user signed up, and then only deliver to those users that signed up within that window of time. The recursive scheduling approach is more accurate and entirely self contained—when and if the digest interval changes the scheduling will pick it up automatically once our code deploys.An extensive discussion on the Oban issue tracker prompted this example along with the underlying feature that made it possible.","ref":"reliable-scheduling.html#more-flexible-than-cron-scheduling","title":"Reliable Scheduled Jobs - More Flexible Than CRON Scheduling","type":"extras"},{"doc":"Reporting Job ProgressMost applications provide some way to generate an artifact—something that may take the server a long time to accomplish. If it takes several minutes to render a video, crunch some numbers or generate an export, users may be left wondering whether your application is working. Providing periodic updates to end users assures them that the work is being done and keeps the application feeling responsive.Reporting progress is something that any background job processor with unlimited execution time can do! Naturally, we'll look at an example built on Oban.","ref":"reporting-progress.html","title":"Reporting Job Progress","type":"extras"},{"doc":"Users of our site can export a zip of all the files they have uploaded. A zip file (no, not a tar, our users don't have neck-beards) is generated on the fly, when the user requests it. Lazily generating archives is great for our server's utilization, but it means that users may wait a while when there are many files. Fortunately, we know how many files will be included in the zip and we can use that information to send progress reports! We will compute the archive's percent complete as each file is added and push a message to the user.Before We StartIn the forum question that prompted this guide the work was done externally by a port process. Working with ports is well outside the scope of this guide, so I've modified it for the sake of simplicity. The result is slightly contrived as it puts both processes within the same module, which isn't necessary if the only goal is to broadcast progress. This guide is ultimately about coordinating processes to report progress from a background job, so that's what we'll focus on (everything else will be rather hand-wavy).","ref":"reporting-progress.html#use-case-exporting-a-large-zip-file","title":"Reporting Job Progress - Use Case: Exporting a Large Zip File","type":"extras"},{"doc":"Our worker, the creatively titled ZippingWorker, handles both building the archive and reporting progress to the client. Showing the entire module at once felt distracting, so we'll start with only the module definition and the perform/1 function:defmodule MyApp.Workers.ZippingWorker do use Oban.Worker, queue: :exports, max_attempts: 1 alias MyApp.{Endpoint, Zipper} def perform(%_{args: %{&quot;channel&quot; =&gt; channel, &quot;paths&quot; =&gt; paths}}) do build_zip(paths) await_zip(channel) end # ... endThe function accepts an Oban Job with a channel name and a list of file paths, which it immediately passes on to the private build_zip/1: defp build_zip(paths) do job_pid = self() Task.async(fn -&gt; zip_path = Zipper.new() paths |&gt; Enum.with_index(1) |&gt; Enum.each(fn {path, index} -&gt; :ok = Zipper.add_file(zip_path, path) send(job_pid, {:progress, trunc(index / length(paths) * 100)}) end) send(job_pid, {:complete, zip_path}) end) endThe function grabs the current pid, which belongs to the job, and kicks off an asynchronous task to handle the zipping. With a few calls to a fictional Zipper module the task works through each file path, adding it to the zip. After adding a file the task sends a :progress message with the percent complete back to the job. Finally, when the zip finishes, the task sends a :complete message with a path to the archive.The asynchronous call spawns a separate process and returns immediately. In order for the task to finish building the zip we need to wait on it. Typically we'd use Task.await/2, but we'll use a custom receive loop to track the task's progress: defp await_zip(channel) do receive do {:progress, percent} -&gt; Endpoint.broadcast(channel, &quot;zip:progress&quot;, percent) await_zip(channel) {:complete, zip_path} -&gt; Endpoint.broadcast(channel, &quot;zip:complete&quot;, zip_path) after 30_000 -&gt; Endpoint.broadcast(channel, &quot;zip:failed&quot;, &quot;zipping failed&quot;) raise RuntimeError, &quot;no progress after 30s&quot; end endThe receive loop blocks execution while it waits for :progress or :complete messages. When a message comes in it broadcasts to the provided channel and the client receives an update (this example uses Phoenix Channels, but any other PubSub type mechanism would work). As a safety mechanism we have an after clause that will timeout after 30 seconds of inactivity. If the receive block times out we notify the client and raise an error, failing the job.","ref":"reporting-progress.html#coordinating-processes","title":"Reporting Job Progress - Coordinating Processes","type":"extras"},{"doc":"Reporting progress asynchronously works in Oban because anything that blocks a worker's perform/1 function will keep the job executing. Jobs aren't executed inside of a transaction, which alleviates any limitations on how long a job can run.This technique is suitable for any single long running job where an end user is waiting on the results. You can read the &quot;Batching Jobs for Monitoring&quot; guide if you need to combine multiple jobs into a single output by creating batch jobs.","ref":"reporting-progress.html#made-possible-by-unlimited-execution","title":"Reporting Job Progress - Made Possible by Unlimited Execution","type":"extras"},{"doc":"Handling Expected FailuresReporting job errors by sending notifications to an external service is essential to maintaining application health. While reporting is essential, noisy reports for flaky jobs can become a distraction that gets ignored. Sometimes we expect that a job will error a few times. That could be because the job relies on an external service that is flaky, because it is prone to race conditions, or because the world is a crazy place. Regardless of why a job fails, reporting every failure may be undesirable.","ref":"expected-failures.html","title":"Handling Expected Failures","type":"extras"},{"doc":"One solution for reducing noisy error notifications is to start reporting only after a job has failed several times. Oban uses Telemetry to make reporting errors and exceptions a simple matter of attaching a handler function. In this example we will extend Honeybadger reporting from the Oban.Telemetry documentation, but account for the number of processing attempts.To start, we'll define a Reportable protocol with a single reportable?/2 function:defprotocol MyApp.Reportable do @fallback_to_any true def reportable?(worker, attempt) end defimpl MyApp.Reportable, for: Any do def reportable?(_worker, _attempt), do: true endThe Reportable protocol has a default implementation which always returns true, meaning it reports all errors. Our application has a FlakyWorker that's known to fail a few times before succeeding. We don't want to see a report until after a job has failed three times, so we'll add an implementation of Reportable within the worker module:defmodule MyApp.Workers.FlakyWorker do use Oban.Worker defimpl MyApp.Reportable do @threshold 3 def reportable?(_worker, attempt), do: attempt &gt; @threshold end @impl true def perform(%{args: %{&quot;email&quot; =&gt; email}}) do MyApp.ExternalService.deliver(email) end endThe final step is to call reportable?/2 from our application's error reporter, passing in the worker module and the attempt number:defmodule MyApp.ErrorReporter do alias MyApp.Reportable def handle_event(_, _, %{attempt: attempt, worker: worker} = meta, _) do if Reportable.reportable?(worker, attempt) do context = Map.take(meta, [:id, :args, :queue, :worker]) Honeybadger.notify(meta.error, context, meta.stacktrace) end end endAttach the failure handler somewhere in your application.ex module::telemetry.attach(&quot;oban-errors&quot;, [:oban, :job, :exception], &amp;ErrorReporter.handle_event/4, nil)With the failure handler attached you will start getting error reports only after the third error.Giving Time to RecoverIf a service is especially flaky you may find that Oban's default backoff strategy is too fast. By defining a custom backoff function on the FlakyWorker we can set a linear delay before retries:# inside of MyApp.Workers.FlakyWorker @impl true def backoff(attempt, base_amount \\\\ 60) do attempt * base_amount endNow the first retry is scheduled 60s later, the second 120s later, and so on.Building BlocksElixir's powerful primitives of behaviours, protocols and event handling make flexible error reporting seamless and extendible. While our Reportable protocol only considered the number of attempts, this same mechanism is suitable for filtering by any other meta value.Explore the [event metadata][meta] that Oban provides for job failures to see how you can configure reporting by by worker, queue, or even specific arguments.","ref":"expected-failures.html#use-case-silencing-initial-notifications-for-flaky-services","title":"Handling Expected Failures - Use Case: Silencing Initial Notifications for Flaky Services","type":"extras"},{"doc":"Splitting Queues Between NodesRunning every job queue on every node isn't always ideal. Imagine that your application has some CPU intensive jobs that you'd prefer not to run on nodes that serve web requests. Perhaps you start temporary nodes that are only meant to insert jobs but should never execute any. Fortunately, we can control this by configuring certain node types, or even single nodes, to run only a subset of queues.","ref":"splitting-queues.html","title":"Splitting Queues Between Nodes","type":"extras"},{"doc":"One notorious type of CPU intensive work is video processing. When our application is transcoding multiple videos simultaneously it is a major drain on system resources and may impact response times. To avoid this we can run dedicated worker nodes that don't serve any web requests and handle all of the transcoding.While it's possible to separate our system into web and worker apps within an umbrella, that wouldn't allow us to dynamically change queues at runtime. Let's look at an environment variable based method for dynamically configuring queues at runtime.Within config.exs our application is configured to run three queues: default, media and events:config :my_app, Oban, repo: MyApp.Repo, queues: [default: 15, media: 10, events: 25]We will use an OBAN_QUEUES environment variable to override the queues at runtime. For illustration purposes the queue parsing all happens within the application module, but it would work equally well in releases.exs.defmodule MyApp.Application do @moduledoc false use Application def start(_type, _args) do children = [ MyApp.Repo, MyApp.Endpoint, {Oban, oban_opts()} ] Supervisor.start_link(children, strategy: :one_for_one, name: MyApp.Supervisor) end defp oban_opts do env_queues = System.get_env(&quot;OBAN_QUEUES&quot;) :my_app |&gt; Application.get_env(Oban) |&gt; Keyword.update(:queues, [], &amp;queues(env_queues, &amp;1)) end defp queues(&quot;*&quot;, defaults), do: defaults defp queues(nil, defaults), do: defaults defp queues(false, _), do: false defp queues(values, _defaults) when is_binary(values) do values |&gt; String.split(&quot; &quot;, trim: true) |&gt; Enum.map(&amp;String.split(&amp;1, &quot;,&quot;, trim: true)) |&gt; Keyword.new(fn [queue, limit] -&gt; {String.to_existing_atom(queue), String.to_integer(limit)} end) end endThe queues function's first three clauses ensure that we can fall back to the queues specified in our configuration (or false, for testing). The fourth clause is much more involved, and that is where the environment parsing happens. It expects the OBAN_QUEUES value to be a string formatted as queue,limit pairs and separated by spaces. For example, to run only the default and media queues with a limit of 10 and 5 respectively, you would pass the string default,5 media,10.Note that the parsing clause has a couple of safety mechanisms to ensure that only real queues are specified:It automatically trims while splitting values, so extra whitespace like won't break parsing (i.e. default,3)It only converts the queue string to an existing atom, hopefully preventing typos that would start a random queue (i.e. defalt)Usage ExamplesIn development (or when using mix rather than releases) we can specify the environment variable inline:OBAN_QUEUES=&quot;default,10 media,5&quot; mix phx.server # default: 10, media: 5We can also explicitly opt in to running all of the configured queues:OBAN_QUEUES=&quot;*&quot; mix phx.server # default: 15, media: 10, events: 25Finally, without OBAN_QUEUES set at all it will implicitly fall back to the configured queues:mix phx.server # default: 15, media: 10, events: 25","ref":"splitting-queues.html#use-case-isolating-video-processing-intensive-jobs","title":"Splitting Queues Between Nodes - Use Case: Isolating Video Processing Intensive Jobs","type":"extras"},{"doc":"This environment variable based solution is more flexible than running separate umbrella apps because we can reconfigure at any time. In a limited environment, like staging, we can run all the queues on a single node using the exact same code we use in production. In the future, if other workers start to utilize too much CPU or RAM we can shift them to the worker node without any code changes.This guide was prompted by an inquiry on the Oban issue tracker.","ref":"splitting-queues.html#flexible-across-all-environments","title":"Splitting Queues Between Nodes - Flexible Across all Environments","type":"extras"},{"doc":"ChangelogAll notable changes to this project will be documented in this file.The format is based on Keep a Changelog and this project adheres to Semantic Versioning.","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"","ref":"changelog.html#unreleased","title":"Changelog - Unreleased","type":"extras"},{"doc":"[Oban.Migration] Conditionally skip altering oban_job_state if the cancelled state is already present. This allows for a smoother upgrade path for users running PG &lt; 12. See the notes for v2.3.0 for more details.","ref":"changelog.html#2-3-3-2020-11-10","title":"Changelog - 2.3.3 — 2020-11-10","type":"extras"},{"doc":"[Oban.Migration] Restore indexes possibly removed by changing oban_job_state. This only applies to PG older than version 12.[Oban.Plugins.Pruner] Prune cancelled jobs along with completed or discarded.","ref":"changelog.html#2-3-2-2020-11-06","title":"Changelog - 2.3.2 — 2020-11-06","type":"extras"},{"doc":"[Oban.Migration] Conditionally alter oban_job_state if the PG version is 12 or greater. This is vastly faster than the renaming, adding and dropping required for older PG versions.","ref":"changelog.html#2-3-1-2020-11-06","title":"Changelog - 2.3.1 — 2020-11-06","type":"extras"},{"doc":"Migration Required (V9)Migration V9 brings the new cancelled state, a cancelled_at column, and job meta.Older versions of PostgreSQL, prior to version 12, don't allow altering an enum within a transaction. If you're running an older version and want to prevent a table rewrite you can add the new cancelled at state before running the V9 migration.Create a migration with @disable_ddl_transaction true declared and run the command ALTER TYPE oban_job_state ADD VALUE IF NOT EXISTS 'cancelled'. The V9 migration will see that the cancelled value exists and won't attempt to modify the enum.After you've sorted adding the cancelled state (or ignored the issue, because you're either running PG &gt;= 12 or don't have many jobs retained), generate a new migration:mix ecto.gen.migration upgrade_oban_jobs_to_v9Next, call Oban.Migrations in the generated migration:defmodule MyApp.Repo.Migrations.UpdateObanJobsToV9 do use Ecto.Migration defdelegate up, to: Oban.Migrations defdelegate down, to: Oban.Migrations endOban will manage upgrading to V9 regardless of the version your application is currently using, and it will roll back a single version.Added[Oban.Job] Add new meta field for storing arbitrary job data that isn't appropriate as args.[Oban.Job] Introduce a cancelled state, along with a new cancelled_at timestamp field. Cancelling a job via Oban.cancel_job (or via Oban Web) now marks the job as cancelled rather than discarded.[Oban.Worker] Add from_string/1 for improved worker module resolution.[Oban.Telemetry] Pass the full job schema in telemetry metadata, not only select fields. Individual fields such as args, worker, etc. are still passed for backward compatibility. However, their use is deprecated and they are no longer documented.Fixed[Oban.Notifier] Fix resolution of Oban.Notifier child process in Oban.Notifier.listen/2.[Oban.Queue.Producer] Fix cancelling jobs without a supervised process. In some circumstances, namely a hot code upgrade, the job's process could terminate without the producer tracking it and leave the job un-killable.[Oban] Only convert invalid changesets into ChangesetError from insert!. This prevents unexpected errors when insert! is called within a transaction after the transaction has rolled back.","ref":"changelog.html#2-3-0-2020-11-06","title":"Changelog - 2.3.0 — 2020-11-06","type":"extras"},{"doc":"Added[Oban] Replace local dynamically composed names with a registry. This dramatically simplifies locating nested children, avoids unnecessary atom creation at runtime and improves the performance of config lookups.[Oban.Repo] The new Oban.Repo module wraps interactions with underlying Ecto repos. This ensures consistent prefix and log level handling, while also adding full dynamic repo support.The Oban.Repo wrapper should be used when authoring plugins.[Oban.Job] Augment the unique keys option with replace_args, which allows enqueuing a unique job and replacing the args subsequently. For example, given a job with these args: %{some_value: 1, id: 123}Attempting to insert a new job: %{some_value: 2, id: 123} |&gt; MyJob.new(schedule_in: 10, replace_args: true unique: [keys: [:id]]) |&gt; Oban.insert()Will result in a single job with the args: %{some_value: 2, id: 123}[Oban] Add Oban.check_queue/1,2 for runtime introspection of a queue's state. It displays a compactly formatted breakdown of the queue's configuration, a list of running jobs, and static identity information.[Oban] Add Oban.retry_job/1,2, used to manually retry a discarded or retryable job.[Oban.Telemetry] Include tags as part of the metadata for job events.Changed[Oban.Worker] The default backoff algorithm now includes a small amount of jitter. The jitter helps prevent jobs that fail simultaneously from repeatedly retrying together.Fixed[Oban.Job] Don't include discarded state by default when accounting for uniqueness.[Oban.Cron] Fully support weekeday ranges in crontab expressions. Previously, given a weekday range like MON-WED the parser would only retain the MON.","ref":"changelog.html#2-2-0-2020-10-12","title":"Changelog - 2.2.0 — 2020-10-12","type":"extras"},{"doc":"Fixed[Oban.Testing] Modify the behaviour checking in perform_job/3 so that it considers all behaviours.[Oban.Plugins.Pruner] Use distinct names when running multiple Oban instances in the same application.Added[Oban] In addition to the standard queue-name / limit keyword syntax it is now possible to define per-queue options such as :poll_interval, :dispatch_cooldown, :paused, and even override the :producer module.[Oban] Cancelling a running job now records an error on the job if it was running at the time it was cancelled.[Oban.Job] Allow using :discarded as a unique state and expose all possible states through Oban.Job.states/0.[Oban.Worker] Allow returning {:discard, reason} from perform/1, where the reason is recorded in the job's errors array. If no reason is provided then a default of &quot;None&quot; is used. All discarded jobs will have an error now, whether discarded manually or automatically.[Oban.Job] Add support for uniqueness across sub-args with the :keys attribute. This makes it possible to only use a few values from a job to determine uniqueness. For example, a job that includes a lot of metadata but must be considered unique based on a :url field can specify keys: [:url].Changed[Oban.Worker] Wrap {:error, reason} and {:discard, reason} in a proper Oban.PerformError exception with a customized message. This ensures that the :error value passed to telemetry handlers is an exception and not a raw term.[Oban.Worker] Wrap job timeouts in Oban.TimeoutError with a customized message indicating the worker and timeout value. This replaces the raw :timeout atom that was reported before.[Oban.Worker] Wrap caught exits and throws in Oban.CrashError with a formatted message. This means the :error value passed to telemetry is always a proper exception and easier to report.[Oban.Worker] Stop reporting internal stacktraces for timeouts, discards or error tuples. The stacktrace was useless and potentially misleading as it appeared that the error originated from Oban rather than the worker module.","ref":"changelog.html#2-1-0-2020-08-21","title":"Changelog - 2.1.0 — 2020-08-21","type":"extras"},{"doc":"No changes from 2.0.0-rc.3.","ref":"changelog.html#2-0-0-2020-07-10","title":"Changelog - 2.0.0 — 2020-07-10","type":"extras"},{"doc":"Fixed[Oban.Crontab.Cron] Do not raise an ArgumentError exception when the crontab configuration includes a step of 1, which is a valid step value.[Oban.Telemetry] Correctly record duration and queue_time using native time units, but log them using microseconds. Previously they used a mixture of native and microseconds, which yielded inconsistent values.Added[Oban.Telemetry] Include job queue_time in the default logger output.Changed[Oban.Plugins.Pruner] The FixedPruner is renamed to Pruner and allows users to configure the max_age value. This partially reverses the breaking change imposed by the move from prune to the FixedPruner plugin, though there isn't any support for max_len or dynamic pruning.","ref":"changelog.html#2-0-0-rc-3-2020-07-01","title":"Changelog - 2.0.0-rc.3 — 2020-07-01","type":"extras"},{"doc":"Breaking Changes[Oban] The interface for pause_queue/2, resume_queue/2 and scale_queue/3 now matches the recently changed start_queue/2 and stop_queue/2. All queue manipulation functions now have a consistent interface, including the ability to work in :local_only mode. See the notes around start_queue/2 and stop_queue/2 in 2.0.0-rc.1 for help upgrading.[Oban] Replace drain_queue/3 with drain_queue/2, which now has an interface consistent with the other *_queue/2 operations.Where you previously called drain_queue/2,3 like this: Oban.drain_queue(:myqueue, with_safety: false)You'll now it with options, like this: Oban.drain_queue(queue: :myqueue, with_safety: false)Fixed[Oban.Breaker] Prevent connection bomb when the Notifier experiences repeated disconnections.[Oban.Queue.Executor] Fix error reporting when a worker fails to resolve.[Oban.Telemetry] Stop logging the :error value for circuit trip events. The error is a struct that isn't JSON encodable. We include the normalized Postgrex / DBConnection message already, so the error is redundant.Added[Oban.Crontab] Add @reboot special case to crontab scheduling[Oban.Telemetry] Add new :producer events for descheduling and dispatching jobs from queue producers.Removed[Oban] Removed Oban.kill_job/2, which wasn't as flexible as Oban.cancel_job/2. Use Oban.cancel_job/2 instead to safely discard scheduled jobs or killing executing jobs.","ref":"changelog.html#2-0-0-rc-2-2020-06-23","title":"Changelog - 2.0.0-rc.2 — 2020-06-23","type":"extras"},{"doc":"Breaking Changes[Oban.Config] The :verbose setting is renamed to :log. The setting started off as a simple boolean, but it has morphed to align with the log values accepted by calls to Ecto.Repo.To migrate, replace any :verbose declarations: config :my_app, Oban, verbose: false, ...With use of :log instead: config :my_app, Oban, log: false, ...[Oban] The interface for start_queue/3 is replaced with start_queue/2 and stop_queue/2 no longer accepts a queue name as the second argument. Instead, both functions now accept a keyword list of options. This enables the new local_only flag, which allows you to dynamically start and stop queues only for the local node.Where you previously called start_queue/2,3 or stop_queue/2 like this: :ok = Oban.start_queue(:myqueue, 10) :ok = Oban.stop_queue(:myqueue)You'll now call them with options, like this: :ok = Oban.start_queue(queue: :myqueue, limit: 10) :ok = Oban.stop_queue(queue: :myqueue)Or, to only control the queue locally: :ok = Oban.start_queue(queue: :myqueue, limit: 10, local_only: true) :ok = Oban.stop_queue(queue: :myqueue, local_only: true)","ref":"changelog.html#2-0-0-rc-1-2020-06-12","title":"Changelog - 2.0.0-rc.1 — 2020-06-12","type":"extras"},{"doc":"Breaking Changes[Oban.Worker] The perform/2 callback is replaced with perform/1, where the only argument is an Oban.Job struct. This unifies the interface for all Oban.Worker callbacks and helps to eliminate confusion around pattern matching on arguments.To migrate change all worker definitions from accepting an args map and a job struct: def perform(%{&quot;id&quot; =&gt; id}, _job), do: IO.inspect(id)To accept a single job struct and match on the args key directly: def perform(%Job{args: %{&quot;id&quot; =&gt; id}}), do: IO.inspect(id)[Oban.Worker] The backoff/1 callback now expects a job struct instead of an integer. That allows applications to finely control backoff based on more than just the current attempt number. Use of backoff/1 with an integer is no longer supported.To migrate change any worker definitions that used a raw attempt like this: def backoff(attempt), do: attempt * 60To match on a job struct instead, like this: def backoff(%Job{attempt: attempt}), do: attempt * 60[Oban.Telemetry] The format for telemetry events has changed to match the new telemetry span convention. This listing maps the old event to the new one:[:oban, :started] -&gt; [:oban, :job, :start][:oban, :success] -&gt; [:oban, :job, :stop][:oban, :failure] -&gt; [:oban, :job, :exception][:oban, :trip_circuit] -&gt; [:oban, :circuit, :trip][:oban, :open_circuit] -&gt; [:oban, :circuit, :open] In addition, for exceptions the stacktrace meta key has changed from :stack to the standardized :stacktrace.[Oban.Prune] Configurable pruning is no longer available. Instead, pruning is handled by the new plugin system. A fixed period pruning module is enabled as a default plugin. The plugin always retains &quot;prunable&quot; (discarded or complete) jobs for 60 seconds.Remove any :prune, :prune_interval or prune_limit settings from your config. To disable the pruning plugin in test mode set plugins: false instead.[Oban.Beat] Pulse tracking and periodic job rescue are no longer available. Pulse tracking and rescuing will be handled by an external plugin. This is primarily an implementation detail, but it means that jobs may be left in the executing state after a crash or forced shutdown.Remove any :beats_maxage, :rescue_after or :rescue_interval settings from your config.Fixed[Oban.Scheduler] Ensure isolation between transaction locks in different prefixes. A node with multiple prefix-isolated instances (i.e. &quot;public&quot; and &quot;private&quot;) would always attempt to schedule cron jobs at the same moment. The first scheduler would acquire a lock and block out the second, preventing the second scheduler from ever scheduling jobs.[Oban.Query] Correctly prefix unprepared unique queries. Unique queries always targeted the &quot;public&quot; prefix, which either caused incorrect results when there were both &quot;public&quot; and an alternate prefix. In situations where there wasn't a public oban_jobs table at all it would cause cryptic transaction errors.[Oban.Query] Wrap all job fetching in an explicit transaction to enforce FOR UPDATE SKIP LOCKED semantics. Prior to this it was possible to run the same job at the same time on multiple nodes.[Oban.Crontab] Fix weekday matching for Sunday, which is represented as 0 in crontabs.Added[Oban] Bubble up errors and exits when draining queues by passing with_safety: false as an option to Oban.drain_queue/3.[Oban] Add Oban.cancel_job/2 for safely discarding scheduled jobs or killing executing jobs. This deprecates kill_job/2, which isn't as flexible.[Oban.Worker] Support returning {:snooze, seconds} from perform/1 to re-schedule a job some number of seconds in the future. This is useful for recycling jobs that aren't ready to run yet, e.g. because of rate limiting.[Oban.Worker] Support returning :discard from perform/1 to immediately discard a job. This is useful when a job encounters an error that won't resolve with time, e.g. invalid arguments or a missing record.[Oban.Job] Introduce a virtual unsaved_error field, which is populated with an error map after failed execution. The unsaved_error field is set before any calls to the worker's backoff/1 callback, allowing workers to calculate a custom backoff depending on the error that failed the job.[Oban.Worker] Add :infinity option for unique period.[Oban.Telemetry] Add span/3 for reporting normalized :start, :stop and :exception events with timing information.[Oban.Telemetry] Include the configured prefix in all event metadata. This makes it possible to identify which schema prefix a job ran with, which is useful for differentiating errors in a multi-tenant system.[Oban.Telemetry] Include queue_time as a measurement with stop and exception events. This is a measurement in milliseconds of the amount of time between when a job was scheduled to run and when it was last attempted.[Oban.Testing] Add perform_job/2,3 helper to automate validating, normalizing and performing jobs while unit testing. This is now the preferred way to unit test workers.To update your tests replace any calls to perform/1,2 with the new Oban.Testing.perform_job/2,3 helper:defmodule MyApp.WorkerTest do use MyApp.DataCase, async: true use Oban.Testing, repo: MyApp.Repo alias MyApp.Worker test &quot;doing business in my worker&quot; do assert :ok = perform_job(Worker, %{id: 1}) end endThe perform_job/2,3 helper will verify the worker, the arguments and any provided options. It will then verify that your worker returns a valid result and return the value for you to assert on.[Oban.Crontab] Add support for non-standard expressions such as @daily, @hourly, @midnight, etc.[Oban.Crontab] Add support for using step values in conjunction with ranges, enabling expressions like 10-30/2, 15-45/3, etc.Changed[Oban.Notifier] Make the module public and clean up the primary function interfaces. Listening for and delivering notifications is simplified and no longer requires macros for pattern matching.Notifier dispatching performance is slightly improved as well. It is now a no-op if no processes are listening to a notification's channel.[Oban.Query] The completed_at timestamp is no longer set for failed jobs, whether they are put in the discarded or retryable state. However, the information is still available and is recorded in the errors array as the at value with the error for that attempt.This corrects a long standing inconsistency between discarding a job manually or automatically when it exhausts retries.[Oban.Producer] Stop dispatching jobs immediately on queue startup. Instead, only dispatch on the first poll. This makes it possible to send the producer a message or allow sandboxed connection access before the initial dispatch.[Oban.Worker] Limit default backoff calculations to 20 attempts, or roughly 24 days. The change addresses an issue with snoozing, which can increase a job's attempts into the hundreds or thousands. In this situation the algorithm calculates the backoff using a ratio of attempts to max attempts, but is still limited to roughly 24 days.","ref":"changelog.html#2-0-0-rc-0-2020-06-03","title":"Changelog - 2.0.0-rc.0 — 2020-06-03","type":"extras"},{"doc":"Fixed[Oban] Handle the :shutdown message when a job is killed purposefully. Previously the message was ignored, which caused the producer to keep a reference to the discarded job and prevented dispatching additional jobs.Added[Oban.Testing] Add assert_enqueued/2 and refute_enqueued/2 to allow asserting with a timeout, like assert_received.[Oban.Telemetry] Add [:oban, :started] event to report the system time as start_time when jobs start. This enables span tracking for jobs, improving integration with monitoring tools like NewRelic.","ref":"changelog.html#1-2-0-2020-03-05","title":"Changelog - 1.2.0 — 2020-03-05","type":"extras"},{"doc":"Fixed[Oban.Crontab] Allow any number of spaces and/or tabs in cron expressions.[Oban.Pruner] Prevent deadlocks while pruning by ensuring that only a single node can prune at any given time.[Oban.Queue.Executor] Improve handling of nested/linked process failures. Previously if a job spawned a process that crashed (i.e. through Task.async) then the job was left stuck in an executing state.Added[Oban.Worker] Add timeout/1 callback for limiting how long a job may execute. The default value is :infinity, which allows a job to run indefinitely and mirrors the previous behavior.[Oban.Telemetry] Add :error and :stack to trip_circuit event metadata.[Oban.Queue.Executor] Retry success/failure database updates after a job finishes.On occasion something may happen to the database connection and updating a job's state would fail. Now we retry writing with a linear backoff to prevent the job from getting stuck in an executing state.Changed[Oban.Worker] Tighten the spec for perform/2. Now workers are expected to return only :ok, {:ok, result} or {:error, reason}. A warning is logged if any other value is returned—for high throughput systems this may cause performance issues and you should update your worker's return values.Returning a success tuple is supported for testing purposes and backward compatibility, while returning :ok on success if preferred.","ref":"changelog.html#1-1-0-2020-02-17","title":"Changelog - 1.1.0 — 2020-02-17","type":"extras"},{"doc":"No changes from 1.0.0-rc.2.","ref":"changelog.html#1-0-0-2020-01-29","title":"Changelog - 1.0.0 — 2020-01-29","type":"extras"},{"doc":"Fixed[Oban.Migration] Separate adding and modifying new columns in the V8 migration. The columns can't be modified without a flush.Changed[Oban.Testing] Accept a custom prefix when making test assertions.","ref":"changelog.html#1-0-0-rc-2-2020-01-22","title":"Changelog - 1.0.0-rc.2 — 2020-01-22","type":"extras"},{"doc":"Migration Required (V8)This is the first required migration since 0.8.0, released in 09/2019. It brings with it a new column, discarded_at, a streamlined notifications trigger, job prioritiy and job tags.Upgrading only requires running the new migration.First, generate a new migration:mix ecto.gen.migration upgrade_oban_jobs_to_v8Next, call Oban.Migrations in the generated migration:defmodule MyApp.Repo.Migrations.UpdateObanJobsToV8 do use Ecto.Migration def up, do: Oban.Migrations.up(version: 8) def down, do: Oban.Migrations.down() endOban will manage upgrading to V8 regardless of the version your application is currently using, and it will roll back a single version.Added[Oban] Add timezone support for scheduling cronjobs using timezones other than &quot;Etc/UTC&quot;. Using a custom timezone requires a timezone database such as tzdata.[Oban] Add dispatch_cooldown option to configure the minimum time between a producer fetching more jobs to execute.[Oban] Add beats_maxage option to configure how long heartbeat rows are retained in the oban_beats table. Each queue generates one row per second, so rows may accumulate quickly. The default value is now five minutes, down from one hour previously.[Oban.Job] Add discarded_at timestamp to help identify jobs that were discarded and not completed. The timestamp is added by the V8 migration and it is also included in the original create table from V1 as a minor space saving optimization (packing datetime columns together because they use a predictable 4bytes of space).[Oban.Job] Add numerical priority value to control the order of execution for jobs within a queue. The priority can be set between 0 and 3, with 0 being the default and the highest priority.[Oban.Job] Add tags field for arbitrarily organizing associated tags. Tags are a list of strings stored as an array in the database, making them easy to search and filter by.Changed[Oban] Change the default prune value from :disabled to {:maxlen, 1_000}. Many people don't change the default until they realize that a lot of jobs are lingering in the database. It is rare that anybody would want to keep all of their jobs forever, so a conservative default is better than :disabled.[Oban] Change oban_beats retention from one hour to five minutes. The value is now configurable, with a default of 300s. The lower bound is 60s because we require one minute of heartbeat activity to rescue orphaned jobs.[Oban.Queue.Producer] Introduce &quot;dispatch cooldown&quot; as a way to debounce repeatedly fetching new jobs. Repeated fetching floods the producer's message queue and forces the producer to repeatedly fetch one job at a time, which is not especially efficient. Debounced fetching is much more efficient for the producer and the database, increasing maximum jobs/sec throughput so that it scales linearly with a queue's concurrency settings (up to what the database can handle).[Oban.Query] Discard jobs that have exhausted the maximum attempts rather than rescuing them. This prevents repeatedly attempting a job that has consistently crashed the BEAM.[Oban.Query] Use transactional locks to prevent duplicate inserts without relying on unique constraints in the database. This provides strong unique guarantees without requiring migrations.Removed[Oban.Notifications] An overhauled and simplified insert trigger no longer emits update notifications. This was largely an internal implementation detail and wasn't publicly documented, but it does effect the UI.","ref":"changelog.html#1-0-0-rc-1-2020-01-21","title":"Changelog - 1.0.0-rc.1 — 2020-01-21","type":"extras"},{"doc":"Fixed[Oban.Worker] Deep merge unique options between the worker and custom params. Previously the unique options passed to Worker.new/2 would completely override the options stored by use Oban.Worker. This was primarily an issue for crontab jobs, where the period is always passed by the scheduler.Changed[Oban] Allow setting crontab: false or crontab: nil to disable scheduling. This matches the queues behavior and makes it possible to override the default configuration within each environment, i.e. when testing.[Oban.Testing] Better error message for Oban.Testing.assert_enqueued/2","ref":"changelog.html#0-12-1-2019-12-13","title":"Changelog - 0.12.1 — 2019-12-13","type":"extras"},{"doc":"Migration Optional (V7)The queries used to prune by limit and age are written to utilize a single partial index for a huge performance boost on large tables. The new V7 migration will create the index for you—but that may not be ideal for tables with millions of completed or discarded jobs because it can't be done concurrently.If you have an extremely large jobs table you can add the index concurrently in a dedicated migration:create index( :oban_jobs, [&quot;attempted_at desc&quot;, :id], where: &quot;state in (&#39;completed&#39;, &#39;discarded&#39;)&quot;, name: :oban_jobs_attempted_at_id_index, concurrently: true )Added[Oban] Add start_queue/3 and stop_queue/2 for dynamically starting and stopping supervised queues across nodes.[Oban] Add drain_queue/3 to accept drain options. with_scheduled: true allows draining scheduled jobs.[Oban] Expose circuit_backoff as a &quot;twiddly&quot; option that controls how long tripped circuit breakers wait until re-opening.[Oban.Testing] Accept a value/delta tuple for testing timestamp fields. This allows more robust testing of timestamps such as scheduled_at.[Oban.Telemetry] Emit [:oban, :trip_circuit] and [:oban, :open_circuit] events for circuit breaker activity. Previously an error was logged when the circuit was tripped, but there wasn't any way to monitor circuit breakers.Circuit breaker activity is logged by the default telemetry logger (both :trip_circuit and :open_circuit events).Fixed[Oban.Query] Avoid using prepared statements for all unique queries. This forces Postgres to use a custom plan (which utilizes the compound index) rather than falling back to a generic plan.[Oban.Job] Include all permitted fields when converting a Job to a map, preserving any optional values that were either specified by the user or came via Worker defaults.[Oban.Migrations] Guard against missing migration modules in federated environments.Changed[Oban] Allow the multi name provided to Oban.insert/3,4 to be any term, not just an atom.[Oban.Query] Use a consistent and more performant set of queries for pruning. Both pruning methods are optimized to utilize a single partial index.","ref":"changelog.html#0-12-0-2019-11-26","title":"Changelog - 0.12.0 – 2019-11-26","type":"extras"},{"doc":"Fixed[Oban.Pruner] Apply prune_limit when pruning beats. A lot of beats can accumulate when pruning has been disabled. Later, when pruning is enabled, the pruner would try to delete too many beats and would time out.[Oban.Crontab.Scheduler] Use a zero arity function for the transaction callback. The one arity version is only available in Ecto &gt;= 3.2.","ref":"changelog.html#0-11-1-2019-11-13","title":"Changelog - 0.11.1 — 2019-11-13","type":"extras"},{"doc":"Migration Optional (V6)Job id's greater than 2,147,483,647 (PG int limit) can't be inserted into the running array on oban_beats. The array that Ecto defines uses int instead of bigint, which can't store the larger integers. This migration changes the column type to bigint[], a locking operation that may take a few seconds.Added[Oban] Added crontab support for automatically enqueuing jobs on a fixed schedule. A combination of transactional locks and unique jobs prevents scheduling duplicate jobs.Fixed[Oban.Migrations] Add a comment when migrating oban_jobs to V5 and when rolling back down to V4.[Oban.Query] Apply the configured log level to unique queries.[Oban.Notifier] Prevent open connections from accumulating when the circuit is tripped during the connection phase. This change may leave notifications in a state where they aren't listening to all channels.Changed[Oban.Notifier] Replay oban_update notifications to subscribed processes.","ref":"changelog.html#0-11-0-2019-11-06","title":"Changelog - 0.11.0 — 2019-11-06","type":"extras"},{"doc":"Changed[Oban.Notifier] Replay oban_gossip notifications to subscribed processes.","ref":"changelog.html#0-10-1-2019-10-08","title":"Changelog - 0.10.1 — 2019-10-08","type":"extras"},{"doc":"Migration Optional (V5)Tables with a lot of available jobs (hundreds of thousands to several million) are prone to time outs when fetching new jobs. The planner fails to optimize using the index available on queue, state and scheduled_at, forcing both a slow sort pass and an expensive bitmap heap scan.This migration drops the separate indexes in favor of a a single composite index. The resulting query is up to 258,757x faster on large tables while still usable for all of the other maintenance queries.History of the EXPLAIN ANALYZE output as the query was optimized is available here: https://explain.depesz.com/s/9Vh7Changed[Oban.Config] Change the default for verbose from true to false. Also, :verbose now accepts only false and standard logger levels. This change aims to prevent crashes due to conflicting levels when the repo's log level is set to false.Fixed[Oban.Notifier] Restructure the notifier in order to to isolate producers from connection failures. Errors or loss of connectivity in the notification connection no longer kills the notifier and has no effect on the producers.","ref":"changelog.html#0-10-0-2019-10-03","title":"Changelog - 0.10.0 — 2019-10-03","type":"extras"},{"doc":"Added[Oban] Add insert_all/2 and insert_all/4, corresponding to Ecto.Repo.insert_all/3 and Ecto.Multi.insert_all/5, respectively.[Oban.Job] Add to_map/1 for converting a changeset into a map suitable for database insertion. This is used by Oban.insert_all/2,4 internally and is exposed for convenience.Changed[Oban.Config] Remove the default queue value of [default: 10], which was overriden by Oban.start_link/1 anyhow.[Oban.Telemetry] Allow the log level to be customized when attaching the default logger. The default level is :info, the same as it was before.Fixed[Oban.Migrations] Prevent invalid up and down targets when attempting to run migrations that have already been ran. This was primarily an issue in CI, where the initial migration was unscoped and would migrate to the current version while a subsequent migration would attempt to migrate to a lower version.[Oban.Job] Prevent a queue comparison with nil by retaining the default queue (default) when building uniqueness checks.[Oban.Job] Set state to scheduled for jobs created with a scheduled_at timestamp. Previously the state was only set when schedule_in was used.","ref":"changelog.html#0-9-0-2019-09-20","title":"Changelog - 0.9.0 — 2019-09-20","type":"extras"},{"doc":"Changed[Oban.Notifier] Restore the gossip macro and allow oban_gossip channel for notifications.Fixed[Oban.Migrations] Prevent invalid up and down ranges when repeatedly migrating without specifying a version. This issue was seen when running all of the up migrations on a database from scratch, as there would be multiple oban migrations that simply delegated to up and down.","ref":"changelog.html#0-8-1-2019-09-11","title":"Changelog - 0.8.1 — 2019-09-11","type":"extras"},{"doc":"Migration Required (V4)This release requires a migration to V3, with an optional migration to V4. V3 adds a new column to jobs and creates the oban_beats table, while V4 drops a function used by the old advisory locking system.For a smooth zero-downtime upgrade, migrate to V3 first and then V4 in a separate release. The following sample migration will only upgrade to V3:defmodule MyApp.Repo.Migrations.AddObanBeatsTable do use Ecto.Migration def up do Oban.Migrations.up(version: 3) end def down do Oban.Migrations.down(version: 2) end endAdded[Oban.Job] Add an attempted_by column used to track the node, queue and producer nonce that attempted to execute a job.[Oban.Beat] Add a new oban_beats table, used by producers to publish &quot;pulse&quot; information including the node, queue, running jobs and other information previously published by gossip notifications.Beat records older than one hour are pruned automatically. Beat pruning respects the :disabled setting, but it ignores length and age configuration. The goal is to prevent bloating the table with useless historic information—each queue generates one beat a second, 3,600 beat records per hour even when the queue is idle.Changed[Oban.Executor] Don't swallow an ArgumentError when raised by a worker's backoff function.[Oban.Notifier] Remove gossip notifications entirely, superseded by pulse activity written to oban_beats.[Oban.Query] Remove all use of advisory locks![Oban.Producer] Periodically attempt to rescue orphans, not just at startup. By default a rescue is attempted once a minute and it checks for any executing jobs belonging to a producer that hasn't had a pulse for more than a minute.Fixed[Oban.Worker] Validate worker options after the module is compiled. This allows dynamic configuration of compile time settings via module attributes, functions, Application.get_env/3, etc.[Oban.Query] Remove scheduled_at check from job fetching query. This could prevent available jobs from dispatching when the database's time differed from the system time.[Oban.Migrations] Fix off-by-one error when detecting the version to migrate up from.","ref":"changelog.html#0-8-0-2019-09-06","title":"Changelog - 0.8.0 — 2019-09-06","type":"extras"},{"doc":"Fixed[Oban.Query] Release advisory locks in batches rather than individually after a job finishes execution. By tracking unlockable jobs and repeatedly attempting to unlock them for each connection we ensure that eventually all advisory locks are released.The previous unlocking system leaked advisory locks at a rate proportional to the number of connections in the db pool. The more connections, the more locks that wouldn't release. With a default value of 64 for max_locks_per_transaction the database would raise &quot;ERROR: 53200: out of shared memory&quot; after it hit a threshold (11,937 exactly, in my testing).[Oban.Worker] Allow max_attempts to be 1 or more. This used to be possible and was broken unintentionally by compile time validations.","ref":"changelog.html#0-7-1-2019-08-15","title":"Changelog - 0.7.1 — 2019-08-15","type":"extras"},{"doc":"Added[Oban] Added insert/2, insert!/2 and insert/4 as a convenient and more powerful way to insert jobs. Features such as unique jobs and the upcoming prefix support only work with insert.[Oban] Add prefix support. This allows entirely isolated job queues within the same database.[Oban.Worker] Compile time validation of all passed options. Catch typos and other invalid options when a worker is compiled rather than when a job is inserted for the first time.[Oban.Worker] Unique job support through the unique option. Set a unique period, and optionally fields and states, to enforce uniqueness within a window of time. For example, to make a job unique by args, queue and worker for 2 minutes:use Oban.Worker, unique: [period: 120, fields: [:args, :queue, :worker]] Note, unique support relies on the use of Oban.insert/2,4.Changed[Oban.Worker] Remove the perform/1 callback in favor of perform/2. The new perform/2 function receives the job's args, followed by the complete job struct. This new function signature makes it clear that the args are always available, and the job struct is also there when it is needed. A default perform/2 function is not generated automatically by the use macro and must be defined manually.This is a breaking change and all worker modules will need to be updated. Thankfully, due to the behaviour change, warnings will be emitted when you compile after the upgrade. If your perform functions weren't matching on the Oban.Job struct then you can migrate your workers by adding a second _job argument: def perform(%{&quot;some&quot; =&gt; args}, _job) If you were making use of Oban.Job metadata in perform/1 then you can move the job matching to the second argument: def perform(_args, %{attempt: attempt}) See the issue that suggested this change for more details and discussion.[Oban.Producer] Use send_after/3 instead of :timer.send_interval/2 to maintain scheduled dispatch. This mechanism is more accurate under system load and it prevents :poll messages from backing up for each producer.[Oban.Migration] Accept a keyword list with :prefix and :version as options rather than a single version string. When a prefix is supplied the migration will create all tables, indexes, functions and triggers within that namespace. For example, to create the jobs table within a &quot;private&quot; prefix:Oban.Migrate.up(prefix: &quot;private&quot;)","ref":"changelog.html#0-7-0-2019-08-08","title":"Changelog - 0.7.0 — 2019-08-08","type":"extras"},{"doc":"Added[Oban.Query] Added :verbose option to control general query logging. This allows debug query activity within Oban to be silenced during testing and development.[Oban.Testing] Added all_enqueued/1 helper for testing. The helper returns a list of jobs matching the provided criteria. This makes it possible to test using pattern matching, which is more flexible than a literal match within the database.Changed[Oban.Config] All passed options are validated. Any unknown options will raise an ArgumentError and halt startup. This prevents misconfiguration through typos and passing unsupported options.[Oban.Worker] The perform/1 function now receives an Oban.Job struct as the sole argument, calling perform/1 again with only the args map if no clause catches the struct. This allows workers to use any attribute of the job to customize behaviour, e.g. the number of attempts or when a job was inserted into the database.The implementation is entirely backward compatible, provided workers are defined with the use macro. Workers that implement the Oban.Worker behaviour manually will need to change the signature of perform/1 to accept a job struct.[Oban] Child process names are generated from the top level supervisor's name, i.e. setting the name to &quot;MyOban&quot; on start_link/1 will set the notifier's name to MyOban.Notifier. This improves isolation and allows multiple supervisors to be ran on the same node.Fixed[Oban.Producer] Remove duplicate polling timers. As part of a botched merge conflict resolution two timers were started for each producer.","ref":"changelog.html#0-6-0-2019-07-26","title":"Changelog - 0.6.0 — 2019-07-26","type":"extras"},{"doc":"Added[Oban.Pruner] Added :prune_limit option to constrain the number of rows deleted on each prune iteration. This prevents locking the database when there are a large number of jobs to delete.Changed[Oban.Worker] Treat {:error, reason} tuples returned from perform/1 as a failure. The :kind value reported in telemetry events is now differentiated, where a rescued exception has the kind :exception, and an error tuple has the kind :error.Fixed[Oban.Testing] Only check available and scheduled jobs with the assert|refute_enqueued testing helpers.[Oban.Queue.Watchman] Catch process exits when attempting graceful shutdown. Any exits triggered within terminate/2 are caught and ignored. This safely handles situations where the producer exits before the watchman does.[Oban.Queue.Producer] Use circuit breaker protection for gossip events and call queries directly from the producer process. This prevents pushing all queries through the Notifier, and ensures more granular control over gossip errors.","ref":"changelog.html#0-5-0-2019-06-27","title":"Changelog - 0.5.0 — 2019-06-27","type":"extras"},{"doc":"Added[Oban] Add Oban.drain_queue/1 to help with integration testing. Draining a queue synchronously executes all available jobs in the queue from within the calling process. This avoids any sandbox based database connection issues and prevents race conditions from asynchronous processing.[Oban.Worker] Add backoff/1 callback and switch to exponential backoff with a base value as the default. This allows custom backoff timing for individual workers.[Oban.Telemetry] Added a new module to wrap a default handler for structured JSON logging. The log handler is attached by calling Oban.Telemetry.attach_default_logger/0 somewhere in your application code.[Oban.Queue.Producer] Guard against Postgrex errors in all producer queries using a circuit breaker. Failing queries will no longer crash the producer. Instead, the failure will be logged as an error and it will trip the producer's circuit breaker. All subsequent queries will be skipped until the breaker is enabled again approximately a minute later.This feature simplifies the deployment process by allowing the application to boot and stay up while Oban migrations are made. After migrations have finished each queue producer will resume making queries.Changed[Oban] Telemetry events now report timing as %{duration: duration} instead of %{timing: timing}. This aligns with the telemetry standard of using duration for the time to execute something.[Oban] Telemetry events are now broken into success and failure at the event level, rather than being labeled in the metadata. The full event names are now [:oban, :success] and [:oban, :failure].[Oban.Job] Rename scheduled_in to schedule_in for readability and consistency. Both the Oban docs and README showed schedule_in, which reads more clearly than scheduled_in.[Oban.Pruner] Pruning no longer happens immediately on startup and may be configured through the :prune_interval option. The default prune interval is still one minute.Fixed[Oban.Migrations] Make partial migrations more resilient by guarding against missing versions and using idempotent statements.","ref":"changelog.html#0-4-0-2019-06-10","title":"Changelog - 0.4.0 — 2019-06-10","type":"extras"},{"doc":"Migration Required (V2)Added[Oban] Allow setting queues: false or queues: nil to disable queue dispatching altogether. This makes it possible to override the default configuration within each environment, i.e. when testing.The docs have been updated to promote this mechanism, as well as noting that pruning must be disabled for testing.[Oban.Testing] The new testing module provides a set of helpers to make asserting and refuting enqueued jobs within tests much easier.Changed[Oban.Migrations] Explicitly set id as a bigserial to avoid mistakenly generating a uuid primary key.[Oban.Migrations] Use versioned migrations that are immutable. As database changes are required a new migration module is defined, but the interface of Oban.Migrations.up/0 and Oban.Migrations.down/0 will be maintained.From here on all releases with database changes will indicate that a new migration is necessary in this CHANGELOG.[Oban.Query] Replace use of (bigint) with (int, int) for advisory locks. The first int acts as a namespace and is derived from the unique oid value for the oban_jobs table. The oid is unique within a database and even changes on repeat table definitions.This change aims to prevent lock collision with application level advisory lock usage and other libraries. Now there is a 1 in 2,147,483,647 chance of colliding with other locks.[Oban.Job] Automatically remove leading &quot;Elixir.&quot; namespace from stringified worker name. The prefix complicates full text searching and reduces the score for trigram matches.","ref":"changelog.html#0-3-0-2019-05-29","title":"Changelog - 0.3.0 - 2019-05-29","type":"extras"},{"doc":"Added[Oban] Add pause_queue/2, resume_queue/2 and scale_queue/3 for dynamically controlling queues.[Oban] Add kill_job/2 for terminating running jobs and marking them as discarded.[Oban] Add config/0 for retreiving the supervisors config. This is primarily useful for integrating oban into external applications.[Oban.Queue.Producer] Use database triggers to immediately dispatch when a job is inserted into the oban_jobs table.[Oban.Job] Execution errors are stored as a jsonb array for each job. Each error is stored, not just the most recent one. Error entries contains these keys:at The utc timestamp when the error occurred atattempt The attempt number when the error ocurrederror A formatted error message and stacktrace, passed through Exception.blame/3[Oban.Config] Validate all options based on type and allowed values. Any invalid option will raise, preventing supervisor boot.[Oban.Notifier] Broadcast runtime gossip through pubsub, allowing any external system to get stats at the node and queue level.Changed[Oban.Queue.Supervisor] Set the min_demand to 1 for all consumers. This ensures that each queue will run the configured number of concurrent jobs. By default the min_demand is half of max_demand, which means a few slow jobs can prevent the queue from running the expected number of concurrent jobs.[Oban.Job] Change psuedo-states based on job properties into fixed states, this applies to scheduled and retryable.[Oban.Job] The &quot;Elixir&quot; prefix is stripped from worker names before storing jobs in the database. Module lookup performs the same way, but it cleans up displaying the worker name as a string.[Oban.Job] Accept all job fields as changeset parameters. While not encouraged for regular use, this is essential for testing various states.","ref":"changelog.html#0-2-0-2019-05-15","title":"Changelog - 0.2.0 - 2019-05-15","type":"extras"},{"doc":"[Oban] Initial release with base functionality.","ref":"changelog.html#0-1-0-2019-03-10","title":"Changelog - 0.1.0 - 2019-03-10","type":"extras"},{"doc":"Upgrading to v2.0This information is extracted and expanded from the CHANGELOG.This is a big release with a ton of new features and fixes, but also a few breaking changes. This guide will walk you through the upgrade process.","ref":"v2-0.html","title":"Upgrading to v2.0","type":"extras"},{"doc":"Update Oban to the latest version: defp deps do [ {:oban, &quot;~&gt; 2.0.0&quot;} ... ] endIf you're an Oban Web+Pro user you'll also need to bump ObanWeb and add ObanPro: defp deps do [ {:oban, &quot;~&gt; 2.0.0&quot;} {:oban_web, &quot;~&gt; 2.0.0&quot;, organization: &quot;oban&quot;} {:oban_pro, &quot;~&gt; 0.3.0&quot;, organization: &quot;oban&quot;} ... ] end","ref":"v2-0.html#bump-your-deps","title":"Upgrading to v2.0 - Bump Your Deps","type":"extras"},{"doc":"The perform/2 callback is replaced with perform/1, where the only argument is an Oban.Job struct. This unifies the interface for all Oban.Worker callbacks and helps to eliminate confusion around pattern matching on arguments.Change all worker definitions from accepting args and a job to only accept a job and match on the nested args key:- def perform(%{&quot;id&quot; =&gt; id}, _job) + def perform(%Job{args: %{&quot;id&quot; =&gt; id}})For finer control of job backoff the backoff/1 callback now expects a job struct instead of an integer. Change any workers that expect an attempt number to match on the full job struct:- def backoff(attempt) + def backoff(%Job{attempt: attempt})","ref":"v2-0.html#oban-worker-changes","title":"Upgrading to v2.0 - Oban.Worker Changes","type":"extras"},{"doc":"The :verbose setting was renamed to :log to better align with values accepted by Ecto.Repo.config :my_app, Oban, - verbose: false, + log: false ...Pruning is now handled by the new plugin system. Replace :prune in your config and pass a :max_age value to the plugin:config :my_app, Oban, - prune: {:max_age, 60}, - prune_interval: 30_000, - prune_limit: 5000, + plugins: [{Oban.Plugins.Pruner, max_age: 60}] ...In test mode you can disable pruning by setting plugins to false instead:config :my_app, Oban, - prune: false, + plugins: false ...🌟 ObanPro users may opt to use the DynamicPruner instead for finer control. For example, to set per-state retention periods:config :my_app, Oban, plugins: [{ Oban.Pro.Plugins.DynamicPruner, state_overrides: [ completed: {:max_age, {1, :day}}, discarded: {:max_age, {1, :month}} ] }] ...Pulse tracking and orphaned job rescue are removed from base Oban. This change means that you will need to manually rescue any jobs left executing state after a crash or forced shutdown.Remove any :beats_maxage, :rescue_after or :rescue_interval settings from your config:config :my_app, Oban, - beats_maxage: 10, - rescue_after: 30_000, - rescue_interval: 5_000 ...🌟 ObanPro users may use the Lifeline plugin to retain automatic orphaned job rescue with lightweight heartbeat recording:config :my_app, Oban, plugins: [Oban.Pro.Plugins.Lifeline] ...","ref":"v2-0.html#update-your-config","title":"Upgrading to v2.0 - Update Your Config","type":"extras"},{"doc":"The new perform_job test helper automates validating, normalizing and perform jobs in unit tests.To update your tests replace any calls to perform with the new helper:- assert :ok = MyApp.Worker.perform(%{id: 1}, %Oban.Job{}) + assert :ok = perform_job(MyApp.Worker, %{id: 1})The perform_job helper will verify the worker, the arguments and any provided options. It will then verify that your worker returns a valid result and return the value for you to assert on.Within integration tests replace drain_queue/3 with drain_queue/2:- Oban.drain_queue(:myqueue, with_safety: false) + Oban.drain_queue(queue: :myqueue, with_safety: false)","ref":"v2-0.html#update-your-tests","title":"Upgrading to v2.0 - Update Your Tests","type":"extras"},{"doc":"Telemetry event names have changed, along with some of the metadata and timing units. The new event names are consistent and align with the now-standard telemetry:span conventions.Update handlers to and match on the new event names:- def handle_event([:oban, :failure], measure, meta, _) do + def handle_event([:oban, :job, :exception], measure, meta, _) doAnd update the attach calls:- :telemetry.attach(&quot;oban&quot;, [[:oban, :failure]], &amp;handle_event/4, %{}) + :telemetry.attach(&quot;oban&quot;, [[:oban, :job, :exception]], &amp;handle_event/4, %{})Finally, replace references to meta.stack with meta.stacktrace:- Sentry.capture_exception(meta.error, stacktrace: meta.stack, extra: extra) + Sentry.capture_exception(meta.error, stacktrace: meta.stacktrace, extra: extra)Here is a conversion chart for any other handlers you may have:[:oban, :started] -&gt; [:oban, :job, :start] [:oban, :success] -&gt; [:oban, :job, :stop] [:oban, :failure] -&gt; [:oban, :job, :exception] [:oban, :trip_circuit] -&gt; [:oban, :circuit, :trip] [:oban, :open_circuit] -&gt; [:oban, :circuit, :open]","ref":"v2-0.html#update-telemetry","title":"Upgrading to v2.0 - Update Telemetry","type":"extras"},{"doc":"ObanWeb is streamlined to share configuration with Oban and use the new plugin system. Remove any ObanWeb specific configuration:- config :my_app, ObanWeb, repo: MyApp.RepoNext, add the necessary plugins to the Oban config:config :my_app, Oban, plugins: [ Oban.Pro.Plugins.DynamicPruner, Oban.Pro.Plugins.Lifeline, Oban.Web.Plugins.Stats ], ...Now remove ObanWeb from your application's supervision tree:children = [ MyApp.Repo, MyApp.Endpoint, {Oban, oban_opts}, - {ObanWeb, oban_web_opts} ]And delete any existing ObanWeb migrations:- defdelegate up, to: ObanWeb.Migrations - defdelegate down, to: ObanWeb.MigrationsFinally, use the oban_dashboard macro to easily mount in your router:+ import Oban.Web.Router scope &quot;/&quot; do pipe_through :browser - live &quot;/oban&quot;, ObanWeb.DashboardLive, layout: {ObanWeb.LayoutView, &quot;app.html&quot;} + oban_dashboard(&quot;/oban&quot;) end","ref":"v2-0.html#update-obanweb-optional","title":"Upgrading to v2.0 - Update ObanWeb (Optional)","type":"extras"},{"doc":"Overview🌟 Oban.Pro is a collection of plugins, workers and extensions that improve Oban's reliability even further and make difficult workflows possible. These are provided in a separate package that is available through a paid license.Get started with installation.","ref":"pro_overview.html","title":"Overview","type":"extras"},{"doc":"Lifeline — Records active queue details as heartbeats and periodically rescues orphaned jobs.DynamicPruner — Deletes completed or exhausted jobs based on configurable limits with detailed overrides.Reprioritizer — Prevents queue starvation by automatically adjusting priorities to ensure all jobs are eventually processed.","ref":"pro_overview.html#plugins","title":"Overview - Plugins","type":"extras"},{"doc":"Batch — Process jobs in groups while tracking overall progress across all nodes and executing optional callbacks.","ref":"pro_overview.html#workers","title":"Overview - Workers","type":"extras"},{"doc":"InstallationBefore continuing, be sure you have Oban up and running in your app!Oban.Pro is delivered as a private hex package named oban_pro. Before you can pull the package into your application you need to authenticate with the oban organization.$ mix hex.organization auth oban --key YOUR_OBAN_LICENSE_KEY⚠️ You'll also need to authenticate on any other development machines, build servers and CI instances.Now that you're authenticated you're ready to add oban_pro as a dependency for your application. Open mix.exs and add the following line:{:oban_pro, &quot;~&gt; 0.5&quot;, organization: &quot;oban&quot;}Now fetch your dependencies:$ mix deps.getThere isn't any direct configuration for Oban.Pro. Instead, you configure Oban to run plugins directly and use the various workers. For example, to use the Lifeline plugin you'd add it to the Oban config within config.exs:config :my_app, Oban, repo: MyApp.Repo, queues: [alpha: 10, gamma: 10, delta: 10], plugins: [Oban.Pro.Plugins.Lifeline]See the individual plugin docs for a detailed description of what they do, how to use them, and how they're configured.","ref":"pro_installation.html","title":"Installation","type":"extras"},{"doc":"Lifeline Plugin🌟 This plugin is available through Oban.ProThe Lifeline plugin records queue activity as heartbeats and periodically rescues orphaned jobs, i.e. jobs that are stuck in the executing state because the node was shut down before the job could finish. Without the Lifeline plugin you need to manually rescue jobs stuck in the executing state.Lifeline must be included used in order for Oban.Web to run properly.","ref":"lifeline.html","title":"Lifeline Plugin","type":"extras"},{"doc":"To use the Lifeline plugin add the module to your list of Oban plugins in config.exs:config :my_app, Oban, plugins: [Oban.Pro.Plugins.Lifeline] ...There isn't any configuration necessary. By default the plugin will record heartbeats every 1 second, prune old heartbeats every 5 minutes, and rescue orphaned jobs every 1 minute. If necessary you can configure any or all of those intervals:plugins: [{ Oban.Pro.Plugins.Lifeline, delete_interval: :timer.minutes(10), record_interval: :timer.seconds(10), rescue_interval: :timer.minutes(5) }]This configuration will record 10x fewer heartbeats per minute, retain them 2x as long and attempt to rescue 5x less frequently. It optimizes for less database activity, at the expense of fidelity and recovery speed.Note that rescuing orphans relies on recent heartbeats. Be sure that the delete_interval is always longer than the rescue_interval or it will look like all executing jobs are orphaned.","ref":"lifeline.html#using-and-configuring","title":"Lifeline Plugin - Using and Configuring","type":"extras"},{"doc":"When a job's attempt matches its max_attempts its retries are considered &quot;exhausted&quot;. Normally, the Lifeline plugin transitions exhausted jobs to the discarded state and they won't be retried again. It does this for a couple of reasons:To ensure at-most-once semantics. If a long running job interacted with a non idempotent service and was shut down while waiting for a reply you may not want that jot to retry.To prevent infinitely crashing BEAM nodes. Poorly behaving jobs may crash the node (through NIFs, memory exhaustion, etc.) We don't want to repeatedly rescue and rerun a job that repeatedly crashes the entire node.Discarding exhausted jobs may not always be desired. Use the retry_exhausted option if you'd prefer to retry exhausted jobs when they are rescued, rather than discarding them:plugins: [{Oban.Pro.Plugins.Lifeline, retry_exhausted: true}]During rescues, with retry_exhausted: true, a job's max_attempts is incremented and it is moved back to the available state.","ref":"lifeline.html#rescuing-exhausted-jobs","title":"Lifeline Plugin - Rescuing Exhausted Jobs","type":"extras"},{"doc":"Some additional notes about heartbeat records and how you can expect Lifeline to operate:Orphan rescuing is guaranteed to only rescue jobs that belong to dead queue processes or nodes.Heartbeat records are written to the oban_beats table very efficiently, as a single batch.Heartbeat records are only retained for five minutes, by default. This prevents bloat or exhausting the row limit on free tier databases.Only a single node will delete heartbeats or rescue orphans at any given time, which prevents potential deadlocks and churn.","ref":"lifeline.html#implementation-notes","title":"Lifeline Plugin - Implementation Notes","type":"extras"},{"doc":"The Lifeline plugin uses Oban.Telemetry.span/3 to emit standardized plugin events.eventmetadata[:oban, :plugin, :start]:plugin, :action[:oban, :plugin, :stop]:plugin, :action, :duration[:oban, :plugin, :exception]:plugin, :action, :duration, :kind, :reason, :stacktraceThe :action key is one of :delete, :record or :rescue.","ref":"lifeline.html#instrumenting-with-telemetry","title":"Lifeline Plugin - Instrumenting with Telemetry","type":"extras"},{"doc":"Dynamic Cron Plugin🌟 This plugin is available through Oban.ProThe DynamicCron plugin enhances Oban's built in cron scheduler by making it configurable at runtime, globally, across your entire cluster. DynamicCron supports adding, updating, deleting, and pausing cron entries at boot time or runtime. It is an ideal solution for applications that must dynamically start and manage scheduled tasks at runtime.","ref":"dynamic_cron.html","title":"Dynamic Cron Plugin","type":"extras"},{"doc":"Before running the DynamicCron plugin you must run a migration to add the oban_cron table to your database.mix ecto.gen.migration add_oban_cronOpen the generated migration in your editor and call the change function on Oban.Pro.Migrations.DynamicCron:defmodule MyApp.Repo.Migrations.AddObanCron do use Ecto.Migration defdelegate change, to: Oban.Pro.Migrations.DynamicCron endAs with the base Oban tables you can optionally provide a prefix to &quot;namespace&quot; the table within your database. Here we specify a &quot;private&quot; prefix:defmodule MyApp.Repo.Migrations.AddObanCron do use Ecto.Migration def change, do: Oban.Pro.Migrations.DynamicCron.change(&quot;private&quot;) endRun the migration to create the table:mix ecto.migrateNow we can use the DynamicCron plugin and start scheduling periodic jobs!","ref":"dynamic_cron.html#installation","title":"Dynamic Cron Plugin - Installation","type":"extras"},{"doc":"To begin using DynamicCron, add the module to your list of Oban plugins in config.exs:config :my_app, Oban, plugins: [Oban.Pro.Plugins.DynamicCron] ...By itself, without providing a crontab or dynamically inserting cron entries, the plugin doesn't have anything to schedule. To get scheduling started, provide a list of {cron, worker} or {cron, worker, options} tuples to the plugin. The syntax is identical to Oban's built in :crontab option, which means you can copying an existing standard :crontab list into the plugin's :crontab.plugins: [{ Oban.Pro.Plugins.DynamicCron, timezone: &quot;America/Chicago&quot;, crontab: [ {&quot;* * * * *&quot;, MyApp.MinuteJob}, {&quot;0 * * * *&quot;, MyApp.HourlyJob, queue: :scheduled}, {&quot;0 0 * * *&quot;, MyApp.DailyJob, max_attempts: 1}, {&quot;0 12 * * MON&quot;, MyApp.MondayWorker, tags: [&quot;scheduled&quot;]}, {&quot;@daily&quot;, MyApp.AnotherDailyWorker} ] }]For more details about periodic jobs and cron expressions see the documentation on Periodic Jobs.Now, when the dynamic pruner initializes, it will persist those cron entries to the database and start scheduling them according to their CRON expression. The plugin's crontab format is nearly identical to Oban's standard crontab, with a few important enhancements we'll look at soon.Each of the crontab entries are persisted to the database and referenced globally, by all the other connected Oban instances. That allows us to insert, update, or delete cron entries at any time. In fact, changing the schedule or options of an entry in the crontab provided to the plugin will automatically update the persisted entry. To demonstrate, let's modify the MinuteJob we specified so that it runs every other minute in the :scheduled queue:crontab: [ {&quot;*/2 * * * *&quot;, MyApp.MinuteJob, name: &quot;frequent&quot;, queue: :scheduled}, ... ]Now it isn't really a &quot;minute job&quot; any more, and the name is no longer suitable. That is why we added a :name override, so that we could update the name of the worker module as well.crontab: [ {&quot;*/2 * * * *&quot;, MyApp.FrequentJob, name: &quot;frequent&quot;, queue: :scheduled}, ... ]All entries are referenced by name, which defaults to the worker's name and must be unique. You may define the same worker multiple times as long as you provide a name override:crontab: [ {&quot;*/3 * * * *&quot;, MyApp.BasicJob, name: &quot;client-1&quot;, args: %{client_id: 1}}, {&quot;*/3 * * * *&quot;, MyApp.BasicJob, name: &quot;client-2&quot;, args: %{client_id: 2}}, ... ]To temporarily disable scheduling jobs you can set the paused flag:crontab: [ {&quot;* * * * *&quot;, MyApp.BasicJob, paused: true}, ... ]To resume the job you must supply paused: false (or use update/2 to resume it manually), simply removing the paused option will have no effect.crontab: [ {&quot;* * * * *&quot;, MyApp.BasicJob, paused: false}, ... ]It is also possible to delete a persisted entry during initialization by passing the :delete option:crontab: [ {&quot;* * * * *&quot;, MyApp.MinuteJob, delete: true}, ... ]One or more entries can be deleted this way. Deleting entries is idempotent, nothing will happen if no matching entry can be found.You can also use :delete to rename entries, if necessary:crontab: [ {&quot;* * * * *&quot;, MyApp.FrequentJob, name: &quot;frequent&quot;, delete: true}, {&quot;* * * * *&quot;, MyApp.FrequentJob, name: &quot;frequently&quot;}, ... ]However, it is recommended that you use the update/2 function to rename entries instead.In the next section we'll look at how to list, insert, update and delete jobs dynamically at runtime.","ref":"dynamic_cron.html#using-and-configuring","title":"Dynamic Cron Plugin - Using and Configuring","type":"extras"},{"doc":"Dynamic cron entries are persisted to the database, making it easy to manipulate them through typical CRUD operations. The DynamicCron plugin provides convenience functions to simplify working those operations. In this section we'll walk through each of the available functions and look at some examples.Typespecs📚 In order to bridge the gap between module level docs and a guide, each section includes a typespec for the corresponding function. The snippet below defines the types listed in each section.@type cron_expr :: binary() @type cron_name :: binary() | atom() @type cron_opt :: {:args, Oban.Job.args()} | {:max_attempts, pos_integer()} | {:paused, boolean()} | {:priority, 0..3} | {:name, cron_name()} | {:queue, atom() | binary()} | {:tags, Oban.Job.tags()} | {:worker, module()} @type cron_input :: {cron_expr(), module()} | {cron_expr(), module(), [cron_opt]}Inserting Entries@spec insert([cron_input()]) :: {:ok, [CronEntry.t()]} | {:error, Ecto.Changeset.t()}The insert/1 function takes a list of one or more tuples with the same {expression, worker} or {expression, worker, options} format as the plugin's crontab option:DynamicCron.insert([ {&quot;0 0 * * *&quot;, MyApp.GenericWorker}, {&quot;* * * * *&quot;, MyApp.ClientWorker, name: &quot;client-1&quot;, args: %{client_id: 1}}, {&quot;* * * * *&quot;, MyApp.ClientWorker, name: &quot;client-2&quot;, args: %{client_id: 2}}, {&quot;* * * * *&quot;, MyApp.ClientWorker, name: &quot;client-3&quot;, args: %{client_id: 3}} ])Be aware that insert/1 acts like an &quot;upsert&quot;, making it possible to modify existing entries if the worker or name matches. Still, it is better to use update/2 to make targeted updates.Updating Entries@spec update(cron_name(), [cron_opt()])) :: {:ok, CronEntry.t()} | {:error, Ecto.Changeset.t()}The update/2 function updates a single cron entry, as identified by the worker or name. Any option available when specifying an entry in the crontab list or when calling insert/2 can be updated—that includes the cron expression and the worker.The following call demonstrates updating every possible option:{:ok, _} = DynamicCron.update( &quot;cron-1&quot;, expression: &quot;1 * * * *&quot;, max_attempts: 10, name: &quot;special-cron&quot;, paused: false, priority: 0, queue: &quot;dedicated&quot;, tags: [&quot;client&quot;, &quot;scheduled&quot;], worker: Other.Worker )Naturally, individual options may be updated instead. For example, set paused: true to pause an entry:{:ok, _} = DynamicCron.update(MyApp.ClientWorker, paused: true)Since update/2 operates on a single entry at a time, it is possible to rename an entry without doing the delete/insert dance:# With the worker as the entry name {:ok, _} = DynamicCron.update(MyApp.ClientWorker, name: &quot;client-worker&quot;) # With a custom entry name already set {:ok, _} = DynamicCron.update(&quot;cron-1&quot;, name: &quot;special-cron&quot;)Deleting Entries@spec delete(cron_name()) :: {:ok, CronEntry.t()} | {:error, Ecto.Changeset.t()}The delete/1 function operates on individual entries, by worker or name. You can use it to delete entries at runtime, rather than hard-coding the :delete flag into the crontab list at compile time.# With the worker as the entry name {:ok, _} = DynamicCron.delete(Worker) # With a custom name {:ok, _} = DynamicCron.delete(&quot;cron-1&quot;)Listing EntriesUse all/0 to retrieve all persisted cron entries:entries = DynamicCron.all()This returns a list of Oban.Pro.Cron schemas with raw attributes. The all/0 function is provided as a convenience to inspect persisted entries.As Oban.Pro.Cron is an Ecto schema you're free to query the table however you wish using Ecto.Query. For example, you can list all of the entries with a name like &quot;client-&quot;:import Ecto.Query, only: [where: 3] Oban.Pro.Cron |&gt; where([c], ilike(c.name, &quot;client-%&quot;)) |&gt; MyApp.Repo.all()You can use functions like update or update_all to modify cron jobs in place, but it is highly recommended that you use update/2 to ensure that options are set correctly and to prevent breakage.","ref":"dynamic_cron.html#runtime-updates","title":"Dynamic Cron Plugin - Runtime Updates","type":"extras"},{"doc":"All DynamicCron functions have an alternate clause that accepts an Oban instance name as the first argument. This is in line with base Oban functions such as Oban.insert/2, which allow you to seamlessly work with multiple Oban instances and across multiple database prefixes. For example, you can use list/1 to list all cron entries for the instance named ObanPrivate:entries = DynamicCron.all(ObanPrivate)Likewise, to insert a new entry using the configuration associated with the ObanPrivate instance:{:ok, _} = DynamicCron.insert(ObanPrivate, [{&quot;* * * * *&quot;, PrivateWorker}])","ref":"dynamic_cron.html#isolation-and-namespacing","title":"Dynamic Cron Plugin - Isolation and Namespacing","type":"extras"},{"doc":"The DynamicCron plugin uses Oban.Telemetry.span/3 to emit standardized plugin events.eventmetadata[:oban, :plugin, :start]:plugin[:oban, :plugin, :stop]:plugin, :duration[:oban, :plugin, :exception]:plugin, :duration, :kind, :reason, :stacktrace","ref":"dynamic_cron.html#instrumenting-with-telemetry","title":"Dynamic Cron Plugin - Instrumenting with Telemetry","type":"extras"},{"doc":"Dynamic Pruning Plugin🌟 This plugin is available through Oban.ProThe DynamicPruner plugin enhances the default FixedPruner plugin's behaviour by allowing you to customize how long or how many jobs are retained in the jobs table. Where the FixedPruner will only retain jobs for 60 seconds, with the DynamicPruner you can specify either a maximum age or a maximum length and provide custom rules for specific queues, workers and job states.","ref":"dynamic_pruner.html","title":"Dynamic Pruner Plugin","type":"extras"},{"doc":"To start using the DynamicPruner add the module to your list of Oban plugins in config.exs:config :my_app, Oban, plugins: [Oban.Pro.Plugins.DynamicPruner] ...Without any additional options the pruner operates in maximum length mode (max_len) and retains a conservative 1,000 completed or discarded jobs. To increase the number of jobs retained you can provide your own mode configuration:plugins: [{Oban.Pro.Plugins.DynamicPruner, mode: {:max_len, 50_000}}]Now the pruner will retain the most recent 50,000 jobs instead.A fixed limit on the number of jobs isn't always ideal. Often you want to retain jobs based on their age instead. For example, if you your application needs to ensure that a duplicate job hasn't been enqueued within the past 24 hours you need to retain jobs for at least 24 hours; a fixed limit simply won't work. For that we can use maximum age (max_age) mode instead:plugins: [{Oban.Pro.Plugins.DynamicPruner, mode: {:max_age, 60 * 60 * 48}}]Here we've specified max_age using seconds, where 60 * 60 * 48 is the number of seconds in two days.Calculating the number of seconds in a period isn't especially readable, particularly when you have numerous max_age declarations in overrides (see below). For clarity you can specify the age's time unit as :second, :minute, :hour, :day or :month. Here is the same 48 hour configuration from above, but specified in terms of days:plugins: [{Oban.Pro.Plugins.DynamicPruner, mode: {:max_age, {2, :days}}}]Now you can tell exactly how long jobs should be retained, without reverse calculating how many seconds an expression represents.","ref":"dynamic_pruner.html#using-and-configuring","title":"Dynamic Pruner Plugin - Using and Configuring","type":"extras"},{"doc":"The mode option is indiscriminate when determining which jobs to prune. It pays no attention to which queue they are in, what worker the job is for, or which state they landed in. The DynamicPruner allows you to specify per-queue, per-worker and per-state overrides that fine tune pruning.We'll start with a simple example of limiting the total number of retained jobs in the events queue:plugins: [{ Oban.Pro.Plugins.DynamicPruner, mode: {:max_age, {7, :days}}, queue_overrides: [events: {max_len: 1_000}] }]With this configuration most jobs will be retained for seven days, but we'll only keep the latest 1,000 jobs in the events queue. We can extend this further and override all of our queues (and omit the default mode entirely):plugins: [{ Oban.Pro.Plugins.DynamicPruner, queue_overrides: [ default: {:max_age, {6, :hours}}, analysis: {:max_age, {1, :day}}, events: {:max_age, {10, :minutes}}, mailers: {:max_age, {2, :weeks}}, media: {:max_age, {2, :months}} ] }]When pruning by queue isn't granular enough you can provide overrides by worker instead:plugins: [{ Oban.Pro.Plugins.DynamicPruner, worker_overrides: [ &quot;MyApp.BusyWorker&quot;: {:max_age, {1, :day}}, &quot;MyApp.SecretWorker&quot;: {:max_age, {1, :second}}, &quot;MyApp.HistoricWorker&quot;: {:max_age, {1, :month}} ] }]You can also override by state, which allows you to keep discarded jobs for inspection while quickly purging cancelled or successfully completed jobs:plugins: [{ Oban.Pro.Plugins.DynamicPruner, state_overrides: [ cancelled: {:max_age, {1, :hour}}, completed: {:max_age, {1, :day}}, discarded: {:max_age, {1, :month}} ] }]Naturally you can mix and match overrides to finely control job retention:plugins: [{ Oban.Pro.Plugins.DynamicPruner, mode: {:max_age, {7, :days}}, queue_overrides: [events: {:max_age, {10, :minutes}}], worker_overrides: [&quot;MyApp.SecretWorker&quot;: {:max_age, {1, :second}}], state_overrides: [discarded: {:max_age, {2, :days}}] }]","ref":"dynamic_pruner.html#providing-overrides","title":"Dynamic Pruner Plugin - Providing Overrides","type":"extras"},{"doc":"With the default settings the DynamicPruner will only delete 10,000 jobs each time it prunes. The limit exists to prevent connection timeouts and excessive table locks. A busy system can easily insert more than 10,000 jobs per minute during standard operation. If you find that jobs are accumulating despite active pruning you can override the limit.Here we set the delete limit to 25,000:plugins: [{ Oban.Pro.Plugins.DynamicPruner, mode: {:max_len, 100_000}, limit: 25_000 }]Deleting in PostgreSQL is very fast, and the 10k default is rather conservative. Feel free to increase the limit to a number that your system can handle.","ref":"dynamic_pruner.html#keeping-up-with-inserts","title":"Dynamic Pruner Plugin - Keeping Up With Inserts","type":"extras"},{"doc":"Some additional notes about pruning in general and nuances of the DynamicPruner plugin:Pruning is best-effort and performed out-of-band. This means that all limits are soft; jobs beyond a specified age may not be pruned immediately after jobs complete.Pruning is only applied to jobs that are completed or discarded (has reached the maximum number of retries or has been manually killed). It'll never delete a new job, a scheduled job or a job that will be retried.Only a single node will prune at any given time, which prevents potential deadlocks between transactions.","ref":"dynamic_pruner.html#implementation-notes","title":"Dynamic Pruner Plugin - Implementation Notes","type":"extras"},{"doc":"The DynamicPruner plugin uses Oban.Telemetry.span/3 to emit standardized plugin events.eventmetadata[:oban, :plugin, :start]:plugin[:oban, :plugin, :stop]:plugin, :duration[:oban, :plugin, :exception]:plugin, :duration, :kind, :reason, :stacktrace","ref":"dynamic_pruner.html#instrumenting-with-telemetry","title":"Dynamic Pruner Plugin - Instrumenting with Telemetry","type":"extras"},{"doc":"Reprioritizer Plugin🌟 This plugin is available through Oban.ProUsing mixed priorities in a queue ensures that certain jobs are executed before others. For example, a queue that processes jobs from various customers may prioritize customers that are in a higher tier or plan. All high priority (0) jobs are guaranteed to run before any with lower priority (1..3), which is wonderful for the higher tier customers but can lead to resource starvation. When there is a constant flow of high priority jobs the lower priority jobs will never get the chance to run.The Reprioritizer plugin automatically adjusts lower job's priorities so that all jobs are eventually processed.","ref":"reprioritizer.html","title":"Reprioritizer Plugin","type":"extras"},{"doc":"To use the Reprioritizer plugin add the module to your list of Oban plugins in config.exs:config :my_app, Oban, plugins: [Oban.Pro.Plugins.Reprioritizer] ...Without any additional options the plugin will automatically increase the priority of any jobs that are available for 5 minutes or more. To reprioritize after less time waiting you can configure the :after time:plugins: [{Oban.Pro.Plugins.Reprioritizer, after: :timer.minutes(2)}]Now lower job priorities are bumped after 2 minutes of waiting, and every minute after that. To lower the reprioritization frequency you can change the :interval along with the :after time:plugins: [{ Oban.Pro.Plugins.Reprioritizer, after: :timer.minutes(10), interval: :timer.minutes(5) }]Here we've specified that job priority will increase every 5 minutes after the first 10 minutes of waiting.","ref":"reprioritizer.html#using-and-configuring","title":"Reprioritizer Plugin - Using and Configuring","type":"extras"},{"doc":"The after option applies to jobs for any workers across all queues. The Reprioritizer plugin allows you to specify per-queue and per-worker overrides that fine tune reprioritization.Let's configure reprioritization for the analysis queue so that it nudges jobs after only 1 minute:plugins: [{ Oban.Pro.Plugins.Reprioritizer, queue_overrides: [analysis: :timer.minutes(1)] }]We can also effectively disable reprioritization for all other queues by setting the period to :infinity:plugins: [{ Oban.Pro.Plugins.Reprioritizer, after: :infinity, queue_overrides: [analysis: :timer.minutes(1)] }]If per-queue overrides aren't enough we can override on a per-worker basis instead:plugins: [{ Oban.Pro.Plugins.Reprioritizer, interval: :timer.seconds(15), worker_overrides: [ &quot;MyApp.HighSLAWorker&quot;: :timer.seconds(30), &quot;MyApp.LowSLAWorker&quot;: :timer.minutes(10) ] }]Naturally you can mix and match overrides to finely control reprioritization:plugins: [{ Oban.Pro.Plugins.Reprioritizer, interval: :timer.minutes(2), after: :timer.minutes(5), queue_overrides: [media: :timer.minutes(10)], worker_overrides: [&quot;MyApp.HighSLAWorker&quot;: :timer.seconds(30)] }]","ref":"reprioritizer.html#providing-overrides","title":"Reprioritizer Plugin - Providing Overrides","type":"extras"},{"doc":"The Reprioritizer plugin uses Oban.Telemetry.span/3 to emit standardized plugin events.eventmetadata[:oban, :plugin, :start]:plugin[:oban, :plugin, :stop]:plugin, :duration[:oban, :plugin, :exception]:plugin, :duration, :kind, :reason, :stacktrace","ref":"reprioritizer.html#instrumenting-with-telemetry","title":"Reprioritizer Plugin - Instrumenting with Telemetry","type":"extras"},{"doc":"Batch Worker🌟 This worker is available through Oban.ProA Batch worker links the execution of many jobs as a group and runs optional callbacks after jobs are processed. This allows your application to coordinate the execution of tens, hundreds or thousands of jobs in parallel. It is built as an ergonomic abstraction over the top of the standard Oban.Worker.","ref":"batch.html","title":"Batch Worker","type":"extras"},{"doc":"While Batch workers are built within your application, they all rely on the BatchManager plugin from Oban.Pro. The BatchManager is responsible for tracking the execution of jobs within a batch and reliably enqueuing callback jobs.Start by adding BatchManager to your list of Oban plugins in config.exs:config :my_app, Oban, plugins: [Oban.Pro.Plugins.BatchManager]With the BatchManager plugin set to run we can create our batch workers. Let's define a worker that delivers daily emails in a large batch:defmodule MyApp.EmailBatch do use Oban.Pro.Workers.Batch, queue: :mailers @impl true def process(%Job{args: %{&quot;email&quot; =&gt; email}}) do MyApp.Mailer.daily_update(email) end endNote that we define a process/1 callback instead of perform/1 because perform/1 is used behind the scenes to coordinate regular execution and callbacks within the same worker. The process/1 function receives an Oban.Job struct, just like perform/1 would and it should return the same accepted values, i.e. :ok, {:ok, value}, {:error, error}.The process/1 function above only looks for an &quot;email&quot; key in the job args, but a &quot;batch_id&quot; is also available. We'll modify the function to extract the batch_id as well:def process(%Job{args: %{&quot;batch_id&quot; =&gt; batch_id, &quot;email&quot; =&gt; email}}) do {:ok, reply} = MyApp.Mailer.daily_update(email) track_delivery(batch_id, reply) :ok endNow the hypothetical track_delivery/2 function will store the delivery details for retrieval later, possibly by one of our handler callbacks.","ref":"batch.html#using-and-configuring","title":"Batch Worker - Using and Configuring","type":"extras"},{"doc":"The preferred way to create batches is to call new_batch/1,2 with a list of job args and a list of options:[%{email: &quot;foo@example.com&quot;}, %{email: &quot;bar@example.com&quot;}] |&gt; MyApp.EmailBatch.new_batch() |&gt; Oban.insert_all() # Generate the batch with options [%{email: &quot;foo@example.com&quot;}, %{email: &quot;bar@example.com&quot;}] |&gt; MyApp.EmailBatch.new_batch(schedule_in: 60, priority: 1, max_attempts: 3) |&gt; Oban.insert_all()The new_batch/1,2 function generates a list of job changesets and automatically injects a unique batch_id into each job's args. A Batch worker is a regular Oban.Worker under the hood, which means you can use new/2 to insert jobs as well, provided you use a deterministic batch id.","ref":"batch.html#inserting-batches","title":"Batch Worker - Inserting Batches","type":"extras"},{"doc":"After jobs in the batch are processed the BatchManager may insert callback jobs for the same worker. There are three batch handler callbacks that your worker may define:handle_attempted — called after all jobs in the batch were attempted at least once, regardless of whether they succeeded or not.handle_completed — called after all jobs in the batch have a completed state. This handler may never be called if one or more jobs keep failing or any are discarded.handle_discarded — called after any jobs in the batch have a discarded state.handle_exhausted — called after all jobs in the batch have either a completed or discarded state.Each handler callback receives a single Oban.Job struct with the batch_id as an argument and should return :ok. The callbacks are executed as separate isolated jobs, so they may be retried or discarded like any other job.Here we'll implement each of the optional handler callbacks and have them print out the batch status along with the batch_id:defmodule MyApp.BatchWorker do use Oban.Pro.Workers.Batch @impl true def handle_attempted(%Job{args: %{&quot;batch_id&quot; =&gt; batch_id}}) do IO.puts(&quot;Attempted \#{batch_id}&quot;) end @impl true def handle_completed(%Job{args: %{&quot;batch_id&quot; =&gt; batch_id}}) do IO.puts(&quot;Completed \#{batch_id}&quot;) end @impl true def handle_discarded(%Job{args: %{&quot;batch_id&quot; =&gt; batch_id}}) do IO.puts(&quot;Discarded \#{batch_id}&quot;) end @impl true def handle_exhausted(%Job{args: %{&quot;batch_id&quot; =&gt; batch_id}}) do IO.puts(&quot;Exhausted \#{batch_id}&quot;) end ... endThe callbacks receive the same batch_id as the regular job invocations, which enables patterns such as map/reduce.","ref":"batch.html#handler-callbacks","title":"Batch Worker - Handler Callbacks","type":"extras"},{"doc":"By default a batch_id is generated as a version 4 random UUID. This is more than sufficient to ensure that batches are unique between workers and nodes for any period of time. However, if you require more control you can override batch_id generation at the worker level or pass a value directly to the new_batch/2 function.To override the batch_id for a particular worker you override the gen_id callback:defmodule MyApp.BatchWorker do use Oban.Pro.Workers.Batch # Generate a 24 character long random string instead @impl true def gen_id do 24 |&gt; :crypto.strong_rand_bytes() |&gt; Base.encode64() end ... endThe gen_id/0 callback is suited for random/non-deterministic id generation. If you'd prefer to use a deterministic id instead you can pass the batch_id in as an option to new_batch/2:MyApp.BatchWorker.new_batch(list_of_args, batch_id: &quot;custom-batch-id&quot;)Using this technique you can verify the batch_id in tests or append to the batch manually after it was originally created. For example, you can add to a batch that is scheduled for the future:batch_id = &quot;daily-batch-\#{Date.utc_today()}&quot; midnight = Date.utc_today() |&gt; NaiveDateTime.new(~T[11:59:59]) |&gt; elem(1) |&gt; DateTime.from_naive!(&quot;Etc/UTC&quot;) # Create the initial batch initial_args |&gt; MyApp.BatchWorker.new_batch(batch_id: batch_id, schedule_at: midnight) |&gt; Oban.insert_all() # Add items to the batch later in the day %{batch_id: batch_id, other_arg: &quot;other&quot;} |&gt; MyApp.BatchWorker.new(schedule_at: midnight) |&gt; Oban.insert()When jobs in the batch execute later that day at midnight they'll all be tracked together.","ref":"batch.html#generating-batch-ids","title":"Batch Worker - Generating Batch IDs","type":"extras"},{"doc":"PostgreSQL's binary protocol has a limit of 65,535 parameters that may be sent in a single call. That presents an upper limit on the number of rows that may be inserted at one time, and therefor the number of jobs that may be inserted in a batch all at once.There is a simple workaround that will allow you to create arbitrarily large batches by inserting them in chunks.reducer = fn {changesets, index}, multi -&gt; Oban.insert_all(multi, &quot;batch_\#{index}&quot;, changesets) end list_of_args |&gt; MyApp.BatchWorker.new_batch() |&gt; Enum.chunk_every(5_000) |&gt; Enum.with_index() |&gt; Enum.reduce(Ecto.Multi.new(), reducer) |&gt; MyApp.Repo.transaction()A few words of explanation:This chunks every 5,000 jobs, you can use a smaller number but anything under 65,000 should work.The Enum.with_index/1 call is necessary to provide a unique name for each multi, without that you'll get an error about the batch names conflicting.The jobs are all inserted together within a transaction and won't start executing until they all go in.","ref":"batch.html#inserting-large-batches","title":"Batch Worker - Inserting Large Batches","type":"extras"},{"doc":"Callback jobs are only enqueued if your worker defines the corresponding callback, e.g. a worker that only defines handle_attempted/1 will only have a callback for that event.Callback jobs are unique, with an infinite period.The BatchManager uses debouncing to minimize queries and reduce overall load on your database.","ref":"batch.html#implementation-notes","title":"Batch Worker - Implementation Notes","type":"extras"},{"doc":"Workflow Worker🌟 This worker is available through Oban.ProWorkflow workers compose together with arbitrary dependencies between jobs, allowing sequential, fan-out and fan-in execution workflows. Workflows are fault tolerant, can be homogeneous or heterogeneous, and scale horizontally across all available nodes. Workflows are unique to Oban—only possible due to historic job retention.","ref":"workflow.html","title":"Workflow Worker","type":"extras"},{"doc":"Though you build Workflow workers within your application, they rely on Oban Pro's WorkflowManager plugin for coordination.Start by adding WorkflowManager to your Oban plugins in config.exs:config :my_app, Oban, plugins: [Oban.Pro.Plugins.WorkflowManager]The WorkflowManager doesn't require any configuration. Once added, you can start creating workflows. Each worker within a workflow must use Oban.Pro.Workers.Workflow rather than Oban.Worker, otherwise jobs will execute without any dependency management.Our example workflow will coordinate archiving an account. To properly archive the account we must perform the following workers:FinalReceipt — generate a final receipt and email it to the account ownerEmailSubscriber — email each subscriber on the accountBackupPost — up all important posts on the account to external storageDeleteAccount — delete the account from the databaseWe need step 1 to run before step 2 or 3, but both steps 2 and 3 can run in parallel. After all of the jobs in steps 2 and 3 finish we can finally delete the account from the database. For demonstration purposes we'll define one of the workers with a stubbed-in process/1 function:defmodule MyApp.FinalReceipt do use Oban.Pro.Workers.Workflow, queue: :default @impl true def process(%Job{} = _job) do # Generate the final receipt. :ok end endNote that we define a process/1 callback instead of perform/1 because the perform/1 function coordinates the workflow. The process/1 function receives an Oban.Job struct, just like perform/1 and it should return similar values, i.e. :ok, {:ok, value}.Assuming that you've defined the other workers, we can construct a workflow to enforce sequential execution:alias MyApp.{BackupPost, DeleteAccount, EmailSubscriber, FinalReceipt} def archive_account(acc_id) do FinalReceipt.new_workflow() |&gt; FinalReceipt.add(:receipt, FinalReceipt.new(%{id: acc_id})) |&gt; FinalReceipt.add(:email, EmailSubscriber.new(%{id: acc_id}), deps: [:receipt]) |&gt; FinalReceipt.add(:backup, BackupPost.new(%{id: acc_id}), deps: [:backup]) |&gt; FinalReceipt.add(:delete, DeleteAccount.new(%{id: acc_id}), deps: [:backup, :receipt]) |&gt; Oban.insert_all() endHere we use new_workflow/0,1 to initialize a new workflow (later on we will look at how to customize workflow behaviour with various options). Jobs are then added to the workflow with add/3,4, where the first argument is an existing workflow, the second argument is unique name, the third is a job changeset, and the fourth an optional list of dependencies. Finally, the workflow passes directly to Oban.insert_all/1,2, which handles inserting each of the jobs into the database atomically.All workers that use Workflow have new_workflow and add functions defined, so it doesn't matter which module you use to initialize or to add new jobs.Dependency resolution guarantees that jobs execute in the order receipt -&gt; email -&gt; backup -&gt; delete, even if one of the jobs fails and needs to retry. Any job that defines a dependency will wait for each upstream dependency to complete before it starts.Recall how our original specification stated that there could be multiple email and backup jobs? The workflow we've built only handles one email or backup job at a time.Let's modify the workflow to fan-out to multiple email and backup jobs and then fan-in to the final delete job:def archive_account_workflow(acc_id) do subscribers = subscribers_for_account(acc_id) documents = documents_for_account(acc_id) email_deps = Enum.map(subscribers, &amp;&quot;email_\#{&amp;1.id}&quot;) backup_deps = Enum.map(documents, &amp;&quot;backup_\#{&amp;1.id}&quot;) delete_deps = email_deps ++ backup_deps FinalReceipt.new_workflow() |&gt; FinalReceipt.add(:receipt, FinalReceipt.new(%{id: acc_id})) |&gt; add_email_jobs(subscribers) |&gt; add_backup_jobs(documents) |&gt; DeleteAccount.add(:delete, DeleteAccount.new(%{id: acc_id}), deps: delete_deps) end defp add_email_jobs(workflow, subscribers) do Enum.reduce(subscribers, workflow, fn %{id: id, email: email}, acc -&gt; EmailSubscriber.add(&quot;email_\#{id}&quot;, EmailSubscriber.new(%{email: email}), deps: [:receipt]) end) end defp add_backup_jobs(workflow, subscribers) do Enum.reduce(subscribers, workflow, fn %{id: id}, acc -&gt; BackupPost.add(&quot;backup_\#{id}&quot;, BackupPost.new(%{id: id}), deps: [:receipt]) end) endNow the workflow will run all of the email and backup jobs we need, before deleting the account. To confirm the flow without inserting and executing jobs we can visualize it.","ref":"workflow.html#using-and-configuring","title":"Workflow Worker - Using and Configuring","type":"extras"},{"doc":"Workflows are a type of Directed Acyclic Graph, also known as a DAG. That means we can describe a workflow as a graph of jobs and dependencies, where execution flows between jobs. By converting the workflow into DOT notation, a standard graph description language, we can render visualizations!Dot generation relies on libgraph, which is an optional dependency. You'll need to specify it as a dependency before generating dot output:def deps do [{:libgraph, &quot;~&gt; 0.7&quot;}] endOnce you've installed libgraph, we can use to_dot/1 to convert a workflow. As with new_workflow and add, all workflow workers define a to_dot/1 function that takes a workflow and returns a dot formatted string. For example, calling to_dot/1 with the account archiving workflow from above:FinalReceipt.to_dot(archive_account_workflow(123))Generates the following dot output, where each vertex is a combination of the job's name in the workflow and its worker module:strict digraph { &quot;delete (MyApp.DeleteAccount)&quot; &quot;backup_1 (MyApp.BackupPost)&quot; &quot;backup_2 (MyApp.BackupPost)&quot; &quot;backup_3 (MyApp.BackupPost)&quot; &quot;receipt (MyApp.FinalReceipt)&quot; &quot;email_1 (MyApp.EmailSubscriber)&quot; &quot;email_2 (MyApp.EmailSubscriber)&quot; &quot;backup_1 (MyApp.BackupPost)&quot; -&gt; &quot;delete (MyApp.DeleteAccount)&quot; [weight=1] &quot;backup_2 (MyApp.BackupPost)&quot; -&gt; &quot;delete (MyApp.DeleteAccount)&quot; [weight=1] &quot;backup_3 (MyApp.BackupPost)&quot; -&gt; &quot;delete (MyApp.DeleteAccount)&quot; [weight=1] &quot;receipt (MyApp.FinalReceipt)&quot; -&gt; &quot;backup_1 (MyApp.BackupPost)&quot; [weight=1] &quot;receipt (MyApp.FinalReceipt)&quot; -&gt; &quot;backup_2 (MyApp.BackupPost)&quot; [weight=1] &quot;receipt (MyApp.FinalReceipt)&quot; -&gt; &quot;backup_3 (MyApp.BackupPost)&quot; [weight=1] &quot;receipt (MyApp.FinalReceipt)&quot; -&gt; &quot;email_1 (MyApp.EmailSubscriber)&quot; [weight=1] &quot;receipt (MyApp.FinalReceipt)&quot; -&gt; &quot;email_2 (MyApp.EmailSubscriber)&quot; [weight=1] &quot;email_1 (MyApp.EmailSubscriber)&quot; -&gt; &quot;delete (MyApp.DeleteAccount)&quot; [weight=1] &quot;email_2 (MyApp.EmailSubscriber)&quot; -&gt; &quot;delete (MyApp.DeleteAccount)&quot; [weight=1] }Now we can take that dot output and render it using a tool like graphviz. The following example function accepts a workflow and renders it out as an SVG:defmodule WorkflowRenderer do alias Oban.Pro.Workers.Workflow def render(workflow) do dot_path = &quot;workflow.dot&quot; svg_path = &quot;workflow.svg&quot; File.write!(dot_path, Workflow.to_dot(workflow)) System.cmd(&quot;dot&quot;, [&quot;-T&quot;, &quot;svg&quot;, &quot;-o&quot;, svg_path, dot_path]) end endWith graphviz installed, that will generate a SVG of the workflow:&lt;g class=&quot;graph&quot;&gt;&lt;path fill=&quot;#fff&quot; stroke=&quot;transparent&quot; d=&quot;M0 188V0h1381.58v188H0z&quot;/&gt;&lt;g class=&quot;node&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;ellipse fill=&quot;none&quot; stroke=&quot;#000&quot; cx=&quot;663.44&quot; cy=&quot;-18&quot; rx=&quot;122.68&quot; ry=&quot;18&quot;/&gt;&lt;text text-anchor=&quot;middle&quot; x=&quot;663.44&quot; y=&quot;-14.3&quot; font-family=&quot;Times,serif&quot; font-size=&quot;14&quot;&gt;delete (MyApp.DeleteAccount)&lt;/text&gt;&lt;/g&gt;&lt;g class=&quot;node&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;ellipse fill=&quot;none&quot; stroke=&quot;#000&quot; cx=&quot;125.44&quot; cy=&quot;-90&quot; rx=&quot;125.38&quot; ry=&quot;18&quot;/&gt;&lt;text text-anchor=&quot;middle&quot; x=&quot;125.44&quot; y=&quot;-86.3&quot; font-family=&quot;Times,serif&quot; font-size=&quot;14&quot;&gt;backup_1 (MyApp.BackupPost)&lt;/text&gt;&lt;/g&gt;&lt;g class=&quot;edge&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;path fill=&quot;none&quot; stroke=&quot;#000&quot; d=&quot;M214.88-77.36c96.49 12.55 249.69 32.48 349.69 45.5&quot;/&gt;&lt;path stroke=&quot;#000&quot; d=&quot;M565.21-35.31l9.47 4.76-10.37 2.18.9-6.94z&quot;/&gt;&lt;/g&gt;&lt;g class=&quot;node&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;ellipse fill=&quot;none&quot; stroke=&quot;#000&quot; cx=&quot;394.44&quot; cy=&quot;-90&quot; rx=&quot;125.38&quot; ry=&quot;18&quot;/&gt;&lt;text text-anchor=&quot;middle&quot; x=&quot;394.44&quot; y=&quot;-86.3&quot; font-family=&quot;Times,serif&quot; font-size=&quot;14&quot;&gt;backup_2 (MyApp.BackupPost)&lt;/text&gt;&lt;/g&gt;&lt;g class=&quot;edge&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;path fill=&quot;none&quot; stroke=&quot;#000&quot; d=&quot;M452.15-73.98C494.34-63 551.7-48.08 596.01-36.55&quot;/&gt;&lt;path stroke=&quot;#000&quot; d=&quot;M597.03-39.9l8.8 5.91-10.56.87 1.76-6.78z&quot;/&gt;&lt;/g&gt;&lt;g class=&quot;node&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;ellipse fill=&quot;none&quot; stroke=&quot;#000&quot; cx=&quot;663.44&quot; cy=&quot;-90&quot; rx=&quot;125.38&quot; ry=&quot;18&quot;/&gt;&lt;text text-anchor=&quot;middle&quot; x=&quot;663.44&quot; y=&quot;-86.3&quot; font-family=&quot;Times,serif&quot; font-size=&quot;14&quot;&gt;backup_3 (MyApp.BackupPost)&lt;/text&gt;&lt;/g&gt;&lt;g class=&quot;edge&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;path fill=&quot;none&quot; stroke=&quot;#000&quot; d=&quot;M663.44-71.7v25.59&quot;/&gt;&lt;path stroke=&quot;#000&quot; d=&quot;M666.94-46.1l-3.5 10-3.5-10h7z&quot;/&gt;&lt;/g&gt;&lt;g class=&quot;node&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;ellipse fill=&quot;none&quot; stroke=&quot;#000&quot; cx=&quot;663.44&quot; cy=&quot;-162&quot; rx=&quot;118.08&quot; ry=&quot;18&quot;/&gt;&lt;text text-anchor=&quot;middle&quot; x=&quot;663.44&quot; y=&quot;-158.3&quot; font-family=&quot;Times,serif&quot; font-size=&quot;14&quot;&gt;receipt (MyApp.FinalReceipt)&lt;/text&gt;&lt;/g&gt;&lt;g class=&quot;edge&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;path fill=&quot;none&quot; stroke=&quot;#000&quot; d=&quot;M576.7-149.71c-95.97 12.48-250.38 32.57-351.35 45.71&quot;/&gt;&lt;path stroke=&quot;#000&quot; d=&quot;M225.51-100.49l-10.36-2.18 9.46-4.76.9 6.94z&quot;/&gt;&lt;/g&gt;&lt;g class=&quot;edge&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;path fill=&quot;none&quot; stroke=&quot;#000&quot; d=&quot;M606.39-146.15c-42.15 10.96-99.7 25.94-144.2 37.52&quot;/&gt;&lt;path stroke=&quot;#000&quot; d=&quot;M462.89-105.2l-10.56-.87 8.8-5.9 1.76 6.77z&quot;/&gt;&lt;/g&gt;&lt;g class=&quot;edge&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;path fill=&quot;none&quot; stroke=&quot;#000&quot; d=&quot;M663.44-143.7v25.59&quot;/&gt;&lt;path stroke=&quot;#000&quot; d=&quot;M666.94-118.1l-3.5 10-3.5-10h7z&quot;/&gt;&lt;/g&gt;&lt;g class=&quot;node&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;ellipse fill=&quot;none&quot; stroke=&quot;#000&quot; cx=&quot;944.44&quot; cy=&quot;-90&quot; rx=&quot;137.28&quot; ry=&quot;18&quot;/&gt;&lt;text text-anchor=&quot;middle&quot; x=&quot;944.44&quot; y=&quot;-86.3&quot; font-family=&quot;Times,serif&quot; font-size=&quot;14&quot;&gt;email_1 (MyApp.EmailSubscriber)&lt;/text&gt;&lt;/g&gt;&lt;g class=&quot;edge&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;path fill=&quot;none&quot; stroke=&quot;#000&quot; d=&quot;M722.35-146.33c44.17 11.01 104.81 26.11 151.56 37.76&quot;/&gt;&lt;path stroke=&quot;#000&quot; d=&quot;M874.81-111.95l8.86 5.81-10.55.98 1.69-6.79z&quot;/&gt;&lt;/g&gt;&lt;g class=&quot;node&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;ellipse fill=&quot;none&quot; stroke=&quot;#000&quot; cx=&quot;1236.44&quot; cy=&quot;-90&quot; rx=&quot;137.28&quot; ry=&quot;18&quot;/&gt;&lt;text text-anchor=&quot;middle&quot; x=&quot;1236.44&quot; y=&quot;-86.3&quot; font-family=&quot;Times,serif&quot; font-size=&quot;14&quot;&gt;email_2 (MyApp.EmailSubscriber)&lt;/text&gt;&lt;/g&gt;&lt;g class=&quot;edge&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;path fill=&quot;none&quot; stroke=&quot;#000&quot; d=&quot;M752.66-150.1c101.88 12.45 268.46 32.8 377.11 46.07&quot;/&gt;&lt;path stroke=&quot;#000&quot; d=&quot;M1130.27-107.5l9.5 4.69-10.35 2.26.85-6.95z&quot;/&gt;&lt;/g&gt;&lt;g class=&quot;edge&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;path fill=&quot;none&quot; stroke=&quot;#000&quot; d=&quot;M883.47-73.81c-44.37 11.05-104.5 26.03-150.69 37.54&quot;/&gt;&lt;path stroke=&quot;#000&quot; d=&quot;M733.4-32.82l-10.55-.98 8.85-5.81 1.7 6.79z&quot;/&gt;&lt;/g&gt;&lt;g class=&quot;edge&quot; transform=&quot;translate(4 184)&quot;&gt;&lt;path fill=&quot;none&quot; stroke=&quot;#000&quot; d=&quot;M1139.89-77.2L764.77-31.38&quot;/&gt;&lt;path stroke=&quot;#000&quot; d=&quot;M765.05-27.89l-10.35-2.26 9.5-4.69.85 6.95z&quot;/&gt;&lt;/g&gt;&lt;/g&gt;Looking at the visualized graph we can clearly see how the workflow starts with a single render job, fans-out to multiple email and backup jobs, and finally fans-in to the delete job—exactly as we planned!","ref":"workflow.html#visualizing-workflows","title":"Workflow Worker - Visualizing Workflows","type":"extras"},{"doc":"Workflows use conservative defaults for safe, and relatively quick, dependency resolution. You can customize the safety checks, waiting times and resolution intensity by providing a few top-level options:ignore_cancelled — regard cancelled dependencies as completed rather than halting the workflow. Defaults to false.ignore_discarded — regard discarded dependencies as completed rather than halting the workflow. Defaults to false.waiting_delay — the number of milliseconds to wait between dependency checks. This value, combined with the waiting_limit, dictates how long a job will await upstream dependencies. Defaults to 1000ms.waiting_limit — the number of times to retry dependency checks. This value, combined with waiting_delay, dictates how long a job will await upstream dependencies. Defaults to 10.waiting_snooze — the number of seconds a job will snooze after awaiting executing upstream dependencies. If upstream dependencies are scheduled or retryable then the job snoozes until the latest dependency's scheduled_at timestamp.The following example creates a workflow with all of the available options:workflow = MyWorkflow.new( ignore_cancelled: true, ignore_discarded: true, waiting_delay: :timer.seconds(5), waiting_limit: 5, waiting_snooze: 10 )Internally, the meta field stores options for each job. That makes it possible to set or override workflow options per-job. For example, configure a single job to ignore cancelled dependencies and another to snooze longer:MyWorkflow.new() |&gt; MyWorkflow.add(:a, MyWorkflow.new(%{})) |&gt; MyWorkflow.add(:b, MyWorkflow.new(%{}, meta: %{ignore_cancelled: true}), deps: [:a]) |&gt; MyWorkflow.add(:c, MyWorkflow.new(%{}, meta: %{waiting_snooze: 30}), deps: [:b])","ref":"workflow.html#workflow-options","title":"Workflow Worker - Workflow Options","type":"extras"},{"doc":"By default workflow_id is a version 4 random UUID. This is more than sufficient to ensure that workflows are unique for any period of time. However, if you require more control you can override workflow_id generation at the worker level, or pass a value directly to the new_workflow/2 function.To override the workflow_id for a particular workflow you override the gen_id callback:defmodule MyApp.Workflow do use Oban.Pro.Workers.Workflow # Generate a 24 character long random string instead @impl true def gen_id do 24 |&gt; :crypto.strong_rand_bytes() |&gt; Base.encode64() end ... endThe gen_id/0 callback works for random/non-deterministic id generation. If you'd prefer to use a deterministic id instead you can pass the workflow_id in as an option to new_workflow/2:MyApp.Workflow.new_workflow(workflow_id: &quot;custom-id&quot;)Using this technique you can verify the workflow_id in tests or append to the workflow manually after it was originally created.","ref":"workflow.html#generating-workflow-ids","title":"Workflow Worker - Generating Workflow IDs","type":"extras"},{"doc":"Jobs aren't executed until all upstream dependencies have completed. This includes waiting on retries, scheduled jobs or snoozing.Dependency resolution relies on jobs lingering in the database after execution. If your system prunes job dependencies then the workflow may never complete.","ref":"workflow.html#implementation-notes","title":"Workflow Worker - Implementation Notes","type":"extras"},{"doc":"ChangelogAll notable changes to Oban.Pro will be documented in this file.","ref":"pro-changelog.html","title":"Changelog","type":"extras"},{"doc":"[Oban.Plugins.BatchManager] Check for the correct output from Worker.from_string, fixing callback insertion.","ref":"pro-changelog.html#v0-5-2","title":"Changelog - v0.5.2","type":"extras"},{"doc":"[Oban.Plugins.DynamicPruner] Allow state overrides when pruning cancelled jobs.","ref":"pro-changelog.html#v0-5-1","title":"Changelog - v0.5.1","type":"extras"},{"doc":"[Oban.Pro.Workers.Workflow] A new worker that composes arbitrary jobs together for directed execution. This allows sequential, fan-out and fan-in workflows that are horizontally distributed and fault tolerant.See the Workflow docs for setup and a walk-through with use-cases.Bump Oban dependency to ~&gt; 2.3.0 due to the workflow worker's use of the new meta field in Oban.Job.","ref":"pro-changelog.html#v0-5-0","title":"Changelog - v0.5.0","type":"extras"},{"doc":"[Oban.Pro.Plugins.DynamicPruner] Always delete oldest job when pruning in max_len mode. In max_len mode, when job insertion exceeded job pruning the oldest jobs would persist between prunes.","ref":"pro-changelog.html#v0-4-2","title":"Changelog - v0.4.2","type":"extras"},{"doc":"ChangedUpgrade Oban dependency to ~&gt; 2.2.0 along with fixes for the move to Oban.Registry and Oban.Repo.","ref":"pro-changelog.html#v0-4-1","title":"Changelog - v0.4.1","type":"extras"},{"doc":"Added[Oban.Pro.Plugins.DynamicCron] The DynamicCron plugin is a replacement for Oban's built in cron scheduler. It supports adding, updating, deleting, pausing and resuming periodic jobs at runtime and boot time. It is an ideal solution for applications that must dynamically start and manage scheduled tasks at runtime.See the DynamicCron installation docs for instructions on getting started.[Oban.Pro.Plugins.Lifeline] Add :retry_exhausted option, which will retry exhausted jobs rather than discard them while rescuing.Breaking ChangeEmit standardized [:oban, :plugin, :start | :stop | :exception] events for all plugins. Previously, each plugin would emit its own event, which required extensive :telemetry.attach_many calls. Now all plugins have a standard event name with differentiated metatada, including a :plugin key with the module name responsible for emitting the event.To upgrade replace any events that used :prune, :lifeline or :reprioritizer with :plugin. For example, [:oban, :prune, :start] is now [:oban, :plugin, :start] with the metadata %{plugin: DynamicPruner}.","ref":"pro-changelog.html#v0-4-0","title":"Changelog - v0.4.0","type":"extras"},{"doc":"[Oban.Pro.Plugins.DynamicPruner] Correctly prune discarded jobs that were cancelled before they were ever attempted.","ref":"pro-changelog.html#v0-3-2","title":"Changelog - v0.3.2","type":"extras"},{"doc":"Fixed[Oban.Pro.Workers.Batch] Only invoke batch callbacks that are defined for the current worker.","ref":"pro-changelog.html#v0-3-1","title":"Changelog - v0.3.1","type":"extras"},{"doc":"Fixed[Oban.Pro.Plugins.Lifeline] Ignore unknown messages to the plugin to prevent crashes.ChangedBump Oban dependency to 2.0.0","ref":"pro-changelog.html#v0-3-0","title":"Changelog - v0.3.0","type":"extras"},{"doc":"AddedAdd handle_exhausted/1 callback to Oban.Pro.Workers.Batch worker. This new callback is triggered when all jobs are either completed or discarded.","ref":"pro-changelog.html#v0-2-1-2020-07-01","title":"Changelog - v0.2.1 — 2020-07-01","type":"extras"},{"doc":"FixedConditional improvements to the BatchManager pluginChangedBump Oban dependency to 2.0.0-rc.1","ref":"pro-changelog.html#v0-2-0-2020-06-12","title":"Changelog - v0.2.0 — 2020-06-12","type":"extras"},{"doc":"Initial release including these plugins:Oban.Pro.Plugins.BatchManagerOban.Pro.Plugins.DynamicPrunerOban.Pro.Plugins.LifelineOban.Pro.Plugins.ReprioritizerAs well as the Oban.Pro.Workers.Batch worker.","ref":"pro-changelog.html#v0-1-0-2020-06-03","title":"Changelog - v0.1.0 — 2020-06-03","type":"extras"},{"doc":"Overview🧭 Oban.Web is a view of Oban's inner workings that you host directly within your Phoenix application. Powered by Phoenix Live View, it is extremely lightweight and continuously updated. It builds on features provided by Oban.Pro and is available through a paid license.Get started with installation.","ref":"web_overview.html","title":"Overview","type":"extras"},{"doc":"Live Inspection — Monitor background job activity across all of your nodes in real time.Composable Filtering — Sift through jobs instantly with any combination of queue, state, node, worker and other metadata.Powerful Search — Intelligently search through job arguments instantly, with support for partial matches and auto-correction.Detailed Inspection — View job details including when, where and how it was ran (or how it failed to run).Batch Actions — Cancel, delete and retry selected jobs or all jobs matching the current filters.Queue Controls — Scale, pause, resume and stop queues across all running nodes with a couple of clicks.","ref":"web_overview.html#features","title":"Overview - Features","type":"extras"},{"doc":"InstallationBefore continuing, be sure you have Oban up and running in your app!Oban.Web is built with Phoenix Live View and it relies on a working installation of it in your application. If you don't have Live View installed, follow these instructions to get started.Oban.Web is delivered as a hex package named oban_web, which is published privately under the oban organization. The package is entirely self contained—it doesn’t hook into your asset pipeline at all.Before you can pull the package into your application you need to authenticate with the oban organization.$ mix hex.organization auth oban --key YOUR_OBAN_LICENSE_KEY⚠️ You'll also need to authenticate on any other development machines, build servers and CI instances.Now that you're authenticated you're ready to add oban_web as a dependency for your application. Open mix.exs and add the following line:{:oban_web, &quot;~&gt; 2.1&quot;, organization: &quot;oban&quot;}Now fetch your dependencies:$ mix deps.getThis will fetch both oban_web and oban_pro, if you haven't already installed oban_pro.Both the pro Lifeline plugin and the web Stats plugins are necessary for the dashboard to function properly. Add them to your Oban configuration in config.exs:config :my_app, Oban, repo: MyApp.Repo, queues: [alpha: 10, gamma: 10, delta: 10], plugins: [ Oban.Pro.Plugins.Lifeline, Oban.Web.Plugins.Stats ]After configuration you can mount the dashboard within your application's router.ex:# lib/my_app_web/router.ex use MyAppWeb, :router import Oban.Web.Router ... scope &quot;/&quot; do pipe_through :browser oban_dashboard &quot;/oban&quot; endHere we're using &quot;/oban&quot; as the mount point, but it can be anywhere you like. After you've verified that the dashboard is loading you'll probably want to restrict access to the dashboard via authentication, e.g. with Basic Auth.Installation is complete and you're all set! Start your Phoenix server, point your browser to where you mounted Oban and start monitoring your jobs.Running Multiple DashboardsApplications that run multiple Oban instances can mount a dashboard for each instance. Set the mounted dashboard's :oban_name to match the corresponding supervision tree's name. For example, given two configured Oban isntances, Oban and MyAdmin.Oban:config :my_app, Oban, repo: MyApp.Repo, name: Oban, ... config :my_admin, Oban, repo: MyAdmin.Repo, name: MyAdmin.Oban, ...You can then mount both dashboards in your router:scope &quot;/&quot; do pipe_through :browser oban_dashboard &quot;/oban&quot;, oban_name: Oban oban_dashboard &quot;/oban/admin&quot;, oban_name: MyAdmin.Oban endNote that the default name is Oban, setting oban_name: Oban in the example above was purely for demonstration purposes.Using LongPollingIf you're application is hosted in an environment that doesn't support websockets you can use longpolling as an alternate transport. To start, make sure that your live socket is configured for longpolling:socket &quot;/live&quot;, Phoenix.LiveView.Socket, longpoll: [connect_info: [session: @session_options], log: false]Then specify &quot;longpoll&quot; as your transport:scope &quot;/&quot; do pipe_through :browser oban_dashboard &quot;/oban&quot;, transport: &quot;longpoll&quot; endTrouble installing? Have questions?Take a look at the troubleshooting guide to see if your issue is covered. If not, or if need any help, stop by the #oban channel in Elixir Slack.","ref":"web_installation.html","title":"Installation","type":"extras"},{"doc":"TroubleshootingAuthorizing with HerokuIf your app runs on Heroku using the Elixir Buildpack (rather than Docker) you'll need to use compilation hook to authorize hex before fetching dependencies.First, if you haven't already, set your license key on Heroku:heroku config:set OBAN_WEB_LICENSE_KEY=&quot;YOUR OBAN WEB LICENSE KEY&quot;Next, add a small shell script to your application in ./bin/predeps:#!/bin/bash mix hex.organization auth oban --key &quot;$OBAN_WEB_LICENSE_KEY&quot;Finally, set the predeps script within elixir_buildpack.config:hook_pre_fetch_dependencies=&quot;./bin/predeps&quot;Thanks to Jessie Cooke for the solution to this situation!","ref":"web_troubleshooting.html","title":"Troubleshooting","type":"extras"},{"doc":"ChangelogAll notable changes to Oban.Web will be documented in this file.","ref":"web-changelog.html","title":"Changelog","type":"extras"},{"doc":"ChangedUpgrade minimum Phoenix Live View dependency to 0.15.FixedAllow retrying or deleting cancelled jobs when they were never attempted.","ref":"web-changelog.html#v2-3-1","title":"Changelog - v2.3.1","type":"extras"},{"doc":"AddedCancelling a job transitions the job to the cancelled state rather than discarded. The discarded state is now reserved for jobs that exhaust retry attempts or are purposefully discarded through a {:discard, reason} tuple.Display meta in the job details view.ChangedUpgrade Oban dependency to ~&gt; 2.3.0 to support the new cancelled state, and meta field.","ref":"web-changelog.html#v2-3-0","title":"Changelog - v2.3.0","type":"extras"},{"doc":"ChangedReplace the queue scale slider with a number input and a submit button. Aside from how difficult it was to scale accurately with a slider, it would fire erroneous update events due to DOM changes.","ref":"web-changelog.html#v2-2-3","title":"Changelog - v2.2.3","type":"extras"},{"doc":"ChangedUpgrade Oban dependency to ~&gt; 2.2.0 along with fixes for the move to Oban.Registry and Oban.Repo.","ref":"web-changelog.html#v2-2-2","title":"Changelog - v2.2.2","type":"extras"},{"doc":"FixedPrevent layout jumb between initial render and subsequent mount. This was caused by the lack of a closing tag that was later injected on DOM update.Prevent bleeding events between nested components. The lack of a closing tag seemed to confuse event handling and allowed the wrong component to receive events.Correct touch event handling in the sidebar for mobile devices.","ref":"web-changelog.html#v2-2-1","title":"Changelog - v2.2.1","type":"extras"},{"doc":"AddedSupport multiple isolated dashboards by providing an oban_name option to oban_dashboard/2 calls in the Phoenix router.For example, to mount dashboards for multiple Oban instances in an umbrella:defmodule MyAppWeb.Router do use Phoenix.Router import Oban.Web.Router scope &quot;/&quot;, MyAppWeb do pipe_through [:browser] oban_dashboard &quot;/oban/web&quot;, oban_name: MyWeb.Oban oban_dashboard &quot;/oban/ops&quot;, oban_name: MyOps.Oban oban_dashboard &quot;/oban/ingestion&quot;, oban_name: MyIngestion.Oban end endChangedClicking anywhere on a job row now opens the details view. Previously only clicking the rightmost &quot;expand&quot; icon opened details, and clicking on the worker name would search for other jobs with the same worker. Now the rightmost &quot;document search&quot; icon performs a related job search.Close the job details view when a user clicks a node name, job state, or queue name in the sidebar.Prevent ugly overflow issues when there are long node or queue names in the sidebar. Note that the full name is still visible on hover.","ref":"web-changelog.html#v2-2-0","title":"Changelog - v2.2.0","type":"extras"},{"doc":"FixedCorrect expected plugin naming scheme for compatibility with Oban &gt;= 2.1.Increase stats collection and activation timeouts to compensate for spikes in table size or additional load on the database pool from other processes.","ref":"web-changelog.html#v2-1-1","title":"Changelog - v2.1.1","type":"extras"},{"doc":"AddedSupport alternate live socket transports, namely the &quot;longpoll&quot; transport. Users can specify either &quot;websocket&quot; or &quot;longpoll&quot; directly from the oban_dashboard/2 call in their Phoenix router (which avoids global configuration).FixedSkip full text search for PostgreSQL versions &lt; 11.0. Older PG versions don't have proper support for JSONB vectorization or web style search operations.","ref":"web-changelog.html#v2-1-0","title":"Changelog - v2.1.0","type":"extras"},{"doc":"ChangedUpgrade Oban to 2.0.0, LiveView 0.14 and ObanPro to 0.3.0Add simple load less/load more pagination for browsing through jobsMove details view inline, eliminating modal sizing and scrolling issuesDisplay queue_time and run_time in the job details viewRestore worker filtering using auto-populated search terms","ref":"web-changelog.html#v2-0-0","title":"Changelog - v2.0.0","type":"extras"},{"doc":"ChangedUpgrade to Oban 2.0.0-rc.1, LiveView ~&gt; 0.13, and add a dependency on Oban Pro.Introduce Oban.Web.Router for more convenient and flexible dashboard routing. The new module provides an oban_dashboard/2 macro, which prevents scoping mistakes and provides the foundation of better route support.Here is how to mount a dashboard at /oban with the new macro:defmodule MyAppWeb.Router do use Phoenix.Router import Oban.Web.Router scope &quot;/&quot;, MyAppWeb do pipe_through [:browser] oban_dashboard &quot;/oban&quot; end endAdd bulk actions for all jobs in table view. It's now possible to select one or more jobs in the table and then cancel, run, retry or delete them all.Add a dropdown that controls the refresh rate of jobs and stats in the dashboard. Lowering the refresh rate can reduce query overhead, or pause updates entirely.Expose pause/resume controls and a slider for scaling in the queue side bar.Eliminate migrations. Migrations are no longer necessary for updates or full text searching. You may safely undo previous migrations.Consider job tags along with args and worker name when searching.More intelligent search using &quot;and&quot;, &quot;or&quot; and respecting parenthesis.Consistently toggle filters on and off from the sidebar, rather than matching on a filter list in the header.","ref":"web-changelog.html#v2-0-0-alpha-0","title":"Changelog - v2.0.0-alpha.0","type":"extras"},{"doc":"ChangedUpgrade to Phoenix ~&gt; 1.5, LiveView ~&gt; 0.12 and PubSub ~&gt; 2.0. None of these upgrades required changes to ObanWeb, they are meant to enable upgrades for host applications.","ref":"web-changelog.html#v1-5-0-2020-04-27","title":"Changelog - v1.5.0 — 2020-04-27","type":"extras"},{"doc":"ChangedUpgrade to LiveView ~&gt; 0.10 along with requisite changes to use @inner_content in the layout template. This prevents the view from hanging with a blank screen on load.","ref":"web-changelog.html#v1-4-0-2020-03-24","title":"Changelog - v1.4.0 — 2020-03-24","type":"extras"},{"doc":"FixedPrevent FunctionClauseError when closing the dashboard before it has finished mounting.","ref":"web-changelog.html#v1-3-1-2020-03-18","title":"Changelog - v1.3.1 — 2020-03-18","type":"extras"},{"doc":"ChangedUpgrade to LiveView ~&gt; 0.9 along with the requisite changes to flash handling. Note, this requires pipelines in the host app to use fetch_live_flash instead of fetch_flash.","ref":"web-changelog.html#v1-3-0-2020-03-10","title":"Changelog - v1.3.0 — 2020-03-10","type":"extras"},{"doc":"FixesPrevent flash map key issue on initial mount. The lack of a show key on the empty flash map caused runtime errors.Prevent losing stats cache on restart by shifting table ownership into the supervisor.ChangedClear completed_at and discarded_at timestamps when descheduling jobs.Upgrade to LiveView ~&gt; 0.7.AddedDisplay job priority value in the table and detail views.Display job tags in the detail view.","ref":"web-changelog.html#v1-2-0-2020-02-07","title":"Changelog - v1.2.0 — 2020-02-07","type":"extras"},{"doc":"FixesCorrect syntax used for descheduling and discarding jobsSet the discarded_at timestamp when discarding jobs","ref":"web-changelog.html#v1-1-2-2020-02-07","title":"Changelog - v1.1.2 — 2020-02-07","type":"extras"},{"doc":"FixesFix job detail refreshing on tick events","ref":"web-changelog.html#v1-1-1-2020-02-06","title":"Changelog - v1.1.1 — 2020-02-06","type":"extras"},{"doc":"ChangesAdd verbose setting to control log levels. This command mirrors the behavior in Oban and is respected by all queries.Deprecate stats configuration. The stats module is entirely overhauled so that it only refreshes when one or more users are connected. That prevents it from using any connections or performing any queries while testing, which renders the stats option pointless.Add stats_interval to control how often node and queue counts are refreshed. The default value is every 1s.Add tick_interval to control how often the jobs table and job details are refreshed. The default value is every 500ms.","ref":"web-changelog.html#v1-1-0-2020-02-06","title":"Changelog - v1.1.0 — 2020—02—06","type":"extras"},{"doc":"FixesDisplay discarded_at time in detail timeline view.Prevent jitter when clearing stats cache.","ref":"web-changelog.html#v1-0-1-2020-01-29","title":"Changelog - v1.0.1 — 2020-01-29","type":"extras"},{"doc":"ChangedUpgrade Oban and set dependency to ~&gt; 1.0.0Update stats and the queries that power stats to rely on more frequent refreshes and fewer database calls. The average refresh time dropped from ~310ms to ~180ms.","ref":"web-changelog.html#v1-0-0-2020-01-29","title":"Changelog - v1.0.0 — 2020-01-29","type":"extras"},{"doc":"ChangedUpgrade to LiveView ~&gt; 0.5 and test with 0.6.","ref":"web-changelog.html#v0-8-0-2020-01-23","title":"Changelog - v0.8.0 — 2020-01-23","type":"extras"},{"doc":"AddedAdd detail modal for inspecting job arguments, errors and state timings.","ref":"web-changelog.html#v0-7-0-2020-01-08","title":"Changelog - v0.7.0 — 2020-01-08","type":"extras"},{"doc":"ChangedDisplay duration or distance in words depending on job state. This clarifies timestamp information. Relative timestamps have an formatted absolute value as the &quot;title&quot; attribute as well.Improved ordering for various job states. Many states were reversed, making the table view seem broken.FixesEnsure all assigns are available on render when a disconnected node reconnects. This fixes an issue when viewing the UI in development.","ref":"web-changelog.html#v0-6-3-2019-12-15","title":"Changelog - v0.6.3 — 2019-12-15","type":"extras"},{"doc":"ChangedAdd support for explicitly disabling stats rather than inferring based on queue configuration. This prevents issues in production environments where certain nodes (i.e. web) aren't processing any queues, but the the UI is accessed.ObanWeb configuration now differs from Oban configuration and you'll need to specify it separately. Specify the repo in config.exs:config :my_app, ObanWeb, repo: MyApp.RepoDisable the stats tracker in the test environment:config :my_app, ObanWeb, stats: false","ref":"web-changelog.html#v0-6-2-2019-12-05","title":"Changelog - v0.6.2 — 2019-12-05","type":"extras"},{"doc":"FixesIgnore the stats server when queues is empty.","ref":"web-changelog.html#v0-6-1-2019-11-22","title":"Changelog - v0.6.1 — 2019-11-22","type":"extras"}]